#ifndef CF63oddtf_H
#define CF63oddtf_H

#include "lcd.h"
#include "report.h"
#include <sys/time.h>
#include <limits.h>
#include <pthread.h>
#include <sched.h> /* procjiffy */
#include <stdint.h> /* for types used by cpufreq routines */
// for load_type and tops 
#include "../../clients/lcdproc/machine.h"
//#define TOPCOMPARE // in store_tops.c to verify top calcs
#define CLIENT_CHILD

// disabled oct 14 07 to see if bad file descriptor goes away with this
// enabled  oct 17    with error trapping on write in commit_packet
//                    has yet to crash with this disabled
#define TIMING_REPORT
// live timing may presume platform &| kernel dependent
// signal handling and multithreading capabilities
// & presumes, is wholly inside TIMING_REPORT
#ifdef TIMING_REPORT
#define LIVE_TIMING 
#endif
#ifdef __linux__
#define DO_ULT  /* top_num and topname implementations in store_tops / machine_get_jiffies are linux only */
#endif /* __linux__ */ 
#ifdef LMSENSORS 
#define DO_LM
#endif
#define DO_HDD
#define CF635LEDS /* leds on a 635: not neccessary & unused by other models */

#ifdef CLIENT_CHILD
//#define MY_CUSTOM_LINE
#define PROTOCOL_CHEATING
//#define MAXLINES 44  // 32 temps + 8 fans + 4 possible fan temps
//#define MAXLINES 40+MAX_TOP+MAX_HDD+MAXLM +1   // + 1 custom line : MAXLM yet defined here...
/* FIXME: why not use a malloc for MAXLINES ? */ 
#define MAXLINES 61  //  put conditional defs together in one line ?  
#define  DEFAULT_WIDGET    1  // create a defalt widget or no
#endif // CLIENT_CHILD

#ifdef DO_LM
#include <sensors/sensors.h>
#define MAX_LM      8  // max num lm-sensor items
#endif // DO_LM
#ifdef DO_ULT
#define MAX_TOP     4  // max num to top by rank or name
#endif // DO_ULT
#ifdef DO_HDD
#define MAX_HDD    4  // max # drives
#endif // DO_HDD

//  0 RPT_CRIT,  1 RPT_ERR,  2 RPT_WARNING,  3 RPT_NOTICE,  4 RPT_INFO,  5 RPT_DEBUG
#define RPT_QUE    RPT_DEBUG
#define RPT_CLIALL RPT_DEBUG 
#define RPT_CLICRE RPT_INFO
#define RPT_CLIUP  RPT_DEBUG
#define RPT_AWAIT  RPT_DEBUG 
#define RPT_FANS   RPT_DEBUG 
//#define  FANS_USEC  // special tracking of fan management to the microsecond
#define RPT_TEMPS  RPT_DEBUG
#define RPT_PCHEAT RPT_DEBUG 
#define RPT_LMS    RPT_DEBUG

#define DEFAULT_CELL_WIDTH   6
#define DEFAULT_CELL_HEIGHT  8
#define DEFAULT_CONTRAST     560
#define DEFAULT_DEVICE       "/dev/lcd"
#define DEFAULT_SPEED        B19200
#define DEFAULT_BRIGHTNESS   1000
#define DEFAULT_OFFBRIGHTNESS  0
#define DEFAULT_SIZE        "16x2"
#define DEFAULT_SIZE_CF635  "20x4"
#ifdef CF635LEDS
#define CFLEDBASE 6   // steps from 0 to 100 ( off>green>amber>red ) in 635 led lighting 
#endif
#define DEFAULT_SIZE_CF633  "16x2"
#define DEFAULT_SIZE_CF631  "20x2"

#define DEFAULT_TEMP_REPORT    0
#define DEFAULT_SYSLOG_SENSORS 0
#define DEFAULT_DS18_NAME     "T"  
#define DEFAULT_FAN_NAME      "fan"  
#define DEFAULT_FAN_AUTOREPORT 0
#define DEFAULT_FAN_POWER      100
#define DEFAULT_FAN_LOG_POW    0
#define DEFAULT_FAN_DIVISOR    2
#define DEFAULT_FAN_D_LABEL   "RPM"
#define DEFAULT_FAN_FAILSAFE   1
#define DEFAULT_FAN_GLITCH     0
#define DEFAULT_FAN_SENS_NUM     0    // 0=none; 1-32 possible  
#define DEFAULT_FAN_BY_S_OFF   15   // (deg C  ) 
#define DEFAULT_FAN_BY_S_100   25 
#define DEFAULT_FAN_SAFETY     30   // this much warmer than defaults gets rejected
#define DEFAULT_FAN_A_SEC      30
#define DEFALT_TOPNUM          0    // how deep into ranked top to log


#define TRY_AGAIN 0
#define GOOD_MSG 1
#define GIVE_UP 2
#define CF633_Ping_Command                                 0
#define CF633_Get_Hardware_And_Firmware_Version            1
#define CF633_Write_User_Flash_Area                        2
#define CF633_Read_User_Flash_Area                         3
#define CF633_Store_Current_State_As_Boot_State            4
#define CF633_Reboot                                       5
#define CF633_Clear_LCD_Screen                             6
#define CF633_Set_LCD_Contents_Line_One                    7
#define CF633_Set_LCD_Contents_Line_Two                    8
#define CF633_Set_LCD_Special_Character_Data               9
#define CF633_Read_Height_Bytes_of_LCD_Memory             10
#define CF633_Set_LCD_Cursor_Position                     11
#define CF633_Set_LCD_Cursor_Style                        12
#define CF633_Set_LCD_Contrast                            13
#define CF633_Set_LCD_And_Keypad_Backlight                14
#define CF633_Query_Last_Fan_Pulse_Count                  15
#define CF633_Set_Up_Fan_Reporting                        16
#define CF633_Set_Fan_Power                               17
#define CF633_Read_DOW_Device_Information                 18
#define CF633_Set_Up_Temperature_Reporting                19
#define CF633_Arbitrary_DOW_Transaction                   20
#define CF633_Set_Up_Live_Fan_or_Temperature_Display      21
#define CF633_Send_Command_Directly_to_the_LCD_Controller 22
#define CF633_Configure_Key_Reporting                     23
#define CF633_Read_Keypad_Polled_Mode                     24
#define CF633_Set_Fan_Power_Failsafe                      25
#define CF633_Set_Fan_Tachometer_Glitch_Filter            26
#define CF633_Query_Fan_Power_and_Failsafe_Mask           27
#define CF633_Set_ATX_Power_Switch_Functionality          28
#define CF633_Enable_Disable_and_Reset_the_Watchdog       29
#define CF633_Read_Reporting_and_Status                   30
#define CF633_Send_Data_to_LCD                            31
#define CF633_Key_Legends                                 32
#define CF633_Set_Baud_Rate                               33
#define CF633_Set_GPIO_Pin                                34
#define CF633_Read_GPIO_Pin                               35
#define CF633_Dummy_Unused                                36

/* Those are all the possible keys we know off */
#define KEY_UP_PRESS       1
#define KEY_DOWN_PRESS     2
#define KEY_LEFT_PRESS     3
#define KEY_RIGHT_PRESS    4
#define KEY_ENTER_PRESS    5
#define KEY_EXIT_PRESS     6
#define KEY_UP_RELEASE     7
#define KEY_DOWN_RELEASE   8
#define KEY_LEFT_RELEASE   9
#define KEY_RIGHT_RELEASE  10
#define KEY_ENTER_RELEASE  11
#define KEY_EXIT_RELEASE   12

#define DEFAULT_DS18_TEMP   -1000  // a bad CRC    
#define BAD_CRC_CUT  1 // how many bad crcs to deconfigure on
#define FAN_LOW_POWER_PERCENT  60  // cutoff for low power

MODULE_EXPORT int  CF63oddtf_init (Driver *drvthis);
MODULE_EXPORT void CF63oddtf_close (Driver *drvthis);
MODULE_EXPORT int  CF63oddtf_width (Driver *drvthis);
MODULE_EXPORT int  CF63oddtf_height (Driver *drvthis);
MODULE_EXPORT int  CF63oddtf_cellwidth (Driver *drvthis);
MODULE_EXPORT int  CF63oddtf_cellheight (Driver *drvthis);
MODULE_EXPORT void CF63oddtf_clear (Driver *drvthis);
MODULE_EXPORT void CF63oddtf_flush (Driver *drvthis);
MODULE_EXPORT void CF63oddtf_string (Driver *drvthis, int x, int y, char string[]);
MODULE_EXPORT void CF63oddtf_chr (Driver *drvthis, int x, int y, char c);
MODULE_EXPORT char *CF63oddtf_get_key (Driver *drvthis);
MODULE_EXPORT void CF63oddtf_vbar (Driver *drvthis, int x, int y, int len, int promille, int options);
MODULE_EXPORT void CF63oddtf_hbar (Driver *drvthis, int x, int y, int len, int promille, int options);
MODULE_EXPORT void CF63oddtf_num (Driver *drvthis, int x, int num);
MODULE_EXPORT int  CF63oddtf_icon(Driver *drvthis, int x, int y, int icon);
MODULE_EXPORT void CF63oddtf_set_char (Driver *drvthis, int n, char *dat);
MODULE_EXPORT int  CF63oddtf_get_contrast (Driver *drvthis);
MODULE_EXPORT void CF63oddtf_set_contrast (Driver *drvthis, int promille);
MODULE_EXPORT int  CF63oddtf_get_brightness (Driver *drvthis, int state);
MODULE_EXPORT void CF63oddtf_set_brightness (Driver *drvthis, int state, int promille);
MODULE_EXPORT void CF63oddtf_backlight (Driver *drvthis, int on);

#ifdef FIXMECF635LEDS
MODULE_EXPORT void CF63oddtf_output (Driver *drvthis, int state);
#endif

typedef unsigned char ubyte;
typedef signed char sbyte;
typedef unsigned short word;
typedef unsigned long dword;

/* KeyRing management */
#define KEYRINGSIZE  16
typedef struct {
  unsigned char contents[KEYRINGSIZE];
  int head;
  int tail;
} KeyRing;

/* receive buffer management */
#define RECEIVEBUFFERSIZE  512
typedef struct _reveivebuffer {
  unsigned char contents[RECEIVEBUFFERSIZE];
  int head;
  int tail;
  int peek;
} ReceiveBuffer;

#ifndef DEFAULT_CELL_HEIGHT
#define DEFAULT_CELL_HEIGHT  8
#endif

/* command management */
#define MAX_DATA_LENGTH  22  /* CF635 spec says 0..22 */
#define MAX_COMMAND  35  /* CF635 spec says 0..35 */
//const char *cf633_pack_names[64];  // name from number
const char *cf633_pack_names[MAX_COMMAND+2];  // name from number
//#define RESEND_TIMEOUT 80000  // how long to wait b4 resending
#define RESEND_TIMEOUT 800000  // how long to wait b4 resending
/* in what units now ? how used differently than p->resend_timeout ? */
typedef struct {
  ubyte command;
  ubyte data_length;
  ubyte data[MAX_DATA_LENGTH+1];
  union {
    unsigned char as_bytes[2];
    word as_word;
  } crc;
} COMMAND_PACKET;

/* Constants for userdefchar_mode */
#define NUM_CCs         8 /* max. number of custom characters */

typedef struct cgram_cache {
        char cache[DEFAULT_CELL_HEIGHT];
        int clean;
} CGram;

typedef struct {
  unsigned char romid[8];    // ROM ID of sensor
  unsigned char cfromid[8];  // CF 'ID' of sensor
  char *name;
  unsigned char data[3];     // raw data
  int abstracto;             // temp * 16 , ds18 abstracto-units
#ifdef CLIENT_CHILD
  int dowidget;              //  = line #  
  char *line;
#endif // CLIENT_CHILD
} CF_DS18;

typedef struct {
  char *name;               // pointer to name str of s_val 
  int *abstvalptr;          // pointer to value
                            // temp * 16 , ds18 abstracto-units
  unsigned char number;     // tmp number for passing to update func
  void (*updates)(Driver *, unsigned char);
                            // pointer to function that updates val
} FAN_SENSOR; 

typedef struct {
  char * name;              // a name for syslog or clients
  char autorep;             // packet autoreport fan's RPM pulses y/n 
  char autorep_mask;        // packet autoreport current state y/n 
  char power;               // 0-100 initial on power
  char p_log;               // log current power setting to syslog   
  int divisor;              // commonly 2; depends on fan make / model
  char * d_label;           // usually RPM
  char failsafe;            // y/n   CF633.pdf p.29, command 25
  char glitch;              // y/n   CF633.pdf p.30, command 26
  int by_s_off;             // deg. C above which to power on fan
                            // temp * 16 , ds18 abstracto-units
  int by_s_100;             // deg. C to reach full power 
                            //  linear from by_s_off 
                            // temp * 16 , ds18 abstracto-units
  FAN_SENSOR *f_sens[2];    // 
  char * dt_label;          // name for computed delta t 
  int dt;                   // store computed delta t
                            // temp * 16 , ds18 abstracto-units
  unsigned char heat;       // 1 if this fan blows to increase a temperature
  int a_sec;                // 1-31  adjust this fan's power every 
                            //  a_sec seconds 
  unsigned char data[3];    // raw data
  float rpm_rpt;            //  sensed data
#ifdef CF635LEDS
  unsigned char led;        // 0=off; 1=power; 2=report; 3=blink random 
#endif
#ifdef FANS_USEC
  struct timeval tdid_fan;  // time of last compute this fan power 
#define FSEC 1175644326     // makes fan times easier to read
#endif // FANS_USEC
#ifdef CLIENT_CHILD
  int dowidget;             // 10's = screen #, 0..9 = line # 
  char * line;
  char * line2;
#endif // CLIENT_CHILD
} CF_FAN;

#ifdef DO_HDD
/* Changed to array of pointers 20080404
have not checked if every read / write to pointer members 
is prefaced by if (parent_struct !=NULL)
but is working atm. CHEKME */
typedef struct {
  char *hddname;
  int hd_abstracto;          // temp * 16 , ds18 abstracto-units
#ifdef CLIENT_CHILD
  char *hddline;             // LCD line 
  int hddlcd;
#endif // CLIENT_CHILD
} CF_HDD;
#endif // DO_HDD

#ifdef DO_LM
typedef struct {
  char *lmlabel;               // name / 'label' by sensors.conf 
  char *chipname;               // lm-sensors chip names probed
  char *lmfeatu;               // feature for each lmname probed
#ifdef SENSORS_API_VERSION 
  int  subfnum;                // subfeature number for libsensors3
  const struct sensors_feature *featn;// one found feature
#else
  const struct sensors_feature_data *featn;// one found feature
#endif
  int   lm_abstracto;          // values of last read // temp * 16 , ds18 abstracto-units
  const struct sensors_chip_name *chipn;   // one found chip
#ifdef CLIENT_CHILD
  char *lmline;                // LCD line for lm sensors
  int lmlcd;                   // line display order
#endif // CLIENT_CHILD
} CF_LM;
#endif // DO_LM

#ifdef DO_ULT
typedef struct {
  float val;            // values (%load) of last read
  char *name;          
#ifdef CLIENT_CHILD
  char *line;          // LCD line for tops
  int lcd;             // line display order
#endif // CLIENT_CHILD
} CF_ULT;

typedef struct
{
  char name[16];      // could make this a pointer & malloc
  int pid;
  /* these are used in debugging / verifying top data 
     and sort jiffpid */
  unsigned long int utime, stime;
  long int num_threads;
//unsigned long int utime, stime, cutime, cstime;
} procjiffy_type;
#endif // DO_ULT

typedef struct {
  struct _pthread_cleanup_buffer pt_buffer;
/*
/usr/include/bits/pthreadtypes.h:50:typedef unsigned long int pthread_t;
can this be set to zero if for instance fan_pth not started ?
so pthread_cancel does not segfault ?
*/
  pthread_t listen_pth;
  pthread_t fan_pth;
  pthread_t log_pth;
#ifdef LIVE_TIMING
  pthread_t timing_pth;
#endif // LIVE_TIMING
  char device[200];
  int fd;
  int model;
  char model_str[MAX_DATA_LENGTH+1];
  int newfirmware;
  int usb;
  char user[64]; // what user to run as ?
  struct passwd *user_pwd;
  uid_t userid;
  COMMAND_PACKET priority_packet[3]; // fan power, fan reporting, backlight
  unsigned char priority_packet_need; // 0|1 fan, 0|2 reporting, 0|4 backlight
//#ifdef DEBUG
  char *pkn[MAX_COMMAND+2];  // names of packets
//#endif
  /* dimensions */
  int width, height;
  int cellwidth, cellheight;

  /* framebuffer and buffer for old LCD contents */
  unsigned char *framebuf;
  unsigned char *backingstore;

  /* defineable characters */
  CGram cc[NUM_CCs];
  CGmode ccmode;

  int contrast;
  int brightness;
  int offbrightness;
#ifdef FIXMECF635LEDS
  unsigned int LEDstate;
#endif
  // globals from io.c
  COMMAND_PACKET  incoming_command;
  ReceiveBuffer receivebuffer;
  KeyRing keyring;

  COMMAND_PACKET sendq[UCHAR_MAX+1];  // packets to send
  unsigned char sending;              // waiting for ack on this
  unsigned char endq;                 // end of qued packets
  struct timeval tsent;               // last time of any send
  struct timeval t_ack;               // last time of any ack 
  unsigned char q_depth;              // current depth of que
  unsigned char resend_repeats;       // current packet resent how many times ?
  int resend_timeout;                 // how long to wait in ms before resending
  int q_resent;                       // on resend, what q # was resent?
#ifdef TIMING_REPORT
  COMMAND_PACKET  last_good_ack;      // suspect: sending identical packets
  COMMAND_PACKET  l_g_a[MAX_COMMAND]; // as repeat causes ack failure
  unsigned char q_recover;            // q depth return to 0 from high flag
  unsigned int deepies;               // q depth return to 0 from high count
  unsigned char maxq;                 // dbg: how deep did the que get ?
  struct timeval timing_mk;  
  long listen[3];   // min, max spent litening
  long listen_t;    // total listen loops 
  long msg_type_count[3];             // GOOD TRY GIVE UP counter
  long ack_d[MAX_COMMAND+2][4];     // minimum, maximum, average, timeout times
  unsigned long int ack_c[MAX_COMMAND+2][4]; // total, dropped by que paring, 
                                      // resent from timeout, extra acks
  unsigned long int ack_c_t;          // total sent-recd / non-sensor acked
  unsigned long int prevp;            // previous total 
  long ack_c_t_qd;  // sum of q depths @ each ack, 
                                      // for average que depth
  unsigned long int ack_rep[3];       // key, fan, temp reports rec'd
  unsigned long int q_purged;         // how many packents purged
#endif // TIMING_REPORT 
  CF_FAN fans[4];
  CF_DS18 **ds18;
  unsigned char tempcount;            // # of temp probes reqested
                                      // then # found, if different
  long log_sens;                      // seconds between log reports
  struct timeval tdid_tlog;           // time of last print logline
#ifdef FANS_USEC
  struct timeval tdid_power;          // time of last set fan power 
  struct timeval treq_power;          // time of last request set fan power 
#endif // FANS_USEC
  /* some things need not be in private data  ? 
   * data that log client writes to could be in separate 
   * structure or only in driver client funcs */
  unsigned char incoming_rep;         // currently updating temp or fans
  unsigned char sent_while_bad;       // was a packet sent while incoming_rep?
  unsigned char fan_rep_first;        // first fan in report stream
  unsigned char fan_rep_last;         // last fan in report stream
  char logfile[128];                  // log data to here, not syslog
  FILE *log_fd;                       // logfile file descriptor

  unsigned char log_uptime;           // log uptime with fans and temps?
  double uptime;
#ifdef CLIENT_CHILD
  int prettylcd;                      // let elapse change wildly & use resources
                                      // for fast lcd screen updates
  unsigned char bool_heartbeat;         /* client screens have heartbeat or not */
#endif // CLIENT_CHILD
#ifdef DO_ULT
  unsigned char cpus;                 // number of cpus
  unsigned long int max_ticks_per;    // sum of max cpufreq / cpu 
  unsigned char newj;                 // 0|1 is swapping new|previous jiffies[newj|1-newj] 
  int dt_to_clktk;
  struct timeval delta_t[2];          /* passed to get jiffes to compute elapse */
  int *countofnamedjiffs;          // if a topname is a wilcard, this can count
                                      // how many pids fit that wildcard
                                      // to aid in comparing oddtf values to top / ps & debug 
  LinkedList *jiffies[3];             // [0|1] current & previous reads, flipping
                                      // [2] the dt|timeslice difference between old and current
  unsigned long int **named_jiffs[2];
  unsigned char log_load;             // log load with fans and temps?
  unsigned char log_load_s;           // log system load with fans and temps?
  unsigned char log_load_u;           // log user load with fans and temps?
  load_type load;                     // most recent load data
  float load_t, load_s, load_u;       // total, system, user
  long elapse;                        // elapse since last get_jiffies
  int topnum;                         // how deep into ranked top cpu load
  struct timeval tdid_top;
  CF_ULT **top_num;
  int topname;                        // how many to top by name
  CF_ULT **top_name;
  procjiffy_type *pjt;
#endif // DO_ULT

#ifdef DO_LM
  int lmnum;                          // # lm-s's configured 
  char *lmconf;                       // lm-sensors configuration file
  FILE *lmfile;                       // the opened file;
  CF_LM **lm;
#endif // DO_LM

#ifdef DO_HDD
  int hdd_count;
  int hdd_rep_len;
  CF_HDD **hdd;
#endif // DO_HDD
} PrivateData;

#endif
