/*
 *  This is a multithreaded que based fan and temperature enabled
 *  LCDproc driver for CrystalFontz 631, 633, and 635 LCDs
 *
 *  (get yours from http://crystalfontz.com)
 *  Applicable Data Sheets
 *  http://www.crystalfontz.com/products/631/CFA-631_v1.0.pdf
 *  http://www.crystalfontz.com/products/633/CFA_633_0_6.PDF
 *  http://www.crystalfontz.com/products/635/CFA_635_1_0.pdf
 *  http://www.crystalfontz.com/products/SCAB ??
 *
 *  Copyright (C) 2002 David GLAUDE
 *  Portions Copyright (c) 2002, Joris Robijn
 *  Portions Copyright (C) 2005 Peter Marschall
 *  Portions Copyright (C) 2005 Nicolas Croiset <ncroiset@vdldiffusion.com>
 *  Portions Copyright (C) 2008 Stewart Putnam
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */

/*
 * Driver status
 * 04/04/2002: Working driver
 * 05/06/2002: Reading of return value
 * 02/09/2002: KeyPad handling and return string
 * 03/09/2002: New icon incorporated
 * 27/01/2003: Adapted for CFontz 631
 * 10/04/2004: added temp and fan functions 
 *   /07/2005: packet que, hw probe, LCD lines
 * 16/05/2005: Adapted for CFontz 635
 * 21/12/2005: data logfile with proc, lm-sensors
 * 04/02/2008: pthreads, as CF63oddtf, docbook
 * THINGS NOT DONE:
 * + No support for multiple instance (require private structure)
 * + No cache of custom char usage (like in MtxOrb)
 *
 * THINGS DONE:
 * + Stopping the live rpting (of temperature)
 * + Stopping the rpting of temp and fan (is it necessary after reboot)
 * + Use of library for hbar and vbar (good but library could be better)
 * + Support for keypad (Using a KeyRing)
 * + BigNum (for CF635 only: it is a 4-line display)
 * + Output support (LED control on a CF635 only)
 *
 * THINGS TO DO:
 * + Make the caching at least for heartbeat icon
 * + Create and use the library (for custom char handling)
 *
 * See separate file CF63oddtf.changelog
*/
/*
 * ===========================================================================
 * 633 Test code for linux.
 * Copyright 2002, David GLAUDE
 * Partial copyright
 * 2001 Crystalfontz America, Inc. 
 * brent@crystalfontz.com
 * Written by Brent A. Crosby 
 * www.crystalfontz.com
 * ===========================================================================
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

// checking gnome-cups-icon�!, gibberish source in csv log...
// is this a permanent solution ?
// #include <ctype.h>

#include <termios.h>
#include <time.h>
#include <dirent.h>
#include <limits.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <wait.h>
// appears uneeded, which manpage said to include ?
#include <pwd.h> // for addjusting prvalages of clones
#include <pthread.h>

#include <sys/types.h>
#include <sys/time.h>
#ifdef __linux__
/* to try gettid */
#define _GNU_SOURCE
//#include <sched.h> /* sched_getaffinity */
#include <sys/syscall.h>
#endif

// was working with mispelling here
// signal.h included elsewhere or not needed?
#ifdef TIMING_REPORT
#ifdef LIVE_TIMING
#include <signal.h>
#endif //  LIVE_TIMING
#endif //  TIMING_REPORT

/* http://www.thescripts.com/forum/thread215657.html */
/* on odd errors working with & printing floats */
/* #include <math.h>  didn't do it */

#include "lcd.h"
#include "CF63oddtf.h"
#include "report.h"
#include "lcd_lib.h"
#include "CFontz-charmap.h"
#include "timing.h"
#ifdef DO_LM
#include <sensors/sensors.h>
/* sensors/chips.h should not be included for newer libsensors version */
//#include <sensors/chips.h>
#endif
#include <sensors/error.h>
#include "../../clients/lcdproc/machine.h"
#include "../../shared/sockets.h"
#ifdef CLIENT_CHILD /* includes */
#include "../screen.h"
#include "../widget.h"
#include "../clients.h"
#endif

/*********************************************************
 * create any custom lines as globals here
 * and increase MAXLINES in oddtf dot h 
 * and create ptr below within #ifdef PROTOCOL _ CHEATING
 * ******************************************************/
#ifdef MY_CUSTOM_LINE
char *watts=NULL;
float watts_est=0;
#endif // MY_CUSTOM_LINE

#define CF633_KEY_UP      1
#define CF633_KEY_DOWN      2
#define CF633_KEY_LEFT      3
#define CF633_KEY_RIGHT      4
#define CF633_KEY_ENTER      5
#define CF633_KEY_ESCAPE    6
#define CF633_KEY_UP_RELEASE    7
#define CF633_KEY_DOWN_RELEASE    8
#define CF633_KEY_LEFT_RELEASE    9
#define CF633_KEY_RIGHT_RELEASE    10
#define CF633_KEY_ENTER_RELEASE    11
#define CF633_KEY_ESCAPE_RELEASE  12
#define CF631_KEY_UL_PRESS    13
#define CF631_KEY_UR_PRESS    14
#define CF631_KEY_LL_PRESS    15
#define CF631_KEY_LR_PRESS     16
#define CF631_KEY_UL_RELEASE    17
#define CF631_KEY_UR_RELEASE    18
#define CF631_KEY_LL_RELEASE    19
#define CF631_KEY_LR_RELEASE    20

#define CELLWIDTH  DEFAULT_CELL_WIDTH
#define CELLHEIGHT  DEFAULT_CELL_HEIGHT


/* LEDs dispatch */
#define CF635_NUM_LEDs  8

/* Vars for the server core */
MODULE_EXPORT char *api_version = API_VERSION;
MODULE_EXPORT int stay_in_foreground = 0;
MODULE_EXPORT int supports_multiple = 0;
MODULE_EXPORT char *symbol_prefix = "CF63oddtf_";

/* Internal functions */
/* static void CF63oddtf_linewrap (int on); */
/* static void CF63oddtf_autoscroll (int on);  */
static void CF63oddtf_hidecursor (Driver *drvthis);
static void CF63oddtf_reboot (Driver *drvthis);
static void CF63oddtf_init_vbar (Driver *drvthis);
static void CF63oddtf_init_hbar (Driver *drvthis);
static void CF63oddtf_no_autodisp(Driver *drvthis);
static void CF63oddtf_hardware_clear (Driver *drvthis);
static void set_fan_fsg (Driver* drvthis);
static void set_temp_rpt(Driver* drvthis, unsigned char out[4]);
static int packet_neq(COMMAND_PACKET *packet1, COMMAND_PACKET *packet2);
int ds_init (Driver* drvthis);
int fan_init (Driver* drvthis);
int fan_get_temp (Driver* drvthis, unsigned char fan, unsigned char type, char *target, char *buf);
void query_hw_ver (Driver *drvthis);

void send_packet(Driver *drvthis, COMMAND_PACKET *out);
void commit_packet(Driver* drvthis,  COMMAND_PACKET *packet);
long time_diff (Driver *drvthis, struct timeval past, struct timeval now);
int           check_for_packet(Driver *drvthis, unsigned char expected_length);
#ifdef DEBUG
void          show_packet (Driver* drvthis, int rep_level, COMMAND_PACKET *packet);
#endif

#ifdef TIMING_REPORT
void timing_dump(Driver *drvthis);
void clear_timing (Driver *drvthis);
#ifdef LIVE_TIMING
static void install_signal(Driver *drvthis);
//static void live_timing(int signal_num);
static void live_timing(int signal_num, siginfo_t *sigfo, void *v_ptr);
//static void live_timing_reset(int signal_num);
static void live_timing_reset(int signal_num, siginfo_t *sigfo, void *v_ptr);
static void timing_loop(Driver *drvthis);
#endif
#endif

void spawn_clones (Driver *drvthis, unsigned char bitmask);
void clones_drop_priv (Driver *drvthis, char *calling_clone, uid_t new_uid);

void csv_header (Driver *drvthis);
int fan_loop(Driver * drvthis);
static void csv_log (Driver *drvthis); 
static void purge_que (Driver *drvthis, const char *caller); 
static void manage_fans  (Driver *drvthis);
#ifdef CF635LEDS
void set_led(Driver *drvthis, unsigned char fan, unsigned char power);
void compute_led_by_rpm (Driver *drvthis, unsigned char fan);
void led_rand (Driver *drvthis, unsigned char fan);
#endif
void set_fan_rpt (Driver *drvthis);
void set_fan_pow (Driver *drvthis);
int compute_temp (Driver *drvthis, unsigned char temp);
float compute_fan_rpm (Driver *drvthis, unsigned char fan);

int listen_packets (Driver * drvthis);
static void sel_wait(Driver *drvthis);
static void sensors_packet(Driver *drvthis);

int csv_cli (Driver * drvthis);
#ifdef DO_ULT
int ult_init(Driver *drvthis);
void store_load(Driver *drvthis);
static int sort_jiffpid (void *a, void *b);
int machine_get_jiffies(Driver *drvthis, LinkedList *procs, struct timeval *taken);
void store_tops(Driver *drvthis);
void store_tops_init(Driver *drvthis);
static int sort_jiffies (void *a, void *b);
#endif // DO_ULT
#ifdef DO_LM
int lm_init(Driver *drvthis);
int buildlm(Driver *drvthis);
static void update_lm (Driver *drvthis, unsigned char lmn);
#endif //DO_LM
#ifdef DO_HDD
static void hdd_init(Driver *drvthis);
static void hddtemp_temps(Driver *drvthis, unsigned char number);
#endif // DO_HDD
static void client_loop(Driver * drvthis);

#ifdef MY_CUSTOM_LINE
static void update_wattsline (Driver *drvthis, unsigned char unused);
#endif // MY_CUSTOM_LINE

pthread_mutex_t quemutex=PTHREAD_MUTEX_INITIALIZER;
#ifdef CLIENT_CHILD /* globals & dependent func declarations */
int order[MAXLINES];  // .conf value for ordering created lines, user arranging lines into screens
char *client_text[MAXLINES]; // the temp or fan or other text stored and updated in the driver module
unsigned char t_or_f_num[MAXLINES]; // which element in the fan or temp or lm or ult list is this one 
#ifdef PROTOCOL_CHEATING
char * server_widget[MAXLINES]; // will be set to widget-> text once it is created
#endif // PROTOCOL_CHEATING


// globals used by cli child
int lines=0; // # of lines actually created
int cf_dsk=-1;
int lcd_wid = 0;
int lcd_hgt = 0;
int lcd_cellwid = 0;
int lcd_cellhgt = 0;
int csv_init(Driver * drvthis); 
// updateline is an array of pointers to functions, not a function itself
// pointer to function is set to line update function for that fan / temp / lm / ult
// each update function takes a (Driver * ,  number of the fan / temp / lm / ulti ) as args
static void (*updateline[MAXLINES])(Driver *, unsigned char);

static void set_up_templine (Driver *drvthis, unsigned char temp);
static void set_up_fanline (Driver *drvthis, unsigned char fan);
#ifdef DO_ULT
static void set_up_top_name_line (Driver *drvthis, unsigned char topn);
#endif // DO_ULT
#ifdef DO_LM
static void set_up_lmline (Driver *drvthis, unsigned char lmn);
#endif //DO_LM
#ifdef DO_HDD
static void set_up_hddline (Driver *drvthis, unsigned char lmn);
#endif // DO_HDD

int wait_success(Driver * drvthis, int options, char **other);
int cli_scrn(Driver * drvthis); 
int read_connect_string(Driver * drvthis);
int split( char * str, char delim, char * parts[], int maxparts ); 
#endif // CLIENT_CHILD

void          EmptyKeyRing(KeyRing *kr);
int           AddKeyToKeyRing(KeyRing *kr, unsigned char key);
unsigned char GetKeyFromKeyRing(KeyRing *kr);
void          EmptyReceiveBuffer(ReceiveBuffer *rb);
void          SyncReceiveBuffer(ReceiveBuffer *rb, int fd, unsigned int number);
int           BytesAvail(ReceiveBuffer *rb);
unsigned char GetByte(ReceiveBuffer *rb);
int           PeekBytesAvail(ReceiveBuffer *rb);
void          SyncPeekPointer(ReceiveBuffer *rb);
void          AcceptPeekedData(ReceiveBuffer *rb);
unsigned char PeekByte(ReceiveBuffer *rb);
int  get_crc(char *buf, int len, int seed);

/*
 * Opens com port and sets baud correctly...
 */
MODULE_EXPORT int
CF63oddtf_init (Driver *drvthis)
{
  struct termios portset;
  int speed = DEFAULT_SPEED;
  int tmp;

  PrivateData *p;
  report(RPT_INFO, "%s init(%p)", __FUNCTION__, drvthis );

  /* Allocate and store private data */
  p = (PrivateData *) calloc(1, sizeof(PrivateData));

  if (p == NULL)
    return -1;
  if (drvthis->store_private_ptr(drvthis, p))
    return -1;

  /* Initialize the PrivateData structure */
  p->cellwidth = DEFAULT_CELL_WIDTH;
  p->cellheight = DEFAULT_CELL_HEIGHT;
  p->ccmode = standard;
#ifdef FIXMECF635LEDS
  p->LEDstate = 0xFFFF;
#endif
  p->sending=0;
  p->q_depth=0;
  for (tmp=0; tmp < UCHAR_MAX; tmp ++){
    p->sendq[tmp+1].command=0;
  }
  p->sendq[0].command=CF633_Dummy_Unused;
  p->endq=1;
  p->resend_repeats=0;
  p->resend_timeout=RESEND_TIMEOUT;
  p->priority_packet[0].command=CF633_Set_Fan_Power;
  p->priority_packet[0].data_length=4;
  p->priority_packet[1].command=CF633_Set_Up_Fan_Reporting;
  p->priority_packet[1].data_length=1;
  p->priority_packet[2].command=CF633_Set_LCD_And_Keypad_Backlight;
  p->priority_packet[2].data_length=1;
  p->priority_packet_need=0;
/* FIXME test: what if I start listen loop here, before much else is init? 
  2012 05 24 seems to be working well this way 
*/
  /* init and spawn listen loop */
  p->fd=0;
  spawn_clones (drvthis, 1+2); 

//#ifdef DEBUG 
  p->pkn[ 0]=strdup("Ping_Command");
  p->pkn[ 1]=strdup("Get_Hw_And_Fw_Ver");
  p->pkn[ 2]=strdup("Write_User_Flash_Area");
  p->pkn[ 3]=strdup("Read_User_Flash_Area");
  p->pkn[ 4]=strdup("Store_Current_State_As_Boot_State");
  p->pkn[ 5]=strdup("Reboot");
  // shortest 6
  p->pkn[ 6]=strdup("Clear_LCD_Screen");
  p->pkn[ 7]=strdup("Set_LCD_Contents_Line_One");
  p->pkn[ 8]=strdup("Set_LCD_Contents_Line_Two");
  p->pkn[ 9]=strdup("Set_LCD_Special_Character_Data");
  p->pkn[10]=strdup("Read_Eight_Bytes_of_LCD_Memory");
  p->pkn[11]=strdup("Set_LCD_Cursor_Position");
  p->pkn[12]=strdup("Set_LCD_Cursor_Style");
  p->pkn[13]=strdup("Set_LCD_Contrast");
  p->pkn[14]=strdup("Set_LCD_And_Keypad_Backlight");
  // longest seen often 28
  p->pkn[15]=strdup("Query_Last_Fan_Pulse_Count");
  p->pkn[16]=strdup("Set_Up_Fan_Reporting");
  p->pkn[17]=strdup("Set_Fan_Power");
  // shortest seen often 13
  p->pkn[18]=strdup("Read_DOW_Device_Information");
  p->pkn[19]=strdup("Set_Up_Temperature_Reporting");
  p->pkn[20]=strdup("Arbitrary_DOW_Transaction");
  p->pkn[21]=strdup("Set_Up_Live_Display");
  p->pkn[22]=strdup("Send_Command_Directly_to_the_LCD_Controller");
  // longest 43
  p->pkn[23]=strdup("Configure_Key_Reporting");
  p->pkn[24]=strdup("Read_Keypad_Polled_Mode");
  p->pkn[25]=strdup("Set_Fan_Power_Failsafe");
  p->pkn[26]=strdup("Set_Fan_Tachometer_Glitch_Filter");
  p->pkn[27]=strdup("Query_Fan_Power_and_Failsafe_Mask");
  p->pkn[28]=strdup("Set_ATX_Power_Switch_Functionality");
  p->pkn[29]=strdup("Enable_Disable_and_Reset_the_Watchdog");
  p->pkn[30]=strdup("Read_Reporting_and_Status");
  p->pkn[31]=strdup("Send_Data_to_LCD");
  p->pkn[32]=strdup("Key_Legends");
  p->pkn[33]=strdup("Set_Baud_Rate");
  p->pkn[34]=strdup("Set_GPIO_Pin");
  p->pkn[35]=strdup("Read_GPIO_Pin");
  p->pkn[36]=strdup("Dummy_Unused");
//#endif
  EmptyKeyRing(&p->keyring);
  EmptyReceiveBuffer(&p->receivebuffer);

  /* Which device should be used */
  strncpy(p->device, drvthis->config_get_string (drvthis->name, "Device", 0, DEFAULT_DEVICE), sizeof(p->device));
  p->device[sizeof(p->device)-1] = '\0';
  report(RPT_INFO,"%s Device (in config) is '%s'", __FUNCTION__, p->device);

  /* Which speed ? CF633 support 19200 only, CF631 & CF635 USB use 115200. */
  tmp = drvthis->config_get_int (drvthis->name, "Speed", 0, DEFAULT_SPEED);
  report(RPT_INFO,"%s Speed (in config) is '%d'", __FUNCTION__, tmp);
  if (tmp == 19200) speed = B19200;
  else if (tmp == 115200) speed = B115200;
  else {
    report(RPT_WARNING, "%s Speed must be 19200 or 11500. Using default value %d.",
        DEFAULT_SPEED, __FUNCTION__);
    speed = DEFAULT_SPEED;
  }

  /* Am I USB or not? */
  p->usb = drvthis->config_get_bool(drvthis->name, "USB", 0, 0);
  if (p->usb)
    report(RPT_INFO, "%s USB is indicated (in config).",  __FUNCTION__);

  /* Set up io port correctly, and open it... */
  report(RPT_DEBUG, "%s Opening device: %s",  __FUNCTION__, p->device);
  p->fd = open(p->device, (p->usb) ? (O_RDWR | O_NOCTTY) : (O_RDWR | O_NOCTTY | O_NDELAY));
  int errno_sav=errno;
  if (p->fd == -1) {
    report(RPT_CRIT, "%s open (%s) failed (%s)",
        __FUNCTION__, p->device, strerror (errno_sav));
    return -1;
  }
  else {
    report(RPT_INFO, "%s opened %s, p->fd=%d.",  __FUNCTION__, 
      p->device, p->fd);
  }
  tmp=tcgetattr (p->fd, &portset);
  errno_sav=errno;
  if (tmp==-1) {
    report(RPT_CRIT, "%s tcgetattr (p->fd, &portset) failed (%s)",
         __FUNCTION__, strerror (errno_sav));
    return -1;
  }

  /* We use RAW mode */
  if (p->usb) {
    report(RPT_CRIT, "%s p->usb", __FUNCTION__);
    // The USB way
    portset.c_iflag &= ~( IGNBRK | BRKINT | PARMRK | ISTRIP
          | INLCR | IGNCR | ICRNL | IXON );
    portset.c_oflag &= ~OPOST;
    portset.c_lflag &= ~( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
    portset.c_cflag &= ~( CSIZE | PARENB | CRTSCTS );
    portset.c_cflag |= CS8 | CREAD | CLOCAL ;
    portset.c_cc[VMIN] = 0;
    portset.c_cc[VTIME] = 0;
  } else {
#ifdef HAVE_CFMAKERAW
    /* The easy way */
    cfmakeraw(&portset);
#else
    /* The hard way */
    portset.c_iflag &= ~( IGNBRK | BRKINT | PARMRK | ISTRIP
                | INLCR | IGNCR | ICRNL | IXON );
    portset.c_oflag &= ~OPOST;
    portset.c_lflag &= ~( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
    portset.c_cflag &= ~( CSIZE | PARENB | CRTSCTS );
    portset.c_cflag |= CS8 | CREAD | CLOCAL ;
#endif
  }

  /* Set port speed */
  tmp=cfsetospeed (&portset, speed);
  errno_sav=errno;
  if (tmp==-1) {
    report(RPT_CRIT, "%s cfsetospeed (&portset, speed) failed (%s)",
         __FUNCTION__, strerror (errno_sav));
    return -1;
  }
  tmp=cfsetispeed (&portset, B0);
  errno_sav=errno;
  if (tmp==-1) {
    report(RPT_CRIT, "%s cfsetispeed (&portset, B0) failed (%s)",
         __FUNCTION__, strerror (errno_sav));
    return -1;
  }

  /* Do it... */
  tmp=tcsetattr (p->fd, TCSANOW, &portset);
  errno_sav=errno;
  if (tmp==-1) {
    report(RPT_CRIT, "%s tcgetattr (p->fd, TCSANOW, &portset) failed (%s)",
         __FUNCTION__, strerror (errno_sav));
    return -1;
  }
  /* init and spawn listen loop */
//  spawn_clones (drvthis, 1+2); 
  
  /* with a viable query_hw_ver what can be discarded ? */
  query_hw_ver (drvthis);

  /* Reboot display? takes 2 seconds! get it done first */
  if (drvthis->config_get_bool(drvthis->name, "Reboot", 0, 0)) {
    CF63oddtf_reboot (drvthis );
  }
  
  /* make sure the frame buffer is there... */
  p->framebuf = (unsigned char *) malloc(p->width * p->height);
  if (p->framebuf == NULL) {
    report(RPT_CRIT, "%s unable to create framebuffer.",  __FUNCTION__);
    return -1;
  }
  memset(p->framebuf, ' ', p->width * p->height);

  /* make sure the framebuffer backing store is there... */
  p->backingstore = (unsigned char *) malloc(p->width * p->height);
  if (p->backingstore == NULL) {
    report(RPT_CRIT, "%s Unable to create framebuffer backing store.",  __FUNCTION__);
    return -1;
  }
  memset(p->backingstore, ' ', p->width * p->height);

  /* Which contrast */
  tmp = drvthis->config_get_int (drvthis->name, "Contrast", 0, DEFAULT_CONTRAST);
  report(RPT_INFO, "%s Contrast (in config) is '%d'", __FUNCTION__, tmp);
  if ((tmp < 0) || (tmp > 1000)) {
    report(RPT_WARNING, "%s Contrast must be between 0 and 1000. Using default value.",  __FUNCTION__);
    tmp = DEFAULT_CONTRAST;
  }
  p->contrast = tmp;

  /* Which backlight brightness */
  tmp = drvthis->config_get_int (drvthis->name, "Brightness", 0, DEFAULT_BRIGHTNESS);
  report(RPT_INFO,"CF_init: Brightness (in config) is '%d'", tmp);
  if ((tmp < 0) || (tmp > 1000)) {
    report(RPT_WARNING, "%s Brightness must be between 0 and 1000. Using default value.",  __FUNCTION__);
    tmp = DEFAULT_BRIGHTNESS;
  }
  p->brightness = tmp;

  /* Which backlight-off "brightness" */
  tmp = drvthis->config_get_int (drvthis->name, "OffBrightness", 0, DEFAULT_OFFBRIGHTNESS);
  report(RPT_INFO,"%s OffBrightness (in config) is '%d'", __FUNCTION__, tmp);
  if ((tmp < 0) || (tmp > 1000)) {
    report(RPT_WARNING, "%s OffBrightness must be between 0 and 1000. Using default value.",  __FUNCTION__);
    tmp = DEFAULT_OFFBRIGHTNESS;
  }
  p->offbrightness = tmp;
  
  /* Do temperatures or no from config*/
  p->tempcount = drvthis->config_get_int(drvthis->name, "Temperatures", 0, DEFAULT_TEMP_REPORT);

  /* contrast hints */
  if ( (p->model==635) && ((p->contrast < 300) ||  (p->contrast > 700)) ){
    // 65-126 hw native by 635.pdf =~ 254 -500 ; 300 - 700 by experience
    report(RPT_INFO, "%s Contrast %d outside suggested normal %d range 300 - 700.", 
         __FUNCTION__, p->model, p->contrast);
  }
  else if ( ((p->model==631)||(p->model==633) ) && ((p->contrast < 50) ||  (p->contrast > 500)) ){
      // 0-29 hw native by 633.pdf =~ 0-600 on 1-1000 config scale 
    report(RPT_INFO, "%s Contrast %d outside suggested normal %d range 50 - 500.", 
         __FUNCTION__, p->model, p->contrast);
  }

  /*  Temp and Fan syslog rpting or no */
  p->log_sens = (long)drvthis->config_get_int(drvthis->name, "LogInterval", 0, DEFAULT_SYSLOG_SENSORS );
  if (p->log_sens > 1800) {
    report(RPT_NOTICE, "%s LogInterval = %d shortened to 30 minutes (1800 seconds.)", 
         __FUNCTION__, p->log_sens);
    p->log_sens=1800;
  }

  /* and now send other init packets */
  
  /* Set display-specific stuff.. */
  CF63oddtf_hidecursor (drvthis);

  CF63oddtf_set_contrast (drvthis, p->contrast);

  CF63oddtf_hardware_clear (drvthis);

#ifdef CF635LEDS
  /* turn LEDs off on a CF635 */
  for (tmp=0; tmp < 4 ; tmp ++) 
    set_led(drvthis, tmp, 0);
  /* CF63oddtf_output(drvthis, 0); */
#endif
  
  if (p->tempcount >0)
    if (!ds_init(drvthis)) return -1;

#ifdef DO_LM
  lm_init(drvthis);
#endif  
#ifdef DO_ULT // does DO_LM depend on this too?
  ult_init(drvthis);
#endif // DO_ULT
#ifdef DO_HDD  
  /* Log all hddtemps  or no ? */
  tmp = drvthis->config_get_bool(drvthis->name, "loghddtemp", 0, 0);
  if (tmp !=0){
    hdd_init(drvthis);
  }
  else p->hdd_count=0;
#endif
  if (!fan_init(drvthis)) return -1;
  
  /* what else to set on completion of init? */
  gettimeofday(&p->tdid_tlog,NULL);
  p->tdid_tlog.tv_sec-=p->log_sens;

  //spawn_clones (drvthis, 4); 
  //spawn_clones (drvthis, 8); 
  //spawn_clones (drvthis, 16); 
  spawn_clones (drvthis, 28); 
/*
  report(RPT_CRIT, "%s EXPLICIT EXIT ON DEBUGGING.",  __FUNCTION__);
  exit(0);
*/
return 0;
}


/*
 * Clean-up
 */
MODULE_EXPORT void
CF63oddtf_close (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  int cmd;
  pthread_mutex_destroy(&quemutex);
#ifdef __linux__
  debug(RPT_INFO, "pthread_self % 15s = %d tid = %d", 
    __FUNCTION__, pthread_self(), syscall(SYS_gettid));
#endif
#ifdef LIVE_TIMING
  // get a report
  // under 100% load not getting up 
  // nanosleep below helps 
  timing_dump(drvthis);
#endif // TIMING_REPORT
  // terminate log child first, it writes most 
  if (pthread_cancel(p->log_pth))
    report (RPT_CRIT, "p->log_pth cancel error");
    else report (RPT_DEBUG, "p->log_pth cancel ok");
  if (pthread_cancel(p->listen_pth))
    report (RPT_CRIT, "listen_pth cancel error");
    else report (RPT_DEBUG, "listen_pth cancel ok");
/*
2009 05 17
SIGSEV, on pthread cancel fan ntread
fan thread is not running
==18969== Thread 1:
==18969== Invalid read of size 4
==18969==    at 0x5BB3C51: pthread_cancel (in /lib/libpthread-2.7.so)
==18969==    by 0x5788AFC: CF63oddtf_close (CF63oddtf.c:645)
==18969==    by 0x4127DD: driver_unload (driver.c:189)
==18969==    by 0x41198A: drivers_unload_all (drivers.c:148)
==18969==    by 0x408270: exit_program (main.c:921)
==18969==    by 0x505BF5F: (within /lib/libc-2.7.so)
==18969==    by 0x50C7FF0: nanosleep (in /lib/libc-2.7.so)
==18969==    by 0x50F34C3: usleep (in /lib/libc-2.7.so)
==18969==    by 0x408A17: main (main.c:871)
==18969==  Address 0x90 is not stack'd, malloc'd or (recently) free'd
==18969==
==18969== Process terminating with default action of signal 11 (SIGSEGV)
==18969==  Access not within mapped region at address 0x90
==18969==    at 0x5BB3C51: pthread_cancel (in /lib/libpthread-2.7.so)
==18969==    by 0x5788AFC: CF63oddtf_close (CF63oddtf.c:645)
==18969==    by 0x4127DD: driver_unload (driver.c:189)
==18969==    by 0x41198A: drivers_unload_all (drivers.c:148)
==18969==    by 0x408270: exit_program (main.c:921)
*/
  if ((p->fan_pth!=0) && (pthread_cancel(p->fan_pth)))
    report (RPT_CRIT, "fan_pth cancel error");
    else report (RPT_DEBUG, "fan_pth cancel ok");
#ifdef TIMING_REPORT
#ifdef LIVE_TIMING
  if (pthread_cancel(p->timing_pth))
    // if this is called by SIGUSR, can it cancel iself ?
    report (RPT_CRIT, "timing_pth cancel error");
    else report (RPT_DEBUG, "timing_pth cancel ok");
#endif // LIVE_TIMING
#endif // TIMING_REPORT
  // try to display the Thanks / Goodby screen ...
  CF63oddtf_flush (drvthis); 
  struct timespec nod = {0,  015000000}; nanosleep(&nod, NULL); // let goodby screen get sent
  //   110000 = get 1st line only usually
  p->log_sens=0;  //  stop all logging by that thread
#ifdef DEBUG
  for (cmd=0; cmd < MAX_COMMAND+2; cmd ++)
    free (p->pkn[cmd]);
#endif
  for (cmd=0; cmd < 4; cmd ++){
    if (p->fans[cmd].name) free (p->fans[cmd].name);
    if (p->fans[cmd].d_label) free (p->fans[cmd].d_label);
    if (p->fans[cmd].dt_label) free (p->fans[cmd].dt_label);
    if (p->fans[cmd].f_sens[0]) free (p->fans[cmd].f_sens[0]);
    if (p->fans[cmd].f_sens[1]) free (p->fans[cmd].f_sens[1]);
#ifdef CLIENT_CHILD /* close & free fans */
    if (p->fans[cmd].line) free (p->fans[cmd].line);
    if (p->fans[cmd].line2) free (p->fans[cmd].line2);
#endif
    /* p->fans[cmd] was not alloced */
  }
  for (cmd=0; cmd < p->tempcount ; cmd ++){
    if (p->ds18[cmd]->name!=NULL) free (p->ds18[cmd]->name);
#ifdef CLIENT_CHILD /* close & free ds */
    if (p->ds18[cmd]->line!=NULL) free (p->ds18[cmd]->line);
#endif
    if (p->ds18[cmd]) free (p->ds18[cmd]);
  }
  if (p->ds18!=NULL) free (p->ds18);
#ifdef DO_ULT
  // stop any other sytem info reads too
  p->log_uptime = 0;
  p->log_load = 0;
  p->log_load_s = 0;
  p->log_load_u = 0;
  for (cmd=0; cmd < p->topnum; cmd ++){
    if (p->top_num[cmd]->name) free (p->top_num[cmd]->name);
    if (p->top_num[cmd]) free (p->top_num[cmd]);
  }
  if (p->top_num) free (p->top_num);
  for (cmd=0; cmd < p->topname; cmd ++){
    if (p->top_name[cmd]->name) free (p->top_name[cmd]->name);
#ifdef CLIENT_CHILD /* close & free top_name */
    if (p->top_name[cmd]->line) free (p->top_name[cmd]->line);
#endif
    if (p->top_name[cmd]) free (p->top_name[cmd]);
    if ((p->named_jiffs[0]) && (p->named_jiffs[0][cmd])) free (p->named_jiffs[0][cmd]);
    if ((p->named_jiffs[1]) && (p->named_jiffs[1][cmd])) free (p->named_jiffs[1][cmd]);
  }
  if (p->countofnamedjiffs) free(p->countofnamedjiffs); 
  if (p->named_jiffs[0]) free (p->named_jiffs[0]);
  if (p->named_jiffs[1]) free (p->named_jiffs[1]);
  if (p->top_name) free (p->top_name);
  /* at any moment, 2 or 3 might be in use */
  for (cmd=0; cmd < 3; cmd ++){
    if (p->jiffies[cmd]!=NULL){ 
      LL_Rewind(p->jiffies[cmd]);
      do {
        procjiffy_type *tofree = (procjiffy_type *) LL_Get(p->jiffies[cmd]);
        if (tofree != NULL) { free(tofree); }
      } while (LL_Next(p->jiffies[cmd]) == 0);
      LL_Destroy(p->jiffies[cmd]);
      p->jiffies[cmd]=NULL;
    }
  } /* end each of 3 possible LL */
  if (p->pjt !=NULL) free (p->pjt);
#endif // DO_ULT
#ifdef DO_LM
    for (cmd=0; cmd < p->lmnum; cmd ++){
      if (p->lm[cmd]){
        if (p->lm[cmd]->chipname) free (p->lm[cmd]->chipname);
        if (p->lm[cmd]->lmfeatu) free (p->lm[cmd]->lmfeatu);
        if (p->lm[cmd]->lmlabel) free (p->lm[cmd]->lmlabel);
          /* sensors_get_label newly created, free it yourself */
#ifdef CLIENT_CHILD /* close & free lm */
        if (p->lm[cmd]->lmline) free (p->lm[cmd]->lmline);
#endif
        free (p->lm[cmd]);
        report(RPT_CRIT,
         "%s free p->lm[%d].", __FUNCTION__, cmd );
      }
    }
    if (p->lm) free (p->lm);
    if (p->lmconf)  free (p->lmconf);
    if (p->lmfile)  {
      sensors_cleanup();
      fclose (p->lmfile);
    }
#endif // DO_LM
#ifdef DO_HDD
    for (cmd=0; cmd < p->hdd_count; cmd ++){
      if (p->hdd[cmd]->hddname) free (p->hdd[cmd]->hddname);
#ifdef CLIENT_CHILD /* close & free hdd */
      if (p->hdd[cmd]->hddline) free (p->hdd[cmd]->hddline);
#endif
      if (p->hdd[cmd]) free (p->hdd[cmd]);
    }
    if (p->hdd) free (p->hdd);
    report(RPT_CRIT, "%s free p->hdd.", __FUNCTION__ );
#endif // DO_HDD
#ifdef MY_CUSTOM_LINE
    if (watts!=NULL) free (watts);
#endif
  //if (p->user_pwd!=NULL) free (p->user_pwd);
  if (p->log_fd)  fclose (p->log_fd);
  close(p->fd);
  report(RPT_CRIT, "%s free p->fd. p->framebuf, p->backingstore next",
     __FUNCTION__ );
  if (p->framebuf)
    free(p->framebuf);
  p->framebuf = NULL;
  if (p->backingstore)
    free(p->backingstore);
  p->backingstore = NULL;
  drvthis->store_private_ptr(drvthis, NULL);
  free(p);
} // end CF63oddtf_close


/*
 * Returns the display width in characters
 */
MODULE_EXPORT int
CF63oddtf_width (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->width;
}


/*
 * Returns the display height in characters
 */
MODULE_EXPORT int
CF63oddtf_height (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->height;
}


/*
 * Returns the width of a character in pixels
 */
MODULE_EXPORT int
CF63oddtf_cellwidth (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->cellwidth;
}


/*
 * Returns the height of a character in pixels
 */
MODULE_EXPORT int
CF63oddtf_cellheight (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->cellheight;
}

/*
 * Flushes all output to the lcd...
 */
MODULE_EXPORT void
CF63oddtf_flush (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  int i,j;
  COMMAND_PACKET flush;

  if (p->model == 633) {
  /*
   * For CF633 we don't use delta update yet.
   * The protocol only permit update of full or partial line starting from pos 0.
   */
  unsigned char *xp = p->framebuf;
  unsigned char *xq = p->backingstore;
  flush.command=CF633_Set_LCD_Contents_Line_One;
  flush.data_length=p->width;

    for (i = 0; i < p->width; i++) {
      if (*xp++ != *xq++) {
        memcpy(flush.data, p->framebuf, p->width);
        send_packet(drvthis, &flush);
        memcpy(p->backingstore, p->framebuf, p->width);
        break;
      }
    }

    xp = p->framebuf + p->width;
    xq = p->backingstore + p->width;
    flush.command=CF633_Set_LCD_Contents_Line_Two;

    for (i = 0; i < p->width; i++) {
      if (*xp++ != *xq++) {
        memcpy(flush.data, p->framebuf + p->width, p->width);
        send_packet(drvthis, &flush);
        memcpy(p->backingstore + p->width, p->framebuf + p->width, p->width);
        break;
      }
    }
  }
  else { /* (p->model != 633) */
  /*
   * CF631 / CF635 protocol is more flexible and we can do real delta update.
   */

    for (i = 0; i < p->height; i++) {
      // set  pointers to start of the line in frame buffer & backing store
      unsigned char *sp = p->framebuf + (i * p->width);
      unsigned char *sq = p->backingstore + (i * p->width);

      report(RPT_DEBUG, "%s Framebuf: '%.*s'",  __FUNCTION__, p->width, sp);
      report(RPT_DEBUG, "%s Backingstore: '%.*s'",  __FUNCTION__, p->width, sq);
      /* Strategy:
       * - not more than one update command per line
       * - leave out leading and trailing parts that are identical
       */
      
      // set  pointers to end of the line in frame buffer & backing store
      unsigned char *ep = sp + (p->width - 1);
      unsigned char *eq = sq + (p->width - 1);
      int length = 0;

      // skip over leading identical portions of the line
      for (j = 0; (sp <= ep) && (*sp == *sq); sp++, sq++, j++) ;
    
      // skip over trailing identical portions of the line
      for (length = p->width - j; (length > 0) && (*ep == *eq); ep--, eq--, length--) ;

      /* there are differences, ... */
      if (length > 0) {
        /* ... send then to the LCD */
        flush.data[0] = j;  // column
        flush.data[1] = i;  // line
        report(RPT_DEBUG, "%s: l=%d c=%d count=%d string='%.*s'",
           __FUNCTION__, flush.data[0], flush.data[1], length, length, sp);
        flush.command=CF633_Send_Data_to_LCD;
        flush.data_length=length + 2;
        memcpy(&flush.data[2], sp, length);
        send_packet(drvthis, &flush);
      }      
    }  // i < p->height
    memcpy(p->backingstore, p->framebuf, p->width * p->height);
  }
}

/*
 * Return one char from the KeyRing
 */
MODULE_EXPORT char *
CF63oddtf_get_key (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  unsigned char key;

  key = GetKeyFromKeyRing(&p->keyring);

  switch (key) {
    case CF633_KEY_LEFT:
      return "Left";
      break;
    case CF633_KEY_UP:
      return "Up";
      break;
    case CF633_KEY_DOWN:
      return "Down";
      break;
    case CF633_KEY_RIGHT:
      return "Right";
      break;
    case CF633_KEY_ENTER:
      return "Enter";
      break;
    case CF633_KEY_ESCAPE:
      return "Escape";
      break;
    case CF631_KEY_UL_PRESS:
      return "Up";
      break;
    case CF631_KEY_UR_PRESS:
      return "Enter";
      break;
    case CF631_KEY_LL_PRESS:
      return "Down";
      break;
    case CF631_KEY_LR_PRESS:
      return "Escape";
      break;
    case CF633_KEY_UP_RELEASE:
    case CF633_KEY_DOWN_RELEASE:
    case CF633_KEY_LEFT_RELEASE:
    case CF633_KEY_RIGHT_RELEASE:
    case CF633_KEY_ENTER_RELEASE:
    case CF633_KEY_ESCAPE_RELEASE:
    case CF631_KEY_UL_RELEASE:
    case CF631_KEY_UR_RELEASE:
    case CF631_KEY_LL_RELEASE:
    case CF631_KEY_LR_RELEASE:
      // report(RPT_INFO, "%s Returning key release 0x%2x",  __FUNCTION__, key);
      return NULL;
      break;
    default:
      if (key != '\0')
        report(RPT_INFO, "%s Untreated unknown key 0x%2x",  __FUNCTION__, key);
      return NULL;
      break;
  }
  return NULL;
}


/*
 * Prints a character on the lcd display, at position (x,y).
 * The upper-left is (1,1), and the lower right should be (16,2).
 */
MODULE_EXPORT void
CF63oddtf_chr (Driver *drvthis, int x, int y, char c)
{
  PrivateData *p = drvthis->private_data;

  y--;
  x--;

  p->framebuf[(y * p->width) + x] = (p->model == 633)
                              ? c
            : CFontz_charmap[(unsigned) c];
}


/*
 * Prints a character on the lcd display, at position (x,y).
 * The upper-left is (1,1), and the lower right should be (16,2).
 */
static void
CF63oddtf_raw_chr (Driver *drvthis, int x, int y, unsigned char c)
{
  PrivateData *p = drvthis->private_data;

  y--;
  x--;

  p->framebuf[(y * p->width) + x] = c;
}


/*
 * Returns current contrast (in promille)
 * This is only the locally stored contrast, the contrast value
 * cannot be retrieved from the LCD.
 */
MODULE_EXPORT int
CF63oddtf_get_contrast (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  return p->contrast;
}


/*
 *  Changes screen contrast (in promille)
 */
MODULE_EXPORT void
CF63oddtf_set_contrast (Driver *drvthis, int promille)
{
  PrivateData *p = drvthis->private_data;
  COMMAND_PACKET contr;

  /* Check it */
  if (promille < 0 || promille  > 1000)
    return;
  /* store the software value since there is not get */
  p->contrast = promille;
    
  /* map range [0, 1000] to a range that the hardware understands */
  /* on CF633: [0, 50], on CF631 & CF635: [0, 255] */
  contr.data[0] = (p->model == 633)
          ? (ubyte)(p->contrast / 20)
          : (ubyte)((p->contrast * 255) / 1000);
  
  contr.command=CF633_Set_LCD_Contrast;
  contr.data_length=1;
  
  report(RPT_INFO, "%s Contrast set to %d. (CF hardware=%d)",
       __FUNCTION__, p->contrast, contr.data[0]);
   send_packet (drvthis, &contr);
}


/*
 * Retrieves brightness (in promille)
 */
MODULE_EXPORT int
CF63oddtf_get_brightness(Driver *drvthis, int state)
{
  PrivateData *p = drvthis->private_data;

  return (state == BACKLIGHT_ON) ? p->brightness : p->offbrightness;
}


/*
 * Sets on/off brightness (in promille)
 */
MODULE_EXPORT void
CF63oddtf_set_brightness(Driver *drvthis, int state, int promille)
{
  PrivateData *p = drvthis->private_data;

  /* Check it */
  if (promille < 0 || promille > 1000)
    return;

  /* store the software value since there is not get */
  if (state == BACKLIGHT_ON) {
    p->brightness = promille;
    CF63oddtf_backlight(drvthis, BACKLIGHT_ON);
  }
  else {
    p->offbrightness = promille;
    CF63oddtf_backlight(drvthis, BACKLIGHT_OFF);
  }
}


/*
 * Sets the backlight on or off.
 * The hardware support any value between 0 and 100.
 */
MODULE_EXPORT void
CF63oddtf_backlight (Driver *drvthis, int on)
{
  PrivateData *p = drvthis->private_data;
  int hardware_value = (on == BACKLIGHT_ON)
           ? p->brightness
           : p->offbrightness;
  
  static unsigned int redundant_repeats_from_upstream=0;
  
  /* map range [0, 1000] -> [0, 100] that the hardware understands */
  hardware_value /= 10;
  
  if (p->priority_packet[2].data[0] != hardware_value) {
    report(RPT_DEBUG, "%s CF633_Set_LCD_And_Keypad_Backlight %d is new after %d repeats of %d.",
         __FUNCTION__, hardware_value, redundant_repeats_from_upstream, 
        p->priority_packet[2].data[0]);
    p->priority_packet[2].data[0]=hardware_value;
    if (p->q_depth==0) {
      send_packet(drvthis, &p->priority_packet[2]);
    }else {
      p->priority_packet_need|=4; 
    }
    redundant_repeats_from_upstream=0;
  }
  else{
    redundant_repeats_from_upstream++;
    if (redundant_repeats_from_upstream==65535){
      redundant_repeats_from_upstream=0;
      report(RPT_DEBUG, "%s CF633_Set_LCD_And_Keypad_Backlight 65536 repeats of %d.",
         __FUNCTION__, p->priority_packet[2].data[0]);
    }
  }
}


/*
 * Get rid of the blinking cursor
 */
static void
CF63oddtf_hidecursor (Driver *drvthis)
{
  COMMAND_PACKET cursor;
  cursor.command=CF633_Set_LCD_Cursor_Style;
  cursor.data_length=1;
  cursor.data[0]=0;
  send_packet(drvthis, &cursor);
}

/*
 * Stop autodisplays
 */
static void
CF63oddtf_no_autodisp(Driver *drvthis)
{
  // 12 19 05 this is getting error responses on 635 w / scab
  // presumably will with 631 as well
  // scab.pdf says deprecated ( & no 'see new command' )
  //  do we know model already ? 
  PrivateData *p = drvthis->private_data;
  if (p->model!=633) return;
  COMMAND_PACKET packet;
  packet.command=CF633_Set_Up_Live_Fan_or_Temperature_Display;
  packet.data_length=2;
  packet.data[0]=0;
  packet.data[1]=0;

  for (packet.data[0] = 0; packet.data[0] < 8; packet.data[0]++){
    send_packet(drvthis, &packet);
  }
}


/*
 * Reset the display bios
 */
static void 
CF63oddtf_reboot (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  COMMAND_PACKET reboo;
  struct timespec nod = {0,  015000000}; 

  report(RPT_INFO, "%s module reboot requested.",  __FUNCTION__);
  reboo.command=CF633_Reboot;
  reboo.data_length=3;
  reboo.data[0] = 8;
  reboo.data[1] = 18;
  reboo.data[2] = 99;
  send_packet(drvthis, &reboo);
  while (p->q_depth !=0) nanosleep(&nod, NULL); // _reboot 
  
  // next command after reboot takes longer ?
  reboo.command=CF633_Ping_Command;
  reboo.data_length=0;
  send_packet(drvthis, &reboo);
  while (p->q_depth !=0) nanosleep(&nod, NULL); // _reboot
}


/*
 * Sets up for vertical bars.
 */
static void
CF63oddtf_init_vbar (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  char a[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
  };
  char b[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char c[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char d[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char e[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char f[] = {
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };
  char g[] = {
    0, 0, 0, 0, 0, 0,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
    0, 1, 1, 1, 1, 1,
  };

  if (p->ccmode != vbar) {
    if (p->ccmode != standard) {
      /* Not supported(yet) */
      report(RPT_WARNING, "%s Cannot combine two modes using user defined characters",
           __FUNCTION__);
      return;
    }
    p->ccmode = vbar;

    CF63oddtf_set_char (drvthis, 1, a);
    CF63oddtf_set_char (drvthis, 2, b);
    CF63oddtf_set_char (drvthis, 3, c);
    CF63oddtf_set_char (drvthis, 4, d);
    CF63oddtf_set_char (drvthis, 5, e);
    CF63oddtf_set_char (drvthis, 6, f);
    CF63oddtf_set_char (drvthis, 7, g);
  }
}

/*
 * Inits horizontal bars...
 */
static void
CF63oddtf_init_hbar (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;
  char a[] = {
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
    1, 0, 0, 0, 0, 0,
  };
  char b[] = {
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
    1, 1, 0, 0, 0, 0,
  };
  char c[] = {
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
  };
  char d[] = {
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
    1, 1, 1, 1, 0, 0,
  };
  char e[] = {
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
    1, 1, 1, 1, 1, 0,
  };
  char f[] = {
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
  };

  if (p->ccmode != hbar) {
    if (p->ccmode != standard) {
      /* Not supported(yet) */
      report(RPT_WARNING, "%s CF63oddtf_init_hbar: Cannot combine two modes using user defined characters",
           __FUNCTION__);
      return;
    }
    p->ccmode = hbar;

    CF63oddtf_set_char (drvthis, 1, a);
    CF63oddtf_set_char (drvthis, 2, b);
    CF63oddtf_set_char (drvthis, 3, c);
    CF63oddtf_set_char (drvthis, 4, d);
    CF63oddtf_set_char (drvthis, 5, e);
    CF63oddtf_set_char (drvthis, 6, f);
  }
}


/*
 * Draws a vertical bar...
 */
MODULE_EXPORT void
CF63oddtf_vbar (Driver *drvthis, int x, int y, int len, int promille, int options)
{
/* x and y are the start position of the bar.
 * The bar by default grows in the 'up' direction
 * (other direction not yet implemented).
 * len is the number of characters that the bar is long at 100%
 * promille is the number of promilles (0..1000) that the bar should be filled.
 */
  PrivateData *p = drvthis->private_data;

  CF63oddtf_init_vbar(drvthis);
  lib_vbar_static(drvthis, x, y, len, promille, options, p->cellheight, 0);
}


/*
 * Draws a horizontal bar to the right.
 */
MODULE_EXPORT void
CF63oddtf_hbar (Driver *drvthis, int x, int y, int len, int promille, int options)
{
/* x and y are the start position of the bar.
 * The bar by default grows in the 'right' direction
 * (other direction not yet implemented).
 * len is the number of characters that the bar is long at 100%
 * promille is the number of promilles (0..1000) that the bar should be filled.
 */
  PrivateData *p = drvthis->private_data;

  CF63oddtf_init_hbar(drvthis);
  lib_hbar_static(drvthis, x, y, len, promille, options, p->cellwidth, 0);
}


/*
 * Inits big numbers
 */
void
CF63oddtf_init_num(Driver *drvthis)
{
PrivateData *p = drvthis->private_data;

char bignum_ccs[8][CELLWIDTH*CELLHEIGHT] = {
  { 1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0 },

  { 0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0 },

  { 1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0 },

  { 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0 },

  { 1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1 },

  { 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1,
    1, 1, 1, 1, 1, 1 },

  { 1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0,
    1, 1, 1, 0, 0, 0 },

  { 0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 1, 1, 1,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0 }
};

  if (p->ccmode != bignum) {
    int i;

    if (p->ccmode != standard) {
      /* Not supported (yet) */
      report(RPT_WARNING, "%s Cannot combine two modes using user defined characters",
           __FUNCTION__);
      return;
    }

    p->ccmode = bignum;

    for (i = 0; i < NUM_CCs; i++)
      CF63oddtf_set_char(drvthis, i, bignum_ccs[i]);
  }
}


/*
 * Writes a big number.
 * Only works on 4-line displays
 */
MODULE_EXPORT void
CF63oddtf_num(Driver *drvthis, int x, int num)
{
PrivateData *p = drvthis->private_data;

/* each bignum is constructed in a 3 x 4 matrix and consists
 * of only the 8 characters defined above as well as ' '
 *
 * The following table defines the 11 big numbers '0'-'9', ':'
 * and the custom base characters they consist of
 */
char bignum_map[11][4][3] = {
  { /* 0: */
    {  1,  2,  3 },
    {  6, 32,  6 },
    {  6, 32,  6 },
    {  7,  2, 32 } },
  { /* 1: */
    {  7,  6, 32 },
    { 32,  6, 32 },
    { 32,  6, 32 },
    {  7,  2, 32 } },
  { /* 2: */
    {  1,  2,  3 },
    { 32,  5,  0 },
    {  1, 32, 32 },
    {  2,  2,  0 } },
  { /* 3: */
    {  1,  2,  3 },
    { 32,  5,  0 },
    {  3, 32,  6 },
    {  7,  2, 32 } },
  { /* 4: */
    { 32,  3,  6 },
    {  1, 32,  6 },
    {  2,  2,  6 },
    { 32, 32,  0 } },
  { /* 5: */
    {  1,  2,  0 },
    {  2,  2,  3 },
    {  3, 32,  6 },
    {  7,  2, 32 } },
  { /* 6: */
    {  1,  2, 32 },
    {  6,  5, 32 },
    {  6, 32,  6 },
    {  7,  2, 32 } },
  { /* 7: */
    {  2,  2,  6 },
    { 32,  1, 32 },
    { 32,  6, 32 },
    { 32,  0, 32 } },
  { /* 8: */
    {  1,  2,  3 },
    {  4,  5,  0 },
    {  6, 32,  6 },
    {  7,  2, 32 } },
  { /* 9: */
    {  1,  2,  3 },
    {  4,  3,  6 },
    { 32,  1, 32 },
    {  7, 32, 32 } },
  { /* colon: (only 1st column used) */
    { 32, 32, 32 },
    {  0, 32, 32 },
    {  0, 32, 32 },
    { 32, 32, 32 } }
};

  if ((num < 0) || (num > 10))
    return;

  if ((p->width >= 20) && (p->height >= 4)) {
    int y = (p->height - 2) / 2;  /* center vertically */
    int x2, y2;

    CF63oddtf_init_num(drvthis);

    for (x2 = 0; x2 < 3; x2++) {
      for (y2 = 0; y2 < 4; y2++) {
        CF63oddtf_chr(drvthis, x+x2, y+y2, bignum_map[num][y2][x2]);
      }
      if (num == 10)
        x2 = 2; /* = break, for colon only */
    }
  }
  else {
    CF63oddtf_chr(drvthis, x, 1 + (p->height - 1) / 2,
            (num == 10) ? ':' : (num + '0'));
  }
}


/*
 * Sets a custom character from 0 - (NUM_CCs-1)
 *
 * For input, values > 0 mean "on" and values <= 0 are "off".
 *
 * The input is just an array of characters...
 */
MODULE_EXPORT void
CF63oddtf_set_char (Driver *drvthis, int n, char *dat)
{
  PrivateData *p = drvthis->private_data;
  static COMMAND_PACKET st_char= { CF633_Set_LCD_Special_Character_Data,
    9, {0,0,0,0,0,0,0,0,0}};
  static char p_dat[48]; // [p->cellheight*p->cellwidth]

  if ((n < 0) || (n >= NUM_CCs) || (!dat)   || 
      ((st_char.data[0]==n) &&
       ( bcmp(dat, p_dat, p->cellheight*p->cellwidth) == 0)))
    return;
  int row, col;
  st_char.data[0] = n;  /* Custom char to define. xxx */
  for (row = 0; row < p->cellheight; row++) {
    int letter = 0;

    for (col = 0; col < p->cellwidth; col++) {
      p_dat[(row * p->cellwidth) + col]=dat[(row * p->cellwidth) + col];
      letter <<= 1;
      letter |= (dat[(row * p->cellwidth) + col] > 0);
    }
    st_char.data[row+1] = letter;
  }
  send_packet(drvthis, &st_char);
}


/*
 * Places an icon on screen
 */
MODULE_EXPORT int
CF63oddtf_icon (Driver *drvthis, int x, int y, int icon)
{
  PrivateData *p = drvthis->private_data;
  char icons[8][6 * 8] = {
  /* Empty Heart */
    {
     0, 1, 1, 1, 1, 1,
     0, 1, 0, 1, 0, 1,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 1, 0, 0, 0, 1,
     0, 1, 1, 0, 1, 1,
     0, 1, 1, 1, 1, 1,
     },
  /* Filled Heart */
    {
     0, 1, 1, 1, 1, 1,
     0, 1, 0, 1, 0, 1,
     0, 0, 1, 0, 1, 0,
     0, 0, 1, 1, 1, 0,
     0, 0, 1, 1, 1, 0,
     0, 1, 0, 1, 0, 1,
     0, 1, 1, 0, 1, 1,
     0, 1, 1, 1, 1, 1,
     },
  /* arrow_up */
    {
     0, 0, 0, 1, 0, 0,
     0, 0, 1, 1, 1, 0,
     0, 1, 0, 1, 0, 1,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 0, 0, 0,
     },
  /* arrow_down */
    {
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 1, 0, 1, 0, 1,
     0, 0, 1, 1, 1, 0,
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 0, 0, 0,
     },
  /* checkbox_off */
    {
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 1, 1, 1, 1, 1,
     0, 1, 0, 0, 0, 1,
     0, 1, 0, 0, 0, 1,
     0, 1, 0, 0, 0, 1,
     0, 1, 1, 1, 1, 1,
     0, 0, 0, 0, 0, 0,
     },
  /* checkbox_on */
    {
     0, 0, 0, 1, 0, 0,
     0, 0, 0, 1, 0, 0,
     0, 1, 1, 1, 0, 1,
     0, 1, 0, 1, 1, 0,
     0, 1, 0, 1, 0, 1,
     0, 1, 0, 0, 0, 1,
     0, 1, 1, 1, 1, 1,
     0, 0, 0, 0, 0, 0,
     },
  /* checkbox_gray */
    {
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 1, 1, 1, 1, 1,
     0, 1, 0, 1, 0, 1,
     0, 1, 1, 0, 1, 1,
     0, 1, 0, 1, 0, 1,
     0, 1, 1, 1, 1, 1,
     0, 0, 0, 0, 0, 0,
     },
   /* Ellipsis */
    {
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0,
     0, 1, 0, 1, 0, 1,
     },
  };

  /* Yes we know, this is a VERY BAD implementation :-) */
  switch (icon) {
    case ICON_BLOCK_FILLED:
      if (p->model == 633)
        CF63oddtf_chr(drvthis, x, y, 255);
      else
        CF63oddtf_raw_chr(drvthis, x, y, 31);
      break;
    case ICON_HEART_FILLED:
      CF63oddtf_set_char(drvthis, 0, icons[1]);
      CF63oddtf_chr(drvthis, x, y, 0);
      break;
    case ICON_HEART_OPEN:
      CF63oddtf_set_char(drvthis, 0, icons[0]);
      CF63oddtf_chr(drvthis, x, y, 0);
      break;
    case ICON_ARROW_UP:
      if (p->model == 633) {
        CF63oddtf_set_char(drvthis, 1, icons[2]);
        CF63oddtf_chr(drvthis, x, y, 1);
      }
      else
        CF63oddtf_raw_chr(drvthis, x, y, 0xDE);
      break;
    case ICON_ARROW_DOWN:
      if (p->model == 633) {
        CF63oddtf_set_char(drvthis, 2, icons[3]);
        CF63oddtf_chr(drvthis, x, y, 2);
      }
      else
        CF63oddtf_raw_chr(drvthis, x, y, 0xE0);
      break;
    case ICON_ARROW_LEFT:
      if (p->model == 633)
        CF63oddtf_raw_chr(drvthis, x, y, 0x7F);
      else
        CF63oddtf_raw_chr(drvthis, x, y, 0xE1);
      break;
    case ICON_ARROW_RIGHT:
      if (p->model == 633)
        CF63oddtf_raw_chr(drvthis, x, y, 0x7E);
      else
        CF63oddtf_raw_chr(drvthis, x, y, 0xDF);
      break;
    case ICON_CHECKBOX_OFF:
      CF63oddtf_set_char(drvthis, 3, icons[4]);
      CF63oddtf_chr(drvthis, x, y, 3);
      break;
    case ICON_CHECKBOX_ON:
      CF63oddtf_set_char(drvthis, 4, icons[5]);
      CF63oddtf_chr(drvthis, x, y, 4);
      break;
    case ICON_CHECKBOX_GRAY:
      CF63oddtf_set_char(drvthis, 5, icons[6]);
      CF63oddtf_chr(drvthis, x, y, 5);
      break;
    case ICON_SELECTOR_AT_LEFT:
      if (p->model == 633)
        return -1;
      CF63oddtf_raw_chr(drvthis, x, y, 0x10);
      break;
    case ICON_SELECTOR_AT_RIGHT:
      if (p->model == 633)
        return -1;
      CF63oddtf_raw_chr(drvthis, x, y, 0x11);
      break;
    default:
      return -1; /* Let the core do other icons */
  }
  return 0;
}


/*
 * Clears the LCD screen
 */
MODULE_EXPORT void
CF63oddtf_clear (Driver *drvthis)
{
  PrivateData *p = drvthis->private_data;

  memset(p->framebuf, ' ', p->width * p->height);
  p->ccmode = standard;
}


/*
 * Hardware clears the LCD screen
 */
static void
CF63oddtf_hardware_clear (Driver *drvthis)
{
  COMMAND_PACKET clear;
  clear.command=CF633_Clear_LCD_Screen;
  clear.data_length=0;
  send_packet (drvthis, &clear);
}


/*
 * Prints a string on the lcd display, at position (x,y).  The
 * upper-left is (1,1), and the lower right is (p->width, p->height).
 */
MODULE_EXPORT void
CF63oddtf_string (Driver *drvthis, int x, int y, char string[])
{
  PrivateData *p = drvthis->private_data;
  int i;

  /* Convert 1-based coords to 0-based... */
  x--;
  y--;

  for (i = 0; string[i] != '\0'; i++) {
    /* Check for buffer overflows... */
    if ((y * p->width) + x + i > (p->width * p->height))
      break;
    p->framebuf[(y * p->width) + x + i] = (p->model == 633)
                  ? string[i]
                  : CFontz_charmap[(unsigned) string[i]];
  }
}


#ifdef FIXMECF635LEDS
/*
 * Output values using the LEDs of a CF635
 *
* FIXME 63oddtf  led use does not respect / update / use p->LEDstate 
 */
MODULE_EXPORT void
CF63oddtf_output(Driver *drvthis, int state)
{
  PrivateData *p = drvthis->private_data;
  if (p->model != 635)
    return;
  int lednum;
  static const unsigned char CFontz635_LEDs[CF635_NUM_LEDs] = {
    11, 9, 7, 5,  // Green LEDs first, Top first
    12,10, 8, 6,  // Red LEDs next, Top first
  };

  COMMAND_PACKET leds;
  leds.command=CF633_Set_GPIO_Pin;
  leds.data_length=2;
  

  for (lednum = 0; lednum < CF635_NUM_LEDs; lednum++) {
    unsigned int mask = (1 << lednum);
    int on_off = (state & mask);
    
    if ((p->LEDstate & mask) != on_off) {
      leds.data[0]=CFontz635_LEDs[lednum];  
      leds.data[1]=(on_off == 0) ? 0 : 100;
      send_packet (drvthis, &leds);
    }
  }
  p->LEDstate = state;
}
#endif // FIXMECF635LEDS

/*
 * KeyRing handling functions.
 * This separates the producer from the consumer.
 * It is just a small fifo of unsigned char.
 */

/** initialize/empty key ring by resetting its read & write pointers */
void EmptyKeyRing(KeyRing *kr)
{
  kr->head = kr->tail = 0;
}


/** add byte to key ring; return success (byte added) / failure (key ring is full) */
int AddKeyToKeyRing(KeyRing *kr, unsigned char key)
{
  if (((kr->head + 1) % KEYRINGSIZE) != (kr->tail % KEYRINGSIZE)) {
     /*  fprintf(stderr,"We add key: %d\n", key); */

          kr->contents[kr->head % KEYRINGSIZE] = key;
      kr->head = (kr->head + 1) % KEYRINGSIZE;
    return 1;
  }
  
  /* KeyRing overflow: do not accept extra key */
  return 0;
}


/** get byte from key ring (or '\0' if key ring is empty) */
unsigned char GetKeyFromKeyRing(KeyRing *kr)
{
  unsigned char retval = '\0';

  kr->tail %= KEYRINGSIZE;

  if ((kr->head % KEYRINGSIZE) != kr->tail) {
    retval = kr->contents[kr->tail];
          kr->tail = (kr->tail + 1) % KEYRINGSIZE;
  }

  /*  if (retval) fprintf(stderr,"We remove key: %d\n", retval); */
  return retval;
}

/** send outgoing_response to the given handle; calc & send CRC when doing so */
void
send_packet(Driver *drvthis, COMMAND_PACKET *out)
{
  pthread_mutex_lock(&quemutex); /* start send_packet */ 
  PrivateData *p = drvthis->private_data;
  // unsigned char cmd=(0x3F&(out->command));
  if ((p->q_depth==0) || (p->sendq[p->sending].command==CF633_Dummy_Unused) ) {
    // q_d ==0 && pk==Dummy checked in opposite order in which set on acks in listen / sensors ...
    // if between depth==0, commit -> time -> report -> show & memcpy, listen pack wants to send
    // then someting would have to be very broken
    commit_packet(drvthis, out);
    gettimeofday (&p->tsent, NULL);
#ifdef DEBUG
    debug(RPT_QUE, "%s on q     %02x (%03d) %32s  send     %32s %02x %02x (%03d)",
        __FUNCTION__,
        p->sending, p->q_depth, p->pkn[p->sendq[p->sending].command], p->pkn[out->command],
        p->sending, p->endq, p->q_depth+1); 
    show_packet(drvthis, RPT_DEBUG, out);
#endif
    memcpy(&p->sendq[p->sending].command, &out->command, out->data_length+2);
    p->q_depth++;
    pthread_mutex_unlock(&quemutex); /* exit send_packet send now */ 
    return;
  } // end send now OK

/* not sending now, add to que */
  memcpy(&p->sendq[p->endq].command, &out->command, out->data_length+2);
  p->endq++;
  p->q_depth++;
#ifdef DEBUG
  debug(RPT_QUE, "%s on q     %02x (%03d) %32s    add    %32s %02x %02x (%03d)",
     __FUNCTION__, p->sending, p->q_depth-1, p->pkn[p->sendq[p->sending].command],
     p->pkn[out->command], p->sending, p->endq, p->q_depth);
#endif
#ifdef TIMING_REPORT 
  if ( p->q_depth > p->maxq){
    p->maxq=p->q_depth;
#ifdef DEBUG
    debug(RPT_QUE, "%s new maximum que depth: %u.", __FUNCTION__, p->maxq );
#endif
  }
#endif // TIMING_REPORT
  if (p->q_depth> 128) {
#ifdef TIMING_REPORT
    p->q_recover=p->q_depth;
#endif // TIMING_REPORT
    if (p->q_depth==UCHAR_MAX-40){
      purge_que(drvthis, __FUNCTION__);
      pthread_mutex_unlock(&quemutex); /* exit send_packet purge_que */ 
      return; 
    } // q_depth==UCHAR_MAX-40
  } // q_depth > 128
  pthread_mutex_unlock(&quemutex); /* exit send_packet packet was qued */ 
} // end of send packet

/* empty que because it is overflowing / full
 * or a key has been pressed and user wants zero
 * lag.
 */
static void purge_que (Driver *drvthis, const char *caller){
  PrivateData *p = drvthis->private_data;
  //struct timespec nod = {0, 000000100};
  /* Let p->sending throug  p->sending+? proceed narmally so
   * acks and send nexts and resends from sending end
   * of que will not be altered by this cleanup */
  // pthread_mutex_lock(&quemutex); /* purge was called by send p which has lock already */ 
  unsigned char i=p->sending+4, 
           pcount=0, 
           ps=p->sending, 
           pd=p->q_depth,
           scd=0, 
           pe=p->endq;
  report(RPT_NOTICE, "%s PURGING QUE  %02x %02x (%03d) %s, start purge %02x -> %02x",
      __FUNCTION__, p->sending, p->endq, p->q_depth, caller, i, pe);
  p->endq=p->sending+4;
  p->q_depth=4;
  do {
//    if (i%40==0) nanosleep(&nod, NULL);  // break up purge loop? 
#ifdef TIMING_REPORT
        //count dropped packets
	p->ack_c[p->sendq[i].command][1]++;
#endif
        if (p->sendq[i].command==CF633_Set_LCD_Special_Character_Data) scd++;
        p->sendq[i].command=CF633_Dummy_Unused;
        i++;
	pcount++;
  } while (i != pe);
  p->priority_packet_need|=4; // send a current backlight command
  if ((p->q_depth !=4) || (ps!=p->sending))
    debug(RPT_NOTICE, "%s p->sending %02x -> %02x    p->endq %02x -> %02x    (%03d)",
      __FUNCTION__, ps, p->sending,  pe, p->endq, p->q_depth);
#ifdef TIMING_REPORT
  report(RPT_NOTICE, "%s %02x %02x (%03d) >  %02x %02x (%03d)  %03d p, % 03.1f%% S_Ch_D, %s by %s %5lu g, %03.1f%% overload",
    __FUNCTION__, ps, pe, pd, 
    p->sending, p->endq, p->q_depth, pcount,  
    (float)scd/(float)pcount*100,
    p->device, caller, 
    p->ack_c_t - p->prevp, (float)pcount / ((float)p->ack_c_t - (float)p->prevp ) *100);
  p->prevp=p->ack_c_t;
  p->q_purged+=pcount;
  p->q_recover=0; // not a 'natural' recovery from depth
#else
  report(RPT_NOTICE, "%s %02x %02x (%03d) >  %02x %02x (%03d)  %03d p,  %03.1f%% S_Ch_D, %s by %s",
    __FUNCTION__,  ps, pe, pd, 
    p->sending, p->endq, p->q_depth, pcount,
    (float)scd/(float)pcount*100,
    p->device, caller);
#endif
  // pthread_mutex_unlock(&quemutex); /* purge by send p which has a lock already*/ 
} // end purge_que 

/** calculate CRC over given buffer with given length */
int
get_crc(char *buf, int len, int seed)
{

  /* CRC lookup table to avoid bit-shifting loops. */
  static const word crcLookupTable[256] = {
    0x00000, 0x01189, 0x02312, 0x0329B, 0x04624, 0x057AD, 0x06536, 0x074BF,
    0x08C48, 0x09DC1, 0x0AF5A, 0x0BED3, 0x0CA6C, 0x0DBE5, 0x0E97E, 0x0F8F7,
    0x01081, 0x00108, 0x03393, 0x0221A, 0x056A5, 0x0472C, 0x075B7, 0x0643E,
    0x09CC9, 0x08D40, 0x0BFDB, 0x0AE52, 0x0DAED, 0x0CB64, 0x0F9FF, 0x0E876,
    0x02102, 0x0308B, 0x00210, 0x01399, 0x06726, 0x076AF, 0x04434, 0x055BD,
    0x0AD4A, 0x0BCC3, 0x08E58, 0x09FD1, 0x0EB6E, 0x0FAE7, 0x0C87C, 0x0D9F5,
    0x03183, 0x0200A, 0x01291, 0x00318, 0x077A7, 0x0662E, 0x054B5, 0x0453C,
    0x0BDCB, 0x0AC42, 0x09ED9, 0x08F50, 0x0FBEF, 0x0EA66, 0x0D8FD, 0x0C974,
    0x04204, 0x0538D, 0x06116, 0x0709F, 0x00420, 0x015A9, 0x02732, 0x036BB,
    0x0CE4C, 0x0DFC5, 0x0ED5E, 0x0FCD7, 0x08868, 0x099E1, 0x0AB7A, 0x0BAF3,
    0x05285, 0x0430C, 0x07197, 0x0601E, 0x014a1, 0x00528, 0x037B3, 0x0263A,
    0x0DECD, 0x0CF44, 0x0FDDF, 0x0EC56, 0x098E9, 0x08960, 0x0BBFB, 0x0AA72,
    0x06306, 0x0728F, 0x04014, 0x0519D, 0x02522, 0x034AB, 0x00630, 0x017B9,
    0x0EF4E, 0x0FEC7, 0x0CC5C, 0x0DDD5, 0x0A96A, 0x0B8E3, 0x08A78, 0x09BF1,
    0x07387, 0x0620E, 0x05095, 0x0411C, 0x035A3, 0x0242A, 0x016B1, 0x00738,
    0x0FFCF, 0x0EE46, 0x0DCDD, 0x0CD54, 0x0B9EB, 0x0A862, 0x09AF9, 0x08B70,
    0x08408, 0x09581, 0x0A71A, 0x0B693, 0x0C22C, 0x0D3A5, 0x0E13E, 0x0F0B7,
    0x00840, 0x019C9, 0x02B52, 0x03ADB, 0x04E64, 0x05FED, 0x06D76, 0x07CFF,
    0x09489, 0x08500, 0x0B79B, 0x0A612, 0x0D2AD, 0x0C324, 0x0F1BF, 0x0E036,
    0x018C1, 0x00948, 0x03BD3, 0x02A5A, 0x05EE5, 0x04F6C, 0x07DF7, 0x06C7E,
    0x0A50A, 0x0B483, 0x08618, 0x09791, 0x0E32E, 0x0F2A7, 0x0C03C, 0x0D1B5,
    0x02942, 0x038CB, 0x00A50, 0x01BD9, 0x06F66, 0x07EEF, 0x04C74, 0x05DFD,
    0x0B58B, 0x0A402, 0x09699, 0x08710, 0x0F3AF, 0x0E226, 0x0D0BD, 0x0C134,
    0x039C3, 0x0284A, 0x01AD1, 0x00B58, 0x07FE7, 0x06E6E, 0x05CF5, 0x04D7C,
    0x0C60C, 0x0D785, 0x0E51E, 0x0F497, 0x08028, 0x091A1, 0x0A33A, 0x0B2B3,
    0x04A44, 0x05BCD, 0x06956, 0x078DF, 0x00C60, 0x01DE9, 0x02F72, 0x03EFB,
    0x0D68D, 0x0C704, 0x0F59F, 0x0E416, 0x090A9, 0x08120, 0x0B3BB, 0x0A232,
    0x05AC5, 0x04B4C, 0x079D7, 0x0685E, 0x01CE1, 0x00D68, 0x03FF3, 0x02E7A,
    0x0E70E, 0x0F687, 0x0C41C, 0x0D595, 0x0A12A, 0x0B0A3, 0x08238, 0x093B1,
    0x06B46, 0x07ACF, 0x04854, 0x059DD, 0x02D62, 0x03CEB, 0x00E70, 0x01FF9,
    0x0F78F, 0x0E606, 0x0D49D, 0x0C514, 0x0B1AB, 0x0A022, 0x092B9, 0x08330,
    0x07BC7, 0x06A4E, 0x058D5, 0x0495C, 0x03DE3, 0x02C6A, 0x01EF1, 0x00F78
  };

      /* initialize the new crc with the given seed */
  int newCrc = seed;

  /* This algorithm is based on the IrDA LAP example */
  while (len-- > 0)
    newCrc = (newCrc >> 8) ^ crcLookupTable[(newCrc ^ *buf++) & 0xff];

  /* Make this crc match the one's complement that is sent in the packet */
  return (~newCrc);
}




/*============================================================================
 * 635 WinTest Code.
 * Copyright 2001, Crystalfontz America, Inc. Written by Brent A. Crosby
 * www.crystalfontz.com, brent@crystalfontz.com
 *============================================================================
 */

/*---------------------------------------------------------------------------*/
/* This is some code that kind of makes the windows stuff look a little
 * like the DOS/633 interrupt driven serial stuff. Basically there is a
 * circular buffer, and SyncReceiveBuffer() uses ReadFile() to put data
 * into the circular buffer much like the DOS stuff uses an ISR to put
 * the data into the buffer. Then the rest of the functions work like
 * the counterparts in the 633,
 */



/*                  v TailPeek                        */
/* -------------------------------------------------- */
/*           ^ Tail              ^ Head               */


/** initialize/empty receive buffer by resetting its pointers */
void EmptyReceiveBuffer(ReceiveBuffer *rb)
{
  rb->head = rb->tail = rb->peek = 0;
}


/** read given number of bytes from given file handle into receive buffer */
void SyncReceiveBuffer(ReceiveBuffer *rb, int fd, unsigned int number)
{
  unsigned char  buffer[MAX_DATA_LENGTH];
  int    BytesRead;

  if (number > MAX_DATA_LENGTH)
    number = MAX_DATA_LENGTH;
  BytesRead = read(fd, buffer, number);

  if (BytesRead == -1) {
    /* fprintf(stderr,"~~~Problem reading: %s .\n", strerror(errno)); */
  }
  else {
    int  i;

    /* fprintf(stderr,"Read %d Bytes:", BytesRead); */

    /* wrap write pointer to the receive buffer */
    rb->head %= RECEIVEBUFFERSIZE;

    /* store the bytes read */
    for (i = 0; i < BytesRead; i++) {
      /* fprintf(stderr," %02x", buffer[i]); */
      rb->contents[rb->head] = buffer[i];

      /* increment write pointer (wrap if needed) */
      rb->head = (rb->head + 1) % RECEIVEBUFFERSIZE;
    }
    /* fprintf(stderr,"\n"); */
  }
}


/** return number of bytes available for reading in receive buffer */
int BytesAvail(ReceiveBuffer *rb)
{
  int avail_bytes = rb->head - rb->tail;

  if (avail_bytes < 0)
    avail_bytes += RECEIVEBUFFERSIZE;

  return(avail_bytes % RECEIVEBUFFERSIZE);
}


/** get next byte from receive buffer (return '\0' if buffer is empty) */
unsigned char GetByte(ReceiveBuffer *rb)
{
  unsigned char    return_byte = '\0';
  
  /* wrap read pointer to the receive buffer */
  rb->tail %= RECEIVEBUFFERSIZE;

  /* See if there are any more bytes available. */
  if (rb->tail != (rb->head % RECEIVEBUFFERSIZE)) {
    /* There is at least one more byte. */
    return_byte = rb->contents[rb->tail];

    /* Increment read pointer (wrap if needed) */
    rb->tail = (rb->tail + 1) % RECEIVEBUFFERSIZE;
  }

  return(return_byte);
}


/** return number of bytes available for peeking in receive buffer */
int PeekBytesAvail(ReceiveBuffer *rb)
{
  int avail_bytes = rb->head - rb->peek;
  
  if (avail_bytes < 0)
    avail_bytes += RECEIVEBUFFERSIZE;

  return(avail_bytes % RECEIVEBUFFERSIZE);
}


/** sync peek pointer with read pointer */
void SyncPeekPointer(ReceiveBuffer *rb)
{
  rb->peek = rb->tail;
}


/** accept ppeked data by syncing the read pointer to the peek pointer */
void AcceptPeekedData(ReceiveBuffer *rb)
{
  rb->tail = rb->peek;
}


/** peek next byte from receive buffer (return '\0' if buffer is empty) */
unsigned char PeekByte(ReceiveBuffer *rb)
{
  unsigned char return_byte = '\0';

  /* wrap peek pointer to the receive buffer */
  rb->peek %= RECEIVEBUFFERSIZE;

  /* See if there are any more bytes available. */
  if (rb->peek != (rb->head % RECEIVEBUFFERSIZE)) {
    /* There is at least one more byte. */
    return_byte = rb->contents[rb->peek];

    /* Increment the peek pointer (wrap if needed). */
    rb->peek = (rb->peek + 1) % RECEIVEBUFFERSIZE;
  }

  return(return_byte);
}

/*============================================================================
 *                              check_for_packet()
 *
 * check_for_packet() will see if there is a valid packet in the input buffer.
 * If there is, it will copy it into p->incoming_command and return 1. If there
 * is not it will return 0. incoming_packet may get partially filled with
 * garbage if there is not a valid packet available.
 *----------------------------------------------------------------------------
 */

/* Let's return
 * O if we have no message but we should try again immediatly
 * 1 if we have a message correctly identified
 * 2 if we have no message and we should not retry until new input
 * So a loop should run as long as we have no 0
 * If we have a 2 we should avoid comming back there.
 */
int
check_for_packet(Driver *drvthis, unsigned char expected_length)
{
  PrivateData *p = drvthis->private_data;
  int i;
  int testcrc;

  SyncReceiveBuffer(&p->receivebuffer, p->fd, expected_length);

  //First off, there must be at least 4 bytes available in the input stream
  //for there to be a valid command in it (command, length, no data, CRC).
  if (BytesAvail(&p->receivebuffer) < 4) {
    /* fprintf(stderr,"Not enough byte available for even the smallest message.\n"); */
    return(GIVE_UP); /* We don't need to retry before more byte are received */
  }

  /* Look into the buffer without removing the data. */
  SyncPeekPointer(&p->receivebuffer);

  p->incoming_command.command = PeekByte(&p->receivebuffer);

  /* Only commands 0 through MAX_COMMAND are valid */
  if (MAX_COMMAND < (0x3F & p->incoming_command.command)) {
    /* Throw out one byte of garbage. Next pass through should re-sync. */
    GetByte(&p->receivebuffer);
    /* fprintf(stderr,"###: Unknown command.\n"); */
    return(TRY_AGAIN);
  }
  
    /* There is a valid command byte. Get the data_length. */
  p->incoming_command.data_length = PeekByte(&p->receivebuffer);
  
    /* The data length must be within reason. */
  if(MAX_DATA_LENGTH < p->incoming_command.data_length) {
    //Throw out one byte of garbage. Next pass through should re-sync.
    GetByte(&p->receivebuffer);
    /* fprintf(stderr,"###: Too long packet: %d.\n", p->incoming_command.data_length); */
    return(TRY_AGAIN);
  }

  // Now there must be at least p->incoming_command.data_length + sizeof(CRC) bytes
  // still available for us to continue.
  if ((int) PeekBytesAvail(&p->receivebuffer) < (p->incoming_command.data_length + 2)) {
    //It looked like a valid start of a packet, but it does not look
    //like the complete packet has been received yet.
    /* fprintf(stderr,"Not enough read to check the complete message.\n"); */
    return(GIVE_UP); /* Let's not return until more byte are available */
  }

  /* There is enough data to make a packet. Transfer over the data. */
  for (i = 0; i < p->incoming_command.data_length; i++)
    p->incoming_command.data[i] = PeekByte(&p->receivebuffer);

  //Now move over the CRC.
  p->incoming_command.crc.as_bytes[0] = PeekByte(&p->receivebuffer);
  p->incoming_command.crc.as_bytes[1] = PeekByte(&p->receivebuffer);
  //Now check the CRC.

  //Compute the expected CheckSum
  testcrc = get_crc(( char *) &p->incoming_command,
        p->incoming_command.data_length+2, 0xFFFF);
  testcrc = testcrc & 0xFFFF; /* This is TRICKY */

  if(p->incoming_command.crc.as_word == testcrc) {
    //This is a good packet. I'll be horn swaggled. Remove the packet
    //from the serial buffer.
    AcceptPeekedData(&p->receivebuffer);
    //Let our caller know that p->incoming_command has good stuff in it.
    report(RPT_DEBUG, "%s p->incoming_command has good stuff in it:", __FUNCTION__);
#ifdef DEBUG
    show_packet(drvthis, RPT_DEBUG, &p->incoming_command); 
#endif

    return(GOOD_MSG);
    // note: p->incoming_command.data[p->incoming_command.datalength+1]
    // is unknown, so printing p->incoming_command.data, if it is text
    // is not gauranteed a string terminator \0 \n ...
  }

  /* The CRC did not match. Toss out one byte of garbage.
  * Next pass through should re-sync.
  */
  GetByte(&p->receivebuffer);
  /* fprintf(stderr,"###: Wrong CheckSum. computed/real %04x:%04x \n",
     testcrc, p->incoming_command.crc.as_word); */
  return(TRY_AGAIN);
}

#ifdef DEBUG
//============================================================================
void show_packet(Driver* drvthis, int rep_level,  COMMAND_PACKET *packet)
{
  PrivateData *p = drvthis->private_data;
        char target[8];
        char packstr[155]={'\0'}; // MAX_DATA_LENGTH 22 *3 + name + ~36 + slop
            // longest I've seen is ~ 126 ?
        int usedlength=0;
        int i, cmd, top, len;

        top=(0xC0&(packet->command)) >> 6;
        cmd=(0x3F&(packet->command));
        len=packet->data_length;
  if (top == 2){
    if (cmd == CF633_Write_User_Flash_Area) // temp packet 
          sprintf(packstr, "Packet (top %02d, %s Rpt) len %02d [", top, 
	    ((p->ds18!=NULL) && (p->ds18[packet->data[0]]!=NULL) && (p->ds18[packet->data[0]]->name !=NULL)) ? p->ds18[packet->data[0]]->name : "nameless", 
	    len);
    if (cmd == CF633_Get_Hardware_And_Firmware_Version ) // a fan packet 
          sprintf(packstr, "Packet (top %02d, %s Rpt) len %02d [", top, p->fans[packet->data[0]].name, len);
  }
  else { // is a normal response packet or error packet
          sprintf(packstr, "Packet (top %02d, cmd %02d %s ) len %02d [", top, cmd, p->pkn[cmd], len);
  }
  usedlength =strlen(packstr);
  char istext [MAX_DATA_LENGTH];
  if (packet->data_length!=0) {
    for(i=0 ;i < packet->data_length; i++){
      sprintf(target, "%02x ", packet->data[i]);
      sprintf(packstr+usedlength , "%s ",target );
      usedlength+=strlen(target);
      if ((cmd==CF633_Send_Data_to_LCD) && ( i > 1 )){
        istext[i-2]=packet->data[i];
        istext[i-1]=0;
      }
    }
  }
  sprintf(packstr+usedlength, "]\n");
        report(rep_level, "%s", packstr);
  if (cmd==CF633_Send_Data_to_LCD){
    debug(rep_level, "     sData_to_LCD     %02d at %02d x %d '%s'",
        packet->data_length-2, packet->data[0],
        packet->data[1], istext);
  }
  debug(RPT_DEBUG, "%s strlen is %d, data_length*3 is %d",
      __FUNCTION__ , strlen(packstr), packet->data_length *3);
} // end of show packet
#endif // DEBUG

//============================================================================
/* multithreaded : 
 * do not change anything in p -> outgoing command -- might be in use elsewhere
 * do only one write(), no chopping up any remaining incremental writes elsewhere
 */
void commit_packet(Driver* drvthis,  COMMAND_PACKET *packet)
{
  PrivateData *p = drvthis->private_data;
  word tmp_crc;         
  ubyte send_this[2+MAX_DATA_LENGTH]; 
  int retval, errno_sav;
  send_this[0]=packet->command;  
  send_this[1]=packet->data_length;  
      
  if (packet->data_length > 0){
    //write(p->fd, packet.data, packet.data_length);
    memcpy(&send_this[2], &packet->data, packet->data_length);
      }
  tmp_crc= get_crc((char *) packet, 
        packet->data_length
        + 2, 0xFFFF);
  memcpy(&send_this[2+packet->data_length], &tmp_crc, 2);
#ifdef DEBUG  
  show_packet(drvthis, RPT_DEBUG, packet);
#endif
  while (!(retval=write(p->fd, send_this, 4 + packet->data_length))){
    errno_sav=errno;
    if ((errno_sav==EAGAIN) || (errno_sav==EWOULDBLOCK)){
      report(RPT_CRIT, "%s write() error: %s. Resend.",
        __FUNCTION__, strerror (errno_sav));
/* 2012 04 19 This was exiting after 90 minutes. 
 * write() error.  Resource temporarily unavailable EXIT_FAILURE
 * Started after adding my 635 led functions
 *09:11:41 furniture LCDd:      0 Key    3654   Fan  14616  Temp
 *09:11:41 furniture LCDd:  46994 GOOD       0 TRY_AGAIN   25685 GIVE_UP
 *09:11:41 furniture LCDd:  count     %    min t     max t  avg    timeout   dump resent xtra
 *09:11:41 furniture LCDd:   7200  25.06  0.004139  0.045  0.016   0.077     0     0     0    Set_LCD_Special_Character_Data
 *09:11:41 furniture LCDd:    120  00.41  0.003302  0.039  0.018   0.023     0     0     0    Set_Fan_Power
 *09:11:41 furniture LCDd:   2840  09.88  0.003883  0.041  0.021   0.033     0     0     0    Send_Data_to_LCD
 *09:11:41 furniture LCDd:  18564  64.62  0.020156  0.041  0.020   0.800     0     0     0    Set_GPIO_Pin
 *09:11:41 furniture LCDd:  28724         0.003302  0.045  0.019             0     0     0    All
 *09:11:41 furniture LCDd: CFA635:h1.0,v1.3 avg que depth   3.0157 = ~  0.0578 sec., max  14,  times over 128   0
 *09:11:41 furniture LCDd: 28724 good,     0 purged, 0.000 % overload.     15.957 packets/sec, elapse 1800.049
 *09:14:43 furniture LCDd: compute_fan_rpm all leds fan1 rep value low =  133.76
 *09:15:28 furniture LCDd: commit_packet write() error.  Resource temporarily unavailable EXIT_FAILURE
 *All those Set_GPIO_Pin trippled the packets/sec & number of packets. 
*/
    }
    else {
      report(RPT_CRIT, "%s write() error.  %s EXIT_FAILURE",
        __FUNCTION__, strerror (errno_sav));
      CF63oddtf_close (drvthis);
    }  
  }
//  report(RPT_CRIT, "%s sent a ", __FUNCTION__);
//  show_packet(drvthis, RPT_CRIT, packet);
} /* end of commit _ packet */

int
fan_loop(Driver * drvthis)
{
  PrivateData *p = drvthis->private_data;
  int Quit=0;
  int fan_loop_timer=0;
#ifdef FANS_USEC
  struct timeval now;
#endif // FANS_USEC
  struct timespec nod = {0, 125000000}; 
#ifdef __linux__
  debug(RPT_INFO, "pthread_self % 15s = %d tid = %d", 
    __FUNCTION__, pthread_self(), syscall(SYS_gettid));
#endif
  while(!Quit)
  {
    manage_fans(drvthis);
    fan_loop_timer++;
    //struct timeval now;
    // 31-( estimated maximum possible que depth in seconds ) 
    // keep fans from spinning up when que gets way behind
    //if  (time_diff(drvthis, p->tdid_power, now) > 28000000){
    //if  (fan_loop_timer==240){ // 8 loops / sec * 30 sec = 240 loops
    //twice as often as neccessary, fans less likely to spin up on que hang
    if  (fan_loop_timer==120){ // 8 loops / sec * 15 sec = 120 loops
      p->priority_packet_need|=1;
#ifdef FANS_USEC
        gettimeofday(&now,NULL);
        p->treq_power=now; 
        debug(RPT_FANS, "%s need fan pow ack+1       failsafe   %ld.%06ld  %f", 
          __FUNCTION__, now.tv_sec- FSEC , now.tv_usec,
          (float)time_diff(drvthis, p->tdid_power, now)/1000000);
#else // FANS_USEC
        debug(RPT_FANS, "%s need fan pow ack+1       failsafe %d sec.", 
          __FUNCTION__, (fan_loop_timer+1) /8);
#endif // FANS_USEC
      fan_loop_timer=0; 
    } // fan_loop_timer==time to failsafe
    if (1&p->priority_packet_need)  {
      set_fan_pow (drvthis);
    }
    if (2&p->priority_packet_need) {
      set_fan_rpt(drvthis);
    }
    nanosleep(&nod, NULL); // 1/8 sec fan loop  
  } // end while(!Quit)
  return 1;
} // end fan_loop


/* 
 *  show a line of syslog column names 
 */
void csv_log (Driver *drvthis) {
  PrivateData *p = drvthis->private_data;  
  char *syslogline;
  int tmp;
  int lineused=0;
  if (p->log_sens==0) {
    debug(RPT_CLIUP, "%s return on syslogsensors==0", __FUNCTION__);
    return;
  }
#ifdef DO_HDD  
  if (p->hdd_count !=0) { // do this here so no pause for hdd wakeup in middle of logline 
    hddtemp_temps(drvthis, UCHAR_MAX);
  }
#endif
/* 
 * once had a check here to ensure no logging in middle of update stream
 */

  syslogline = calloc (256,1);
  char *buf=calloc(256,1);
  time_t very_now=time(NULL);
  ctime_r(&very_now, buf);
  buf[strlen(buf)-1] ='\0';
  lineused += snprintf (syslogline, 255, buf);
#ifdef DO_ULT  
  if (p->log_uptime!=0) {
    double uptime =0;
    double idle =0;
    machine_get_uptime(&uptime, &idle);
    lineused +=snprintf (&syslogline[lineused], 255, ",%.0f", uptime);
  }
  if ((p->log_load !=0 ) || ( p->log_load_s!=0) || (p->log_load_u!=0)){
    store_load(drvthis);  /* in csv_log */ 
  }
  if ((p->topnum!=0) || (p->topname!=0)){  
    store_tops(drvthis); /* in csv_log */ 
  }
  if (p->log_load!=0) {
    lineused +=snprintf (&syslogline[lineused], 255, ",%03.1f", p->load_t);
  }
  if (p->log_load_s!=0) {
    lineused +=snprintf (&syslogline[lineused], 255, ",%03.1f", p->load_s);
  }
  if (p->log_load_u!=0) {
    lineused +=snprintf (&syslogline[lineused], 255, ",%03.1f", p->load_u);
  }
  for (tmp=0; tmp <p->topnum; tmp++) {
/*
  note to print a % into string use %%,
  to print a % into a stritg that will be printed into a string, use %%%%
*/
    syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
        ",%s", p->top_num[tmp]->name));
    lineused += snprintf (&syslogline[lineused], 255, buf);
    syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
        ",%2.2f", p->top_num[tmp]->val) );
    lineused += snprintf (&syslogline[lineused], 255, buf);
  }
  for (tmp=0; tmp < p->topname; tmp++){
    //if ((p->top_name[tmp]!=NULL) && (p->top_name[tmp]->name!=NULL)) { 
    // fixed: sorted to all non-null < p->topname in ult_init
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%2.2f", p->top_name[tmp]->val ));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    //}
  }// end of each possible topname
#endif // DO_ULT
#ifdef DO_LM
  for (tmp=0; tmp < p->lmnum; tmp++){ 
    update_lm(drvthis, tmp);
    syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
        ",%d", p->lm[tmp]->lm_abstracto/16));
    lineused += snprintf (&syslogline[lineused], 255, buf);
  }
#endif
#ifdef DO_HDD  
  if (p->hdd_count !=0) {
  /* this is done above so hddwakeup delay does not occour while line is being composed
    hddtemp_temps(drvthis, UCHAR_MAX);
  */
    for (tmp=0; tmp  < p->hdd_count ; tmp ++){  // for each hdd
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%f",(float)p->hdd[tmp]->hd_abstracto/16) );
      lineused += snprintf (&syslogline[lineused], 255, buf);
    }// end each drive
  }
#endif // DO_HDD  
  for (tmp=0; tmp <4; tmp++) {
    if (p->fans[tmp].p_log) {
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%03d", p->priority_packet[0].data[tmp]) );
      lineused += snprintf (&syslogline[lineused], 255, buf);
    }
    if (p->fans[tmp].autorep) {
        syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
            ",%5.4f", compute_fan_rpm(drvthis, tmp)));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    }
    if (p->fans[tmp].dt_label != NULL ){
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%3.4f", (float)p->fans[tmp].dt/16));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    }
  } // end of build syslogline for each fan
  if (p->tempcount > 0) {
    for (tmp=0; tmp < p->tempcount; tmp++) {
      compute_temp(drvthis, tmp);
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
            ",%2.4f", (float)(p->ds18[tmp]->abstracto)/16 ));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    } // end of build syslogline for each Temperature  
  } // end of build syslogline for all Temperatures
#ifdef MY_CUSTOM_LINE
      update_wattsline(drvthis,0); // only other call to this is when watts is active screen
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%3.2f", watts_est ));
      lineused += snprintf (&syslogline[lineused], 255, buf);
#endif
  //report(RPT_CRIT, "%s CF633data%s", __FUNCTION__, syslogline);
  fprintf(p->log_fd, "%s \n", syslogline);
  fflush(p->log_fd);
  free (syslogline);
  free (buf);
} // end of csv_log



static void manage_fans  (Driver *drvthis) {
  int fan;
  PrivateData * p = drvthis->private_data; 
  static int per_fan_timer[4]={6400,6400,6400,6400};
#ifdef FANS_USEC
  struct timeval now;
  gettimeofday(&now,NULL); // if fans need adjusting now
#endif // FANS_USEC
  for (fan=0; fan < 4 ; fan ++) {
    // calcs for sensor driven fans
    if ((p->fans[fan].f_sens[0]!=NULL) && (p->fans[fan].f_sens[0]->abstvalptr!=NULL)) {
      /* 2009 05 17 case of f_sens[0] != NULL, but abstvalptr==NULL 
         .. partial deconfiguration ? 
         crashed on dereferencing abstvalptr to get temperat below . 
         Added check here, but partial
         deconfigure could be avoided ? */
      per_fan_timer[fan]++;
      // adjust this fan now ? 
      if (per_fan_timer[fan] + 1 >= (int)p->fans[fan].a_sec * 8 )
	//(time_diff(drvthis,p->fans[fan].tdid_fan, now) >=
        //  (long)p->fans[fan].a_sec*1000000)
        {
        debug(RPT_FANS, "%s fan %d a_sec=%d loops=%d sec=%d.",
          __FUNCTION__, fan+1, p->fans[fan].a_sec, per_fan_timer[fan]+1,
	  (per_fan_timer[fan]+1)/8);
        per_fan_timer[fan]=0;
#ifdef FANS_USEC
        debug(RPT_FANS, "%s fan %d timeout                did %ld.%06ld",
          __FUNCTION__, fan+1,
          p->fans[fan].tdid_fan.tv_sec- FSEC ,
          p->fans[fan].tdid_fan.tv_usec);  
        debug(RPT_FANS, "%s                              now %ld.%06ld  %6.6f",
          __FUNCTION__, now.tv_sec- FSEC ,
          now.tv_usec,
          (float)time_diff (drvthis, p->fans[fan].tdid_fan,now) / 1000000);  
#endif // FANS_USEC
        unsigned char tmp_r_mask=
          p->fans[fan].autorep *
          ( p->priority_packet[0].data[fan]!=0 );
        int temperat;
/* get most recent temp data now! */
        if ((p->fans[fan].f_sens[0]!=NULL) && (p->fans[fan].f_sens[0]->updates!=NULL))
	  p->fans[fan].f_sens[0]->updates(drvthis, p->fans[fan].f_sens[0]->number);
        if ((p->fans[fan].f_sens[1]!=NULL) && (p->fans[fan].f_sens[1]->updates!=NULL))
	  p->fans[fan].f_sens[1]->updates(drvthis, p->fans[fan].f_sens[1]->number);
        if (p->fans[fan].f_sens[1]==NULL) {
	  if ((p->fans[fan].f_sens[0]==NULL)||(p->fans[fan].f_sens[0]->abstvalptr==NULL) ) {
	    report (RPT_NOTICE, "%s fan %d sensor has failed.",
	      __FUNCTION__, fan+1 );
            continue; 
          }
          temperat=*p->fans[fan].f_sens[0]->abstvalptr; 
	  report (RPT_FANS, "%s fan %d simple temperat = %f",
	    __FUNCTION__, fan+1, (float)temperat/16); 
	}
	else {
	  p->fans[fan].dt=temperat= *p->fans[fan].f_sens[0]->abstvalptr 
              - *p->fans[fan].f_sens[1]->abstvalptr;
	  report (RPT_FANS, "%s fan %d relative temperat = %f",
	    __FUNCTION__, fan+1, (float)temperat/16); 
	}
//        float tmp_power=FAN_LOW_POWER_PERCENT+
//          (temperat-p->fans[fan].by_s_off) / 
//          (p->fans[fan].by_s_100-
//          p->fans[fan].by_s_off)
//          *(100-FAN_LOW_POWER_PERCENT);

	report (RPT_FANS, "%s fan %d by_s_off= %f by_s_100=%f",
	    __FUNCTION__, fan+1, (float)p->fans[fan].by_s_off/16, (float)p->fans[fan].by_s_100/16); 
         int tmp_power=
          100*(temperat-p->fans[fan].by_s_off) / 
          (p->fans[fan].by_s_100-
          p->fans[fan].by_s_off);
	report (RPT_FANS, "%s fan %d rough tmp_power= %d\n",
	    __FUNCTION__, fan+1, tmp_power); 
        
        if ( tmp_power > 100){
          tmp_power=100;
        }
        if ( tmp_power < 0 ){
          tmp_power=0;
        }
        if (p->fans[fan].heat){
          tmp_power=100-tmp_power; 
        }
#ifdef CF635LEDS
        if (p->model==635) { 
          if (p->fans[fan].led==1){ /* this fan drives an led based on power */
            set_led(drvthis, fan, tmp_power);
          }
        } /* end it is a 635 */  
#endif
        if (tmp_power > 0) {  /* from 0 to 100 to LOW_POWER_PERCENT to 100 */
          
          tmp_power=(FAN_LOW_POWER_PERCENT+(100-FAN_LOW_POWER_PERCENT)*tmp_power/100);
        }
        if ( (unsigned char)tmp_power != p->priority_packet[0].data[fan]) {
          /* if the power setting for this fan has changed */
#ifdef FANS_USEC
          p->treq_power=now;
#endif // FANS_USEC
          p->priority_packet_need|=1; /* signal commit packet */ 
          debug(RPT_FANS, "%s need fan pow ack+=1 fan %d",
            __FUNCTION__, fan+1);
        }
	report (RPT_FANS, "%s fan %d final tmp_power= %d",
	    __FUNCTION__, fan+1, tmp_power); 
        p->priority_packet[0].data[fan]=(unsigned char)tmp_power;
        // turn off rpts for powered off fans
        if ( tmp_r_mask !=           p->fans[fan].autorep * (p->priority_packet[0].data[fan]!=0 )) {
          p->priority_packet_need|=2;
	  // this is done in set_fan_rpt
	  // p->fans[fan].autorep_mask= p->fans[fan].autorep*(p->priority_packet[0].data[fan]!=0);
        }
        debug(RPT_FANS, 
          "%s Calc %d:   %s=%f, pow %d, priority_packet_need %d.",
          __FUNCTION__, fan+1, 
          p->fans[fan].dt_label != NULL ? p->fans[fan].dt_label : "t",
          (float)temperat/16, p->priority_packet[0].data[fan], p->priority_packet_need);
#ifdef FANS_USEC
        p->fans[fan].tdid_fan=now;
#endif // FANS_USEC
      } // end it is time for this fan
    } // end it is a sensor driven fan
#ifdef CF635LEDS
    if ((p->model==635) && (p->fans[fan].led==3)){
       led_rand (drvthis, fan); /* blink speed controlled in func */
    }
#endif //CF635LEDS 
  } // end check each fan
}// end  manage_fans

#ifdef CF635LEDS
/**
 * Set one model 635 red green led pair to a color understood as 0-100.
 * \param drvthis  Pointer to driver structure.
 * \param fan      fan to led ontology 0-3,  0==fan1, top led;
 * \param power    0-100 -> { off <> green <> amber <> red }
 */
void set_led(Driver *drvthis, unsigned char fan, unsigned char power)
{
  PrivateData *p = drvthis->private_data;
  if (p->model !=635) return;
  static unsigned char ledstate[8] ={ 0,0,0,0, 0,0,0,0};
  static unsigned char oldstate[8] ={255,255,255,255, 255,255,255,255};
  unsigned char simplpow=power/(100/CFLEDBASE);
  switch (simplpow){
    case 0 : /* fan powered off, green & red off */
      ledstate[fan]=ledstate[fan+4]=0;
      break;
    case CFLEDBASE :
      ledstate[fan]=0;
      ledstate[fan+4]=100;
      break;
    default :
       /* changing shades of amber */
        ledstate[fan]=100-(simplpow * (100/CFLEDBASE) ); /* green  */
        ledstate[fan+4]=(simplpow * (100/CFLEDBASE)); /* red */
  } // end power cases
  if (oldstate[fan] != ledstate[fan] ){
//    report(RPT_CRIT, 
//        "%s green fan[%d] pow % 3d(%d/%d): G % 3d -> % 3d", __FUNCTION__ , fan+1, 
//         power, simplpow, CFLEDBASE, oldstate[fan], ledstate[fan]);
    COMMAND_PACKET led;
    led.command=CF633_Set_GPIO_Pin;
    led.data_length=2;
    led.data[0]=11-fan*2; /* green top 11, 9, 7, 5 bottom */
    led.data[1]=ledstate[fan];
    send_packet (drvthis, &led);
    oldstate [fan] = ledstate[fan]; 
  }
  if (oldstate [fan+4] != ledstate[fan+4] ){
//    report(RPT_CRIT, 
//        "%s  red  fan[%d] pow % 3d(%d/%d): R % 3d -> % 3d", __FUNCTION__ , fan+1, 
//         power, simplpow, CFLEDBASE, oldstate[fan+4], ledstate[fan+4]);
    COMMAND_PACKET led;
    led.command=CF633_Set_GPIO_Pin;
    led.data_length=2;
    led.data[0]=12-fan*2; /* red top 12, 10, 8, 6 bottom */
    led.data[1]=ledstate[fan+4];
    send_packet (drvthis, &led);
    oldstate [fan+4] = ledstate[fan+4]; 
  }
  return;
} /* end set led by fan */

/* given a fan number and it's rpm report, set led(s) */
void compute_led_by_rpm (Driver *drvthis, unsigned char fan){
  PrivateData *p = drvthis->private_data;
  if ((p->model !=635) || (p->fans[fan].led <2) || (p->fans[fan].led ==3 )){ 
    return;
  } 
  static unsigned char slow[4]={9,9,9,9};
  if ( slow[fan]%9 !=0 ) { /* update every 9 reports */
    slow[fan]++;
    return;
  } 
  if (p->fans[fan].led==2) { /* is a 635 & this fan's led displays rpm */
    static float frepminmax[8]={INT_MAX,0-INT_MAX, INT_MAX,0-INT_MAX, INT_MAX,0-INT_MAX, INT_MAX,0-INT_MAX};
    static char previous[4]={0,0,0,0}; 
    /* dang sloppy wasteful init & checking */
    if ((frepminmax[fan]==INT_MAX)  && (p->fans[fan].rpm_rpt!=2)){ 
      frepminmax[fan]=frepminmax[fan+4]=p->fans[fan].rpm_rpt;
//      report(RPT_CRIT, "%s fan%d 1st rep value = %f", __FUNCTION__, fan+1, p->fans[fan].rpm_rpt);
    }
    else if ((frepminmax[fan]>p->fans[fan].rpm_rpt) && (p->fans[fan].rpm_rpt!=2)){
      /* ==2 means too low to get a report */
      frepminmax[fan]=p->fans[fan].rpm_rpt;
//      report(RPT_CRIT, "%s fan%d rep value low = % 3.2f  % 3.2f  range: % 3.2f", __FUNCTION__, fan+1, 
//          frepminmax[fan], frepminmax[fan+4], (frepminmax[fan+4] - frepminmax[fan]) );
    }
    else if (frepminmax[fan+4]<p->fans[fan].rpm_rpt){
      frepminmax[fan+4]=p->fans[fan].rpm_rpt;
//      report(RPT_CRIT, "%s fan%d rep valu high = % 3.2f  % 3.2f  range: % 3.2f", __FUNCTION__, fan+1, 
//          frepminmax[fan], frepminmax[fan+4], (frepminmax[fan+4] - frepminmax[fan]) );
    }
    char newrepval=(char)((p->fans[fan].rpm_rpt-frepminmax[fan]) / (frepminmax[fan+4]-frepminmax[fan]) *100);
    if (abs(previous[fan] - newrepval)> 100/CFLEDBASE) {
      set_led(drvthis, fan, newrepval);
//      report(RPT_CRIT, "%s fan%d led = %d", __FUNCTION__, fan+1, newrepval );
      previous[fan]=newrepval;
    }
  } /* end one led for fan  */
  else if (p->fans[fan].led==4) {
    static float frepminmax[2]={INT_MAX,0-INT_MAX};
    if ((frepminmax[0]==INT_MAX)  && (p->fans[fan].rpm_rpt!=2)){
      frepminmax[0]=frepminmax[1]=p->fans[fan].rpm_rpt;
//      report(RPT_CRIT, "%s all leds fan%d 1st rep value = % 3.2f", __FUNCTION__, fan+1, p->fans[fan].rpm_rpt);
    }
    else if ((frepminmax[0]>p->fans[fan].rpm_rpt) && (p->fans[fan].rpm_rpt!=2)){
      /* ==2 means too low to get a report */
      frepminmax[0]=p->fans[fan].rpm_rpt;
//      report(RPT_CRIT, "%s all leds fan%d rep value low = % 3.2f  % 3.2f  range: % 3.2f", __FUNCTION__, fan+1,
//        frepminmax[0], frepminmax[1], frepminmax[1]-frepminmax[0]);
    } 
    else if (frepminmax[1]<p->fans[fan].rpm_rpt){
      frepminmax[1]=p->fans[fan].rpm_rpt;
//      report(RPT_CRIT, "%s all leds fan%d rep valu high = % 3.2f  % 3.2f  range: % 3.2f", __FUNCTION__, fan+1,
//        frepminmax[0], frepminmax[1], frepminmax[1]-frepminmax[0]);
    }
    float todisplay=p->fans[fan].rpm_rpt-frepminmax[0]; /* value to display */
    float spread=frepminmax[1]-frepminmax[0];  /* current range to convert to 0-100 */
//    report(RPT_CRIT, "%s all spread=% 3.2f, this=% 3.2f or %0.2f %%", __FUNCTION__,
//      spread, todisplay, todisplay / spread * 100); 
    unsigned char tmp;
    for (tmp=0; tmp <4; tmp ++){
      static char prevrepval[4]={0,0,0,0};
      unsigned char newrepval;
      if (todisplay > spread / 4 ) { /* more than 25%, so this led is 100 */
        newrepval=100;
//        report(RPT_CRIT, "%s fan %d, pass %d % 3.2f (% 3d)", __FUNCTION__, fan+1, tmp, todisplay, newrepval);
        todisplay -= spread / 4;
      }
      else if (todisplay > 0) { /* not > 25 %, but some fraction for this led */
        /* last led to light, todisplay is between 0 - spead / 4 */
        newrepval=(unsigned char) (todisplay/spread*400);
//        report(RPT_CRIT, "%s fan %d, pass %d : % 3.2f (% 3d)", __FUNCTION__, fan+1, tmp, todisplay, newrepval);
        todisplay=0;            
      }
      else { /* nothing for this led to display */
        newrepval=0;
//        report(RPT_CRIT, "%s fan %d, pass %d : % 3.2f (% 3d)", __FUNCTION__, fan+1, tmp, todisplay, newrepval);
      }
      if (newrepval != prevrepval[tmp]) { /* do not resend identical to previous setting */
//        report(RPT_CRIT, "%s fan %d, pass %d : % 3.2f (% 3d)", __FUNCTION__, fan+1, tmp, todisplay, newrepval);
        set_led(drvthis, 3-tmp, newrepval);
        prevrepval[tmp]=newrepval;
      }
    } /* end each led loop */
  } /* end use all 4 leds for one datum report */
  return;
} /* end of compute_led_by_rpm */ 

/* given a fan number, set led randomly */
void led_rand (Driver *drvthis, unsigned char fan){
  PrivateData *p = drvthis->private_data;
  if ((p->model!=635) || (p->fans[fan].led!=3)){
    return;
  }
/* blinks faster with load
 update will never be triggered by a fan report
 or a fan power change
*/
  if (88 < (float)rand()/ RAND_MAX * 100){
    //report (RPT_CRIT, "%s fan %d blink return on no ULT 88 .", __FUNCTION__, fan+1); 
    return; 
  }
  COMMAND_PACKET led;
  led.command=CF633_Set_GPIO_Pin;
  led.data_length=2;
  led.data[0]=11-fan*2;
  led.data[1]=(char)(((float)RAND_MAX-(float)rand())*100/(float)RAND_MAX);
  //report(RPT_WARNING, "%s f%d_led %d=blink", __FUNCTION__, fan+1, led.data[1]);
  send_packet (drvthis, &led);
  led.data[0]=12-fan*2;
  led.data[1]=(char)(((float)RAND_MAX-(float)rand())*100/(float)RAND_MAX);
  send_packet (drvthis, &led);
  // report (RPT_CRIT, "%s fan %d blinked.", __FUNCTION__, fan+1); 
} /* end led_rand */

#endif /* CF635LEDS, large block of functions */

/*
 * Detect ds18 temperature sensors.
 * This run once at start detection routine could be romoved from driver
 * if driver simply took LCDd.conf Temperatures=x as accurate, and
 * handled potential innacuraccy adequately.
 * Write a separate utility to scan temperatures and report id's.
 *    No storing p->*romid* in running driver
 *    No await packet for romid probes
 *    does sending a s_able_m for nen-existant sensors return an error?
 * Would result in smaller resident CF.so, less mem use too :)
 * put cases for hw_querry & what else in sensors_packet ?
 * init packets now regular packets in que : get rid of await_packet entirely
 */
int 
ds_init (Driver* drvthis)
{
  PrivateData *p = drvthis->private_data;
  unsigned char retry=3;
  unsigned char this_retry=0;

  unsigned char sensors=0;
  unsigned char detected=0;
  unsigned char s_able_m[4]={0,0,0,0};
  unsigned char s_m_mult=1;
  struct timespec nod = {0, 003500000};
  p->ds18=malloc(sizeof( CF_DS18* ) *32);
  
  COMMAND_PACKET dow;
  dow.command=CF633_Read_DOW_Device_Information;
  dow.data_length=1;

  do {
    p->ds18[sensors]=malloc(sizeof( CF_DS18));
    p->ds18[sensors]->name=NULL;
    dow.data[0]=sensors;
//    dow.data_length=1; /* FIXME does this need repeating? off 2012 05 24 */
    send_packet(drvthis, &dow);
    // probe ds18 -> response, /100 on adding receipts
    if (sensors==0) {
      nod.tv_nsec =  3500000;
      nanosleep(&nod, NULL); // 1st ds18 probe respond delay
      //  more likely to detect all on 1st try with a pause ?
      // is the long pause for 1st response dependent on # of sensors?
      // nanosleep(5000000*p->tempcount) perhaps
      nod.tv_nsec = 10000;
    }
    report(RPT_DEBUG, 
        "%s sent[%d]", __FUNCTION__ , sensors);
//    while ((p->q_depth !=0) && (p->incoming_command.data[0] !=sensors )) 
// INTERESTING: why does the above slow things down so much ?
    while(p->q_depth !=0) 
      nanosleep(&nod, NULL); // ds_init 
    report(RPT_DEBUG, "%s wait got [%d]", __FUNCTION__, 
      p->incoming_command.data[0]);
    if (p->incoming_command.data[1]!=0) { //&& (sensors < p->tempcount)) {
      // collect ROM ID's of attatched sensors  
      int counter;
      p->ds18[sensors]->cfromid[0]=p->incoming_command.data[1];
      p->ds18[sensors]->romid[0]=p->incoming_command.data[1];
      // id[0] is sensor type, 0x28 =ds18b20
      /* I could not make sense of the CF ROM ID's
       * as returned by CF633_Read_DOW_Device_Information:
       * for id[1|2|3] the following does an 8-bit-byte 
       * endian reversal.  The 633 appears to order 
       * the sensors 0-31 by what is romid[1] here
       * ( then romid[2], then romid[3] I'm guessing ...
       * which may also be reversed  in order from
       * the manufacturer's 6,...3,2,1.  romids[4,5,6] are 
       * apparently as of yet unused ??? Q4 2004 
       * [7] is a CRC. */
      for (counter=2; counter <9; counter++) {
        p->ds18[sensors]->cfromid[counter-1]=
          p->incoming_command.data[counter];
        }
      for (counter=1; counter <8; counter++) { 
        p->ds18[sensors]->romid[counter] = 
          ((128&(p->ds18[sensors]->cfromid[counter]))>>7)+
          (( 64&(p->ds18[sensors]->cfromid[counter]))>>5)+
          (( 32&(p->ds18[sensors]->cfromid[counter]))>>3)+
          (( 16&(p->ds18[sensors]->cfromid[counter]))>>1)+
          (( 8& (p->ds18[sensors]->cfromid[counter]))<<1)+
          (( 4& (p->ds18[sensors]->cfromid[counter]))<<3)+
          (( 2& (p->ds18[sensors]->cfromid[counter]))<<5)+
          (( 1& (p->ds18[sensors]->cfromid[counter]))<<7);
        report(RPT_DEBUG, "%s p->ds18[%d]->romid[%d]=%02x",
         __FUNCTION__, sensors, counter, p->ds18[sensors]->romid[counter]);
      
      }
      if ( sensors  ==  p->tempcount ) 
	report(RPT_INFO, "%s More sensors than requested are avialable.");
      if (sensors < p->tempcount) {
	report(RPT_DEBUG, "%s s_enable_mask[%d] += %3d",
            __FUNCTION__, (int)(sensors/8),s_m_mult );
        s_able_m[(int)(sensors/8)]+=s_m_mult;
        if (s_m_mult == 128 ) {
          s_m_mult=1;
        }
        else {
          s_m_mult*=2;
        }
      } // end enable this requested sensor
      detected ++;
      sensors ++;
    } // endif do not enable unrequested sensors
    else {
      free(p->ds18[sensors]);
    }
/* when first p->incoming_command.data[1]==0 happens, there are no mare sensors
    else if ( (p->incoming_command.data[1]==0) && 
        (sensors >= p->tempcount)) {
      sensors ++;
    }
*/
    if ((p->incoming_command.data[1]==0) && 
        (p->tempcount > detected ) && 
        (this_retry < retry -1  ) ){
      report(RPT_INFO, 
        "%s detected %d of suggested %d Temperatures, attempt # %d, retry.",
        __FUNCTION__, detected, p->tempcount, this_retry );
      report(RPT_DEBUG, 
        "%s s_able_m built to [%d %d %d %d] ",
        __FUNCTION__, s_able_m[0],s_able_m[1],
        s_able_m[2],s_able_m[3] );
      for (detected=0; detected< sensors; detected ++) free(p->ds18[detected]);
      sensors=0;
      detected=0;
      s_able_m[0]=0;
      s_able_m[1]=0;
      s_able_m[2]=0;
      s_able_m[3]=0;
      report(RPT_DEBUG, "%s s_able_m reset to [%d %d %d %d] ",
        __FUNCTION__, s_able_m[0],s_able_m[1],
        s_able_m[2],s_able_m[3] );
      s_m_mult=1;
      this_retry++;
    }
    else if ( (p->incoming_command.data[1]==0)  && 
        ( this_retry == retry -1 )){
      report(RPT_INFO, 
        "%s detected %d of suggested %d Temperatures, attempt # %d, no retry.",
        __FUNCTION__, detected, p->tempcount, this_retry );
      this_retry++;
    }
    report(RPT_DEBUG, "%s this_retry %d sensors %d", 
      __FUNCTION__, this_retry, sensors);
  } while ((p->incoming_command.data[1]!=0) || ((this_retry < retry) && (sensors < p->tempcount)) );

  if (p->tempcount == detected) { 
    report(RPT_INFO, 
      "%s %d Tmp request; %d Tmp detect, attempt %d.",
      __FUNCTION__, p->tempcount, detected, this_retry);
  }
  else if (p->tempcount > detected) { 
    report(RPT_WARNING, 
      "%s %d temps suggested; %d detected.",
      __FUNCTION__, p->tempcount, detected );
    p->tempcount=detected;
  }
  else if (p->tempcount < detected) { 
    report(RPT_WARNING, "%s %d temps avail; cont with %d as requested.",
      __FUNCTION__,  detected, p->tempcount );
  }
  report(RPT_DEBUG, "%s s_able_m[%3d][%3d][%3d][%3d]", __FUNCTION__,
    s_able_m[0], s_able_m[1],
    s_able_m[2], s_able_m[3]);
  if (p->tempcount !=0) {
    char buf[255];
    // BUFSIZE is 255 from ../../shared/configfile.c
    char target[64];
    char named_by[10];
    report(RPT_CLICRE, "%s  #  romid  rom~     by     name",
      __FUNCTION__);
    for (sensors =0; sensors < p->tempcount; sensors++){ 
      /* set all sensor temp values to starting default*/ 
      p->ds18[sensors]->abstracto=DEFAULT_DS18_TEMP;
      /* try to read name by unique romid */
      sprintf(target, "%02x%02x%02x", p->ds18[sensors]->romid[1],
        p->ds18[sensors]->romid[2],p->ds18[sensors]->romid[3] );  
      snprintf(buf, sizeof(buf), 
        drvthis->config_get_string (drvthis->name,target, 0, target)
      );
      if ((strlen(buf)==0) || (*buf==*target)) {
        // empty or absent name by romid in config 
        sprintf(target, "t%02d_Name", sensors+1);  
          snprintf(buf, sizeof(buf), 
          drvthis->config_get_string
          (drvthis->name,target, 0, target));
        if ((strlen(buf)==0) || (*buf==*target)) {
          // empty or absent name by number from config
          sprintf (buf, DEFAULT_DS18_NAME);
          // add sensor number to default name
          snprintf (buf+ strlen(DEFAULT_DS18_NAME), 3, "%02d", sensors+1);
          snprintf (named_by, 10, "default "); 
        }
        else { /* did read a name by order number */
          snprintf (named_by, 10, "txx_Name"); 
        }
      } 
      else { /* did read a name by rom id */
          snprintf (named_by, 10, "romid   "); 
      }
      p->ds18[sensors]->name = strdup(buf) ; 
      p->ds18[sensors]->data[0]=p->ds18[sensors]->data[1]=0;
      p->ds18[sensors]->data[2]=1;
      report(RPT_CLICRE,
       "%s T%02d %02x%02x%02x %02x%02x%02x %s % 12s ",
        __FUNCTION__, sensors + 1,  
        p->ds18[sensors]->romid[1],  p->ds18[sensors]->romid[2], 
        p->ds18[sensors]->romid[3],  
        p->ds18[sensors]->cfromid[1], p->ds18[sensors]->cfromid[2],
        p->ds18[sensors]->cfromid[3], named_by, 
        p->ds18[sensors]->name);
    } // end of one sensor
    set_temp_rpt(drvthis, s_able_m);
  } // end p->tempcount !=0
  //p->ds18=realloc(p->ds18, sizeof( CF_DS18* ) * p->tempcount );
  return 1;
} // end ds_init

/* 
 * parse fan configs, verify, fill *p->fans
 */
int fan_init (Driver* drvthis){
  PrivateData *p = drvthis->private_data;  
  char target[64];
  int fans=0;
  int tmp;
#ifdef FANS_USEC
  gettimeofday(&p->tdid_power, NULL);
  p->tdid_power.tv_sec-=32;
  p->treq_power=p->tdid_power;
#endif // FANS_USEC
  char buf[255]; // BUFSIZE is 255 from ../../shared/configfile.ci
#ifdef CF635LEDS
  if (p->model==635) {
    p->fans[0].led=p->fans[1].led=p->fans[2].led=p->fans[3].led=0;
    sprintf (target, "leds");  
    sprintf(buf, drvthis->config_get_string (drvthis->name, target, 0, NULL));
    if ((strlen(buf)==3) && (buf[0]=='f') && (tmp=(int)strtol(buf+1, NULL, 10)) && ((tmp>0) && (tmp<5)) ) {
          if (buf[2]=='r') p->fans[tmp-1].led=4;
          else if (buf[2]=='p') p->fans[tmp-1].led=5;
          report(RPT_CLICRE, "%s read %s=%s. fan%d uses all leds for %s value",
            __FUNCTION__, target, buf, tmp, (p->fans[tmp-1].led==4 ? "report" : "power") );
    } /* end read something for leds */
    else { /* did not read leds= entry */
      for (fans =0; fans <  4; fans++){
        sprintf (target, "f%d_led", fans+1);
        sprintf(buf, drvthis->config_get_string (drvthis->name, target, 0, "pow"));
        switch (buf[0]){
          case 'o' :
            p->fans[fans].led=0;  /* off / nothing */
            break;
          case 'p': /* show power */
            p->fans[fans].led=1; 
            break;
          case 'r': /* show report */
            p->fans[fans].led=2;
            break;
          case 'b': /* blink randomly */
            p->fans[fans].led=3;
            break;
          default:
            p->fans[fans].led=0;
            report(RPT_WARNING, "%s f%d_led=off; on unknown value %s.", __FUNCTION__,
            fans+1, buf );
        } 
        report(RPT_CLICRE, "%s read %s=%s.",
          __FUNCTION__, target, buf);
      }/* for each fan */ 
    } /* end did not read leds= entry */ 
    
  } /* end it is a 635 */
#endif // CF635LEDS

  for (fans =0; fans <  4; fans++){
#ifdef FANS_USEC
      p->fans[fans].tdid_fan=p->treq_power;
#endif // FANS_USEC
      p->fans[fans].name=NULL;
      p->fans[fans].d_label=NULL;
      p->fans[fans].data[0]=p->fans[fans].data[1]=0;
      p->fans[fans].rpm_rpt=0;
      char fan_safety_checks=1;
      sprintf (target, "f%d_Name", fans+1);  
      sprintf(buf, drvthis->config_get_string (drvthis->name, target, 0, target));
      if ((strlen(buf)==0) || (*buf==*target) ){  // empty or absent f#_Name line in config
        sprintf(buf, DEFAULT_FAN_NAME);
        snprintf (buf+ strlen(DEFAULT_FAN_NAME), 2, "%d", fans+1); // add fan number to default name
      }
      p->fans[fans].name =(char *)malloc (strlen (buf)+1);
      snprintf(p->fans[fans].name, strlen (buf)+1, buf);
      if (strncmp(buf,"^",1)==0) {
        report(RPT_WARNING, "%s f%d safety checking disabled.",__FUNCTION__, fans+1);
        fan_safety_checks=0;
      }

      sprintf (target, "f%d_D_label", fans+1);  
      sprintf(buf, drvthis->config_get_string (drvthis->name, target, 0, DEFAULT_FAN_D_LABEL));
      if (strlen(buf)==0)  // empty f#_Name line in config
        sprintf(buf, DEFAULT_FAN_D_LABEL);
      p->fans[fans].d_label =(char *)malloc (strlen (buf)+1);
      sprintf(p->fans[fans].d_label, buf);
                          
      sprintf (target, "f%d_Rpt", fans+1);  
      p->fans[fans].autorep= drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_AUTOREPORT);

      sprintf (target, "f%d_Power", fans+1);  
      tmp=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_POWER);
      if ((tmp < 0) || (tmp > 100)) {
        report(RPT_WARNING, "%s f%d_Power=%d invalid, set to 100.", __FUNCTION__, fans+1, tmp);
        tmp=100;
      }
      p->fans[fans].power=tmp;
      p->priority_packet[0].data[fans]=tmp;
      sprintf (target, "f%d_P_Log", fans+1);  
      p->fans[fans].p_log=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_LOG_POW);

      sprintf (target, "f%d_Divisor", fans+1);  
      tmp=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_DIVISOR);
      if (tmp == 0)  {
        report(RPT_WARNING, "%s f%d_Divisor=%d invalid, set to %d.", __FUNCTION__,
            fans+1, tmp, DEFAULT_FAN_DIVISOR);
        tmp=DEFAULT_FAN_DIVISOR;
      }
      p->fans[fans].divisor=tmp;

      sprintf (target, "f%d_Failsafe", fans+1);  
      p->fans[fans].failsafe=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_FAILSAFE);
/*    
      sprintf (target, "f%d_Glitch", fans+1);  
      p->fans[fans].glitch=drvthis->config_get_bool(drvthis->name, target, 0, DEFAULT_FAN_GLITCH);
*/      
      p->fans[fans].f_sens[0]=(FAN_SENSOR *)malloc(sizeof(FAN_SENSOR)); 
      p->fans[fans].f_sens[1]=NULL; 
      p->fans[fans].f_sens[0]->abstvalptr=NULL;
      p->fans[fans].f_sens[0]->name=NULL; 
      fan_get_temp(drvthis, fans, 0, target, buf);
      if (p->fans[fans].f_sens[0]->name==NULL) {
        report (RPT_CLICRE, "%s did not read a f%d_Temp by temperature name or number.", __FUNCTION__, fans+1 );
        free(p->fans[fans].f_sens[0]);
        p->fans[fans].f_sens[0]=NULL; 
      } else { 
        p->fans[fans].f_sens[1]=(FAN_SENSOR *)malloc(sizeof(FAN_SENSOR)); 
        p->fans[fans].f_sens[1]->abstvalptr=NULL;
        p->fans[fans].f_sens[1]->name=NULL; 
//        report (RPT_CRIT, "%s fans=%d",  __FUNCTION__, fans);
        fan_get_temp(drvthis, fans, 1, target, buf);
//      report (RPT_CRIT, "%s fans=%d on return",  __FUNCTION__, fans);
        if (p->fans[fans].f_sens[1]->name==NULL) {
          report (RPT_CLICRE, "%s did not read a f%d_Ref_Temp by temperature name or number.", __FUNCTION__, fans+1 );
          free(p->fans[fans].f_sens[1]);
          p->fans[fans].f_sens[1]=NULL; 
        }  
      }
      if (p->fans[fans].f_sens[1]!=NULL){
        sprintf (target, "f%d_DT_Label", fans+1);  
        sprintf(buf, drvthis->config_get_string (drvthis->name, target, 0, "delta_t"));
        if (strlen(buf)!=0){  // read dt_label in config
          p->fans[fans].dt_label =(char *)malloc (strlen (buf)+1);
          sprintf(p->fans[fans].dt_label, buf);
        } 
      }

      sprintf (target, "f%d_Heat", fans+1);  
      p->fans[fans].heat=drvthis->config_get_bool(drvthis->name, target, 0, 0);

      //if (( p->fans[fans].f_sens[0]->name!=NULL)&&(fan_safety_checks)) { 
      if (( p->fans[fans].f_sens[0]!=NULL)&&(fan_safety_checks)) { 
        if  (p->fans[fans].p_log==0){
          p->fans[fans].p_log=1;
          report(RPT_INFO, "%s f%d_Temp=%s causes f%d_Plog=yes for safety.",
            __FUNCTION__, fans+1,  p->fans[fans].f_sens[0]->name, fans+1);
        }
        if  (p->fans[fans].autorep==0){
          p->fans[fans].autorep=1;
          report(RPT_INFO, "%s f%d_Temp=%s causes f%d_Rpt=yes for safety.",
              __FUNCTION__, fans+1, p->fans[fans].f_sens[0]->name, fans+1);
        }
        if  (p->fans[fans].failsafe==0){ 
          p->fans[fans].failsafe=1;
          report(RPT_INFO, "%s f%d_Temp=%s causes f%d_Failsafe=yes for safety.",
              __FUNCTION__, fans+1, p->fans[fans].f_sens[0]->name, fans+1);
        }
      } // end of simple checks for turning on sensor driven fan speed  
      p->fans[fans].autorep_mask=p->fans[fans].autorep;

      //if ( p->fans[fans].f_sens[0]->name!=NULL) {  
      if ( p->fans[fans].f_sens[0]!=NULL) {  
        sprintf (target, "f%d_T_off", fans+1);  
        p->fans[fans].by_s_off=16*drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_BY_S_OFF);
        if (( p->fans[fans].by_s_off > 16*DEFAULT_FAN_BY_S_OFF+16*DEFAULT_FAN_SAFETY) && (fan_safety_checks)) {
          report(RPT_INFO, "%s f%d_T_off=%d -> %d for safety.",
            __FUNCTION__, fans+1,  p->fans[fans].by_s_off/16, DEFAULT_FAN_BY_S_OFF);
          p->fans[fans].by_s_off=DEFAULT_FAN_BY_S_OFF*16;
        }

        sprintf (target, "f%d_T_100", fans+1);  
        p->fans[fans].by_s_100=16*drvthis->config_get_int(drvthis->name, target, 0, 16*DEFAULT_FAN_BY_S_100);
        if ((p->fans[fans].by_s_100 > 16*DEFAULT_FAN_BY_S_100+16*DEFAULT_FAN_SAFETY)&&(fan_safety_checks)) {
          report(RPT_INFO, "%s f%d_T_100=%d -> %d for safety.", __FUNCTION__, 
            fans+1, p->fans[fans].by_s_100/16, DEFAULT_FAN_BY_S_100);
            p->fans[fans].by_s_100 =DEFAULT_FAN_BY_S_100*16;
        }
        if ((p->fans[fans].by_s_100 <= p->fans[fans].by_s_off)
            &&(fan_safety_checks)) {
          report(RPT_INFO, "%s f%d_T_100=%d is less than f%d_T_off=%d.  Is this intended?",
            __FUNCTION__, fans+1, p->fans[fans].by_s_100/16, fans+1, 
            p->fans[fans].by_s_off/16);
        }
        sprintf (target, "f%d_A_Sec", fans+1);  
        p->fans[fans].a_sec=drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_A_SEC);
        if ( ((p->fans[fans].a_sec > 31) || (p->fans[fans].a_sec < 1)) &&(fan_safety_checks) ) {
          report(RPT_INFO, "%s f%d_A_Sec=%d  not 1-31, reset to default %d.",
            __FUNCTION__, fans+1, p->fans[fans].a_sec, DEFAULT_FAN_A_SEC);
          p->fans[fans].a_sec=DEFAULT_FAN_A_SEC;
        }
      } // end secondary temp adjust data
      else {
        p->fans[fans].by_s_off=DEFAULT_FAN_BY_S_OFF*16;
        p->fans[fans].by_s_100 =DEFAULT_FAN_BY_S_100*16;
        p->fans[fans].a_sec=DEFAULT_FAN_A_SEC;
      }
      report(RPT_CLICRE, 
        "%s f%d, %s, u %s, rpt %d, pow %d, p_log %d, div %d, fs %d, T %s, TA %s, off %.4f, full %.4f, a_sec %d, dt_l %s%s", 
            __FUNCTION__, fans+1, p->fans[fans].name,  p->fans[fans].d_label, 
            p->fans[fans].autorep, p->fans[fans].power, p->fans[fans].p_log,
            p->fans[fans].divisor, p->fans[fans].failsafe, 
            (p->fans[fans].f_sens[0] !=NULL ) ? p->fans[fans].f_sens[0]->name : NULL,
            (p->fans[fans].f_sens[1] !=NULL ) ? p->fans[fans].f_sens[1]->name : NULL,
            (p->fans[fans].f_sens[0] !=NULL ) ? (float)p->fans[fans].by_s_off/16 : 0, 
            //(p->fans[fans].f_sens[0]->name !=NULL ) ? (float)p->fans[fans].by_s_off : 0, 
            (p->fans[fans].f_sens[0] !=NULL ) ? (float)p->fans[fans].by_s_100 /16: 0,
            //(p->fans[fans].f_sens[0]->name !=NULL ) ? (float)p->fans[fans].by_s_100 : 0,
            p->fans[fans].a_sec, p->fans[fans].dt_label, (p->fans[fans].heat==0) ?"":", heat")  ;
            // could put in #ifdef CF635LEDS for format specifier line and arg line?  
  } // end for each fan
  /* fsg used only in init */
  set_fan_fsg (drvthis); 
  set_fan_pow(drvthis);
  set_fan_rpt(drvthis);
  return 1;
} //end fan_init

/*
 * read temps that determine fan speed
 */
int 
fan_get_temp (Driver* drvthis, unsigned char fan, unsigned char type, char *target, char *buf){
  PrivateData *p = drvthis->private_data;  
  int tmp;
    /* try to read a temp by temp name */
      if (type==0) { 
        sprintf (target, "f%d_Temp", fan+1 );  
      } else
      {
        sprintf (target, "f%d_Ref_Temp", fan+1 );  
      } 
      sprintf(buf, drvthis->config_get_string (drvthis->name, target, 0, "" ));
      if (strlen(buf) !=0 ) {
        for (tmp=0; tmp < p->tempcount; tmp++){
          if(0 == strncmp(buf, p->ds18[tmp]->name,  strlen(buf) )) { 
            p->fans[fan].f_sens[type]->number = tmp;
            p->fans[fan].f_sens[type]->abstvalptr = &p->ds18[tmp]->abstracto;
            p->fans[fan].f_sens[type]->name = p->ds18[tmp]->name;
            p->fans[fan].f_sens[type]->updates=( void*) *compute_temp;
            report (RPT_CLICRE, "%s by temperature name  f%d_Temp=  %s", __FUNCTION__, 
              fan+1, p->fans[fan].f_sens[type]->name);
          }
          if (p->fans[fan].f_sens[type]->name!= NULL) return (1);
        }
#ifdef DO_LM
        for (tmp=0; tmp < p->lmnum; tmp++){
          if(0 == strncmp(buf, p->lm[tmp]->lmlabel,  strlen(buf) < strlen (p->lm[tmp]->lmlabel) ? strlen(buf) : strlen (p->lm[tmp]->lmlabel) )) { 
            p->fans[fan].f_sens[type]->number = tmp;
            p->fans[fan].f_sens[type]->abstvalptr = &p->lm[tmp]->lm_abstracto;
            p->fans[fan].f_sens[type]->name = p->lm[tmp]->lmlabel;
            p->fans[fan].f_sens[type]->updates=*update_lm;
            report (RPT_CLICRE, "%s by lm-sensors label  f%d_Temp=  %s", 
              __FUNCTION__, fan+1, p->fans[fan].f_sens[type]->name);
          }
          if (p->fans[fan].f_sens[type]->name!= NULL) return (1);
        }
#endif // DO_LM
#ifdef DO_HDD
        for (tmp=0; tmp < p->hdd_count; tmp++){
          if(0 == strncmp(buf, p->hdd[tmp]->hddname,  strlen(buf) )) { 
            p->fans[fan].f_sens[type]->number = tmp;
            p->fans[fan].f_sens[type]->abstvalptr = &p->hdd[tmp]->hd_abstracto;
            p->fans[fan].f_sens[type]->name = p->hdd[tmp]->hddname;
            p->fans[fan].f_sens[type]->updates=*hddtemp_temps;
            report (RPT_CLICRE, "%s by hdd name   f%d_Temp=  %s", 
              __FUNCTION__, fan+1, p->fans[fan].f_sens[type]->name);
          }
          if (p->fans[fan].f_sens[type]->name!= NULL) return (1);
        }
#endif // DO_HDD
      } 
      if (p->fans[fan].f_sens[type]->name==NULL) {
        /* try to read a temp by temp number */
        tmp = drvthis->config_get_int(drvthis->name, target, 0, DEFAULT_FAN_SENS_NUM);
        if ((tmp != DEFAULT_FAN_SENS_NUM) && ( tmp!= -1) && (tmp < p->tempcount) ) {
          p->fans[fan].f_sens[type]->number = tmp;
          p->fans[fan].f_sens[type]->abstvalptr = &p->ds18[tmp]->abstracto;
          p->fans[fan].f_sens[type]->name = p->ds18[tmp-1]->name; /* fix: added -1 20081209 */
          p->fans[fan].f_sens[type]->updates=NULL;
          report (RPT_CLICRE, "%s by temperature number  f%d_Temp=  %s", 
            __FUNCTION__, fan+1, p->fans[fan].f_sens[type]->name);
        }
      }
  return (1);
} //end fan_get_temp

/*
 * Set fan failsafes to values in *p->fans.
 */
void  
set_fan_fsg ( Driver* drvthis)  
{
  PrivateData *p = drvthis->private_data;  
  unsigned char fans;
  COMMAND_PACKET fan_packet;
  unsigned char bitmask;

  // set failsafe first seems safe
  fan_packet.command=CF633_Set_Fan_Power_Failsafe;   
  fan_packet.data_length=2;   
  fan_packet.data[0]=0;   
  fan_packet.data[1]=255; // 31 7/8 seconds failsafe reset
  bitmask=1;
  for (fans =0; fans <4 ;fans ++){ 
    if (p->fans[fans].failsafe==1) 
      fan_packet.data[0]+=bitmask;
    bitmask=bitmask*2;  
  }  
  debug(RPT_FANS, "%s failsafe bitmask=%d, %d.", __FUNCTION__,
      fan_packet.data[0], fan_packet.data[1] );
  send_packet(drvthis, &fan_packet);
} // end of set_fan_fsg

/*
 * Set the rpting of temperatures.
 */
static void 
set_temp_rpt(Driver *drvthis, unsigned char out[4])
{
  COMMAND_PACKET temp_rep;
  temp_rep.command=CF633_Set_Up_Temperature_Reporting;
  temp_rep.data_length=4;
  temp_rep.data[0]=out[0];
  temp_rep.data[1]=out[1];
  temp_rep.data[2]=out[2];
  temp_rep.data[3]=out[3];
  send_packet (drvthis, &temp_rep);
}  

/* 
 *  show a line of syslog column names 
 */
void csv_header (Driver *drvthis) {
  PrivateData *p = drvthis->private_data;  
  char *syslogline;
  int tmp;
  int lineused=0;
  strncpy(p->logfile, drvthis->config_get_string(drvthis->name, "Logfile", 0, ""), sizeof(p->logfile) );
  if (0==strncmp(p->logfile,"", 1)) {
    report(RPT_INFO, "%s Did not read an output Logfile.",  __FUNCTION__);
    p->log_sens=0;
    return;
  }
  else {
    // more file open error checking
    // LCDd is not a group or user,
    // and this does not need to be root eh ?
    // utmp is done by for logging ? convention ?
    // how to not downgrade other threads ?
/*
    uid_t me = getuid();
    struct passwd * mep;
    mep = getpwuid(me);
    report(RPT_CRIT, "%s Read write to logfile %s as uid  %s",
      __FUNCTION__, p->logfile, mep->pw_name );
    me = geteuid();
    mep = getpwuid(me);
    report(RPT_CRIT, "%s Read write to logfile %s as euid %s",
      __FUNCTION__, p->logfile, mep->pw_name );
*/
    report(RPT_INFO, "%s Will use data Logfile %s",  __FUNCTION__, p->logfile);
    p->log_fd=fopen(p->logfile, "a");
    tmp=errno;
    if (p->log_fd == NULL){
      report(RPT_ERR, "%s Could not open logfile %s, %s.",
           __FUNCTION__, p->logfile, strerror(tmp));
      p->log_sens=0;
      return;
      // & deconfigure / free more ?
    }
  }
  syslogline = calloc (256,1);
  char *buf=calloc(256,1);
  time_t very_now=time(NULL);
  ctime_r(&very_now, buf);
  buf[strlen(buf)-1] ='\0';
  lineused += snprintf (syslogline, 255, buf);
#ifdef DO_ULT  
  if (p->log_uptime!=0) {
    lineused +=snprintf (&syslogline[lineused], 255, ",uptime");
  }
  if (p->log_load!=0) {
    lineused +=snprintf (&syslogline[lineused], 255, ",load");
  }
  if (p->log_load_s!=0) {
    lineused +=snprintf (&syslogline[lineused], 255, ",sload");
  }
  if (p->log_load_u!=0) {
    lineused +=snprintf (&syslogline[lineused], 255, ",uload");
  }
  for (tmp=0; tmp <p->topnum; tmp++) {
/*
  note to print a % into string use %%,
  to print a % into a stritg that will be printed into a string, use %%%%
*/
    syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
        ",top%d,%%%%", tmp+1) );
    lineused += snprintf (&syslogline[lineused], 255, buf);
  }
  for (tmp=0; tmp < p->topname; tmp++){
    //if (p->top_name[tmp]->name!=NULL) { 
    // fixed :sorted to all non-null in ult_init
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%s%%%%", p->top_name[tmp]->name));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    //}
  }// end of each possible topname
#endif // DO_ULT
#ifdef DO_LM
  for (tmp=0; tmp < p->lmnum; tmp++){ 
    syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
        ",%s:", p->lm[tmp]->chipname));
    lineused += snprintf (&syslogline[lineused], 255, buf);
    syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
        "%s",  p->lm[tmp]->lmlabel));
    lineused += snprintf (&syslogline[lineused], 255, buf);
  }
#endif
#ifdef DO_HDD  
  if (p->hdd_count !=0) {
    for (tmp=0; tmp  < p->hdd_count ; tmp ++){  // for each hdd
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%s", p->hdd[tmp]->hddname));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    }// end each drive
  }
#endif // DO_HDD  
  for (tmp=0; tmp <4; tmp++) {
    if (p->fans[tmp].p_log) {
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%s:pwr", p->fans[tmp].name));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    }
    if (p->fans[tmp].autorep) {
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%s:%s", p->fans[tmp].name,p->fans[tmp].d_label));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    }
    if (p->fans[tmp].dt_label != NULL ){
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",%s", p->fans[tmp].dt_label));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    }
  } // end of build syslogline for each fan
  if (p->tempcount > 0) {
    for (tmp=0; tmp < p->tempcount; tmp++) {
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
            ",%s", p->ds18[tmp]->name));
      lineused += snprintf (&syslogline[lineused], 255, buf);
    } // end of build syslogline for each Temperature  
  } // end of build syslogline for all Temperatures
#ifdef MY_CUSTOM_LINE
      syslogline=realloc(syslogline, 1+ strlen(syslogline) + snprintf (buf, 255, 
          ",watts"));
      lineused += snprintf (&syslogline[lineused], 255, buf);
#endif
  //report(RPT_CRIT, "%s CF633data%s", __FUNCTION__, syslogline);
  fprintf(p->log_fd, "%s \n", syslogline);
  char *start_field=syslogline;
  /* print a grep helper line */
  char *comma=index(syslogline, ',');
  int column=1;
  do {
    int written=snprintf(start_field, comma-start_field, "%d", column);
    //int written=snprintf(start_field, comma-start_field -1, "%d", column);
    if ( written > comma-start_field -1) written =0;
    for (; start_field + written < comma ; written++){
      start_field[written]=' ';
    }
    column++;
    start_field=comma+1;
    comma++;
    comma=index(comma, ',');
  } while (comma!=NULL) ;
  snprintf(start_field, &syslogline[lineused]-start_field, "%d", column);
  fprintf(p->log_fd, "%s \n", syslogline);
  fflush(p->log_fd);
  free (syslogline);
  free (buf);
} // end of csv_header


/* 
 * there is a built in way to compare two structures, yes?
 * but if data_length < MAX_DATA_LENGTH then
 * unused parts of PACKETs are compared by builtin ==
 * return 1 = packets differ
 * return -1 = total equality
*/
static int packet_neq
  (COMMAND_PACKET *packet1, COMMAND_PACKET *packet2) {
  int counter;
/*  if (packet1->CRC.as_word != packet2->CRC.as_word) return 1;  
  if (packet1->CRC.as_bytes[0] != packet2->CRC.as_bytes[0]) return 1;  
  if (packet1->CRC.as_bytes[1] != packet2->CRC.as_bytes[1]) return 1;  
*if the client / server bit of ->command is different, CRC different
*/
  if ((0x3F&(packet1->command)) != (0x3F&(packet2->command))) return 1;  
  if (packet1->data_length != packet2->data_length) return 1;  
  for (counter=0; counter < packet1->data_length; counter++)
    if (packet1->data[counter] != packet2->data[counter]) return 1;  
   return -1;
}


/*
 * Listen for packets 
 */
int 
listen_packets(Driver * drvthis)
{
  PrivateData * p = drvthis->private_data;
  struct timeval now;
  int is_msg;
  struct timespec nod = {0, 002000000};
#ifdef __linux__
  debug(RPT_INFO, "pthread_self % 15s = %d tid = %d",
    __FUNCTION__, pthread_self(), syscall(SYS_gettid));
#endif /* ifdef linux */
  while (1) {
    is_msg = check_for_packet(drvthis, MAX_DATA_LENGTH);
    // stuff is incoming ... no elaborate waiting 
    if (is_msg == GOOD_MSG ){
#ifdef TIMING_REPORT
      p->msg_type_count[GOOD_MSG]++;
#endif
      sensors_packet(drvthis);
/*                999999999 = 1 sec
 *              0.017       = average ack for 635 today */  
      nod.tv_nsec=000050000; nanosleep(&nod, NULL); // GOOD_MSG packet wait
      continue;
    } // end is GOOD_MSG
    else if (is_msg == TRY_AGAIN ) {
#ifdef TIMING_REPORT
      p->msg_type_count[TRY_AGAIN]++;
#endif
      // report(RPT_CRIT, "%s TRY_AGAIN",__FUNCTION__);
      nod.tv_nsec=000001000; nanosleep(&nod, NULL); // TRY_AGAIN packet wait
      continue;
    } /* end TRY AGAIN */
    else if (p->q_depth >0) { /* must be GIVE_UP */
#ifdef TIMING_REPORT
      p->msg_type_count[GIVE_UP]++;
#endif
      // and run these checks only if expecting something
      int cmd=(0x3F&(p->sendq[p->sending].command));
      gettimeofday (&now, NULL);
      if ( (cmd != CF633_Dummy_Unused) && 
        //by the pdf is 750000 ? for all CFx 250000 for CF635 ( .. on usb 2.0, not 1.0 ) ?
#ifdef TIMING_REPORT
          ( time_diff(drvthis, p->tsent, now) > p->ack_d[cmd][3])
#else
          ( time_diff(drvthis, p->tsent, now) > p->resend_timeout)
#endif
         ) 
      {
        pthread_mutex_lock(&quemutex); /* start resending */ 
        static COMMAND_PACKET last_repeat; /* initialize me FIXME */
        if (0 != memcmp(&p->sendq[p->sending].command, &last_repeat.command, p->sendq[p->sending].data_length+2)) {
          memcpy(&last_repeat.command, &p->sendq[p->sending].command, p->sendq[p->sending].data_length+2);
          p->resend_repeats=0;
        } // end this is resending a new packet
        p->q_resent=p->sending;
        p->resend_repeats++;
        commit_packet(drvthis, &p->sendq[p->sending]);
        gettimeofday (&p->tsent, NULL);
#ifdef DEBUG
        debug(RPT_QUE, "%s RESEND %02d q %02x %32s             %02x (%03d) = %.3f",
          __FUNCTION__, p->resend_repeats, p->sending,
	  p->pkn[p->sendq[p->sending].command], p->endq, p->q_depth, 
#ifdef TIMING_REPORT
          (float)p->ack_d[cmd][3]/1000000);
#else
          (float)p->resend_timeout/1000000);
#endif
        //show_packet(drvthis, RPT_CRIT, &p->sendq[p->sending]);
#endif // debug
#ifdef TIMING_REPORT
        p->ack_c[p->sendq[p->sending].command][2]++;
#endif
        if (p->resend_repeats == 5) {
/* suspect: sending a packet identical to the previous ( maybe of the same type ) packet sent
 * makes the module not send an ack.  So track the last good ack'ed packet
 * and see if it is identical to the one now being resent 
May 25 18:25:30 furniture LCDd: 
16261 good,     0 purged, 0.000 % overload.     9.033 packets/sec, elapse 1800.090
Packet (top 00, cmd 09 Set_LCD_Special_Character_Data ) len 09 [00 1f 15 00 00 00 11 1b 1f ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [08 10 ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [07 00 ]
listen_packets(CFA635:h1.0,v1.3 on /dev/ttyUSB0) resend_repeats==5 -> next
Packet (top 00, cmd 09 Set_LCD_Special_Character_Data ) len 09 [00 1f 15 00 00 00 11 1b 1f ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [06 30 ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [06 50 ]
listen_packets(CFA635:h1.0,v1.3 on /dev/ttyUSB0) resend_repeats==5 -> next
Packet (top 00, cmd 09 Set_LCD_Special_Character_Data ) len 09 [00 1f 15 00 00 00 11 1b 1f ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [07 00 ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [05 24 ]
listen_packets(CFA635:h1.0,v1.3 on /dev/ttyUSB0) resend_repeats==5 -> next
Packet (top 00, cmd 09 Set_LCD_Special_Character_Data ) len 09 [00 1f 15 00 00 00 11 1b 1f ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [06 30 ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [06 40 ]
listen_packets(CFA635:h1.0,v1.3 on /dev/ttyUSB0) resend_repeats==5 -> next
purge_que PURGING QUE  30 30 (215) send_packet, start purge 34 -> 30
purge_que 30 30 (215) >  30 34 (004)  252 p,  17.9% S_Ch_D, /dev/ttyUSB0 by send_packet 10987 g, 2.3% overload
purge_que PURGING QUE  36 38 (215) send_packet, start purge 3a -> 38
purge_que 36 38 (215) >  36 3a (004)  254 p,  7.1% S_Ch_D, /dev/ttyUSB0 by send_packet   408 g, 62.3% overload
Packet (top 00, cmd 31 Send_Data_to_LCD ) len 05 [11 02 36 32 35 ]
     sData_to_LCD     03 at 17 x 2 '625'
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [08 20 ]
Packet (top 00, cmd 34 Set_GPIO_Pin ) len 02 [07 24 ]
listen_packets(CFA635:h1.0,v1.3 on /dev/ttyUSB0) resend_repeats==5 -> next
*/
#ifdef DEBUG
#ifdef TIMING_REPORT
          show_packet(drvthis, RPT_CRIT, &p->last_good_ack);
          show_packet(drvthis, RPT_CRIT, &p->l_g_a[p->sendq[p->sending].command]);
#endif
          show_packet(drvthis, RPT_CRIT, &p->sendq[p->sending]);
#endif // debug
          report(RPT_CRIT, "%s(%s on %s) resend_repeats==5 -> next",
            __FUNCTION__, p->model_str, p->device);
//          CF63oddtf_close(drvthis);
          p->sendq[p->sending].command=CF633_Dummy_Unused;
          p->sending++;
          p->q_depth--;
          p->resend_repeats=0;
        } // end resend repeats == 5
        pthread_mutex_unlock(&quemutex); /* start resending */ 
      } // end check for lost packet
#ifdef DEBUG
      //debug(RPT_QUE, "%s GIVE_UP %02d p->sending=%03d qd=%03d %s",
      //  __FUNCTION__, p->resend_repeats, p->sending, p->q_depth,
      //   p->pkn[p->sendq[p->sending].command]);
#endif
    } // end GIVE UP and que depth > 0 
// grep "listen_packets loop wait:" /var/log/syslog.1 | cut -f5 -d : | cut -f4 -d' ' | grep -v "100." | sort | more

/* below here, must've been GIVE_UP */
    //sel_wait(drvthis);
    // trial and error notes with GOOD sleep at 1000
    //nod.tv_nsec=    10000; 
    //9075 GOOD       0 TRY_AGAIN   12188 GIVE_UP
    //1289         0.009983  0.037  0.022             0     1     1    All

    //nod.tv_nsec=    50000;
    //4142 GOOD       0 TRY_AGAIN    5497 GIVE_UP
    //614         0.007993  0.038  0.021             0     0     0    All    

    //nod.tv_nsec=   100000;
    //3797 GOOD       0 TRY_AGAIN    4774 GIVE_UP
    //539         0.010979  0.036  0.021             0     0     0    All
    
    //nod.tv_nsec=   500000; 
    //5176 GOOD       0 TRY_AGAIN    5881 GIVE_UP
    //56         0.009806  0.039  0.022             0     0     0    All

    //nod.tv_nsec=  1000000; 
    //3435 GOOD       0 TRY_AGAIN    3121 GIVE_UP
    //501         0.009988  0.040  0.022             0     1     1    All
    
    //nod.tv_nsec=  1500000; 
    //888 GOOD       0 TRY_AGAIN     724 GIVE_UP
    //136         0.009792  0.036  0.022             0     0     0    All
    
    //nod.tv_nsec=  5000000; 
    //1235 GOOD       0 TRY_AGAIN     540 GIVE_UP
    //194         0.008993  0.041  0.024             0     0     0    All

    //nod.tv_nsec= 10000000; 
    //1871 GOOD       0 TRY_AGAIN     445 GIVE_UP
    //282         0.013812  0.043  0.026             0     0     0    All
/* 200812 25 =50000, lenny, new kernel, LCDd is 50%+ cpu!
       1:41 ratio 52109 GOOD       0 TRY_AGAIN  2147763 GIVE_UP
       1:40 ratio 34459 GOOD       0 TRY_AGAIN  1400924 GIVE_UP
       =5000000 4-5 % cpu
       6140 GOOD       0 TRY_AGAIN    3294 GIVE_UP
       =10000000 1%
       3953 GOOD       0 TRY_AGAIN    1112 GIVE_UP
       speculate ? related to choice of 1000 HZ (HZ_1000) in new lenny kernel ?
       nanosleep should be independent of Processor Type and Features / Timer frequency ?
    */

    // nod.tv_nsec=    50000; nanosleep(&nod, NULL); // GIVE UP packet wait
    int errno_sav=0;
/*                 999999999 = 1 sec
 *               0.017       = average ack for 635 today */  
    nod.tv_nsec=    20000000; if (nanosleep(&nod, NULL)) { // GIVE UP packet wait
      errno_sav=errno;
      report(RPT_CRIT, "%s nanosleep interrupted (%s)",
        __FUNCTION__, strerror (errno_sav));

    } // end was interrupted

  } // end while(1)
  return 1;
} // end listen packets

static void
sel_wait(Driver *drvthis){
  PrivateData * p = drvthis->private_data;

#ifdef TIMING_REPORT
  struct timeval now;
    struct timeval maxelapse;
    gettimeofday (&now, NULL);
#endif

    fd_set rfds;
    int retval;
    FD_ZERO(&rfds);
    FD_SET(p->fd, &rfds);
    struct timeval waiting;
    waiting.tv_sec=0;
//    waiting.tv_usec=80000;
    waiting.tv_usec=01000;
    int errno_sav;
    //  wait 3/4 of a second, or until a read would not block or eof 
    retval = select(p->fd+1, &rfds, NULL, NULL, &waiting);
    errno_sav=errno;
    if (retval == -1){
      report(RPT_CRIT, "%s select() error %d %s.",
        __FUNCTION__, errno, strerror(errno_sav));
      //struct timespec nod = {2, 0};
      //nod.tv_sec=2;          nanosleep(&nod, NULL); // b4 e xit on select() error 
      //exit(EXIT_FAILURE);
    }
    else if (retval){
/*
#ifdef DEBUG 
      report(RPT_QUE, "%s               % 33s             waited  %1.3f", __FUNCTION__,
          p->pkn[p->sendq[p->sending].command], 
          (80000-(float)waiting.tv_usec)/1000000   );
#else
      report(RPT_QUE, "%s      select()                                %02d                waited  %1.3f",
          __FUNCTION__, p->sendq[p->sending].command, 
          (80000-(float)waiting.tv_usec)/1000000   );
#endif
*/
#ifdef TIMING_REPORT
      gettimeofday (&maxelapse, NULL);
      unsigned long sel_s=time_diff(drvthis, now, maxelapse);
      p->listen[2]=(p->listen[2]*(double)p->listen_t + sel_s) / (double)(p->listen_t);
      ++p->listen_t;
      if (sel_s <  p->listen[0]){
        p->listen[0]= sel_s;
        debug(RPT_QUE, "%s new min wait %1.6f",
              __FUNCTION__, (float)p->listen[0]/1000000 );
      }
      else if (sel_s >  p->listen[1]){
        p->listen[1]= sel_s;
        debug(RPT_QUE, "listen_packets new max wait %1.6f",
            (float)p->listen[1]/1000000 );
      }
#endif
    } // end normal select return

}// end sel_wait

static void
sensors_packet(Driver *drvthis)
  /*
   *   calling check _ for _ packet from within treat _ packet 
   *   creates infinitely deepening recursion, necessarily....
   *   is this, or something similar still so ?  
   *   threaded separation of sending & listening, with outgoing packet que ...
   */
{
  PrivateData * p = drvthis->private_data;
  int top=(0xC0&(p->incoming_command.command)) >> 6;
  int cmd=(0x3F&(p->incoming_command.command));

  switch (top) {
    case 3 :   // error, malformed packet 
      pthread_mutex_lock(&quemutex); /* malformed packet move to next */ 
#ifdef DEBUG
      report(RPT_CRIT, "%s(%s on %s) malformed packet %d %s -> next",
        __FUNCTION__, p->model_str, p->device, cmd, p->pkn[cmd]);
      show_packet(drvthis, RPT_CRIT, &p->sendq[p->sending]);
#else
      report(RPT_CRIT, "%s(%s on %s) malformed packet %d -> next",
        __FUNCTION__, p->model_str, p->device, cmd );
#endif
      p->sendq[p->sending].command=CF633_Dummy_Unused;
      p->sending++;
      p->q_depth--;
      p->resend_repeats=0;
      // CF63oddtf_close(drvthis);
      pthread_mutex_unlock(&quemutex); /* malformed packet to next done */ 
      return;
    case 1 :  // top==1, is a good ack of a sent packet
      pthread_mutex_lock(&quemutex); /* ack of respsonse packet  */ 
      if  (cmd == p->sendq[p->sending].command) {  // is the expected ack
        if (cmd == CF633_Set_Fan_Power) {
          p->priority_packet_need-=1&p->priority_packet_need;
#ifdef FANS_USEC
          gettimeofday(&p->tdid_power,NULL);
          debug(RPT_FANS, "%s fan pow upd last req %ld.%06ld",
            __FUNCTION__, p->treq_power.tv_sec-FSEC ,
            p->treq_power.tv_usec);  
          debug(RPT_FANS, "%s fan pow upd      did %ld.%06ld  %6.6f",
            __FUNCTION__, p->tdid_power.tv_sec-FSEC,
            p->tdid_power.tv_usec,
            (float)time_diff(drvthis, p->treq_power,p->tdid_power)/1000000);  
#else  // FANS_USEC
          debug(RPT_FANS, "%s fan pow upd ack rec.", __FUNCTION__);
#endif // FANS_USEC
          //return; /*other things need to happen */
        }  //end rec CF633_Set_Fan_Power rpt
        else if (cmd == CF633_Set_Up_Fan_Reporting) {
          p->priority_packet_need-=2&p->priority_packet_need;
          debug(RPT_FANS, "%s fan rpt upd   success ack rec",
            __FUNCTION__ );
          //return; /*other things need to happen */
        } // end CF633_Set_Up_Fan_Reporting rpt
        else if (cmd == CF633_Set_LCD_And_Keypad_Backlight){
          p->priority_packet_need-=4&p->priority_packet_need;
        }
        // end of cases for expected ack of a priority packet
#ifdef TIMING_REPORT
        if (p->q_resent==-3){
          if (p->resend_repeats==0){ // discard times on resent packets
            gettimeofday(&p->t_ack,NULL);
            long this_ack=time_diff(drvthis, p->tsent, p->t_ack);
            if (this_ack < p->ack_d[cmd][0])  p->ack_d[cmd][0]=this_ack;
            if (this_ack > p->ack_d[cmd][1]) { 
	      p->ack_d[cmd][1]=this_ack;
#ifdef DEBUG
              debug(RPT_QUE, "%s max ack++ %32s %d", __FUNCTION__, p->pkn[cmd], p->ack_d[cmd][1]);
#endif
            }
            if (p->ack_d[cmd][2]==0) p->ack_d[cmd][2]=this_ack;
            else {
              p->ack_d[cmd][2]=( ( p->ack_c[cmd][0]*p->ack_d[cmd][2] + this_ack ) / (p->ack_c[cmd][0]+1)  ) ;
            }
            p->ack_c[cmd][0]++;
            p->ack_c_t++;
            p->ack_c_t_qd+=p->q_depth;
            memcpy(&p->last_good_ack.command, &p->sendq[p->sending].command, p->sendq[p->sending].data_length+2);
            memcpy(&p->l_g_a[cmd].command, &p->sendq[p->sending].command, p->sendq[p->sending].data_length+2);
	  } // end (p->resend_repeats==0)
        } // end (p->q_resent==-3)
        else { // this is a resent packet
/* 
  The plan was to increase timeout if packets keep failing.
  But the following data suggests that past 0.045 ack time, packets are simply not acked.
  :increasing timeout past 0.045 wastes time wiating for an ack that will never come.
Apr 20 20:21:51  2012 
 count     %    min t     max t  avg    timeout   dump resent xtra
   7181  51.72  0.004190  0.045  0.016   1.009    11    30     0    Set_LCD_Special_Character_Data
     28  00.20  0.020192  0.041  0.022   0.021     0     0     0    Set_Up_Fan_Reporting
    148  01.06  0.004598  0.040  0.017   0.023     0     0     0    Set_Fan_Power
   2728  19.64  0.006771  0.041  0.021   1.020     8     4     0    Send_Data_to_LCD
   3797  27.34  0.008140  0.040  0.020   1.065    14    19     3    Set_GPIO_Pin
  13883         0.004190  0.045  0.018            33    53     3    All
 CFA635:h1.0,v1.3 avg que depth   7.9098 = ~  0.1410 sec., max 215,  times over 128   0
 13883 good,    33 purged, 0.238 % overload.     7.713 packets/sec, elapse 1800.024
 FIXME / CHECKME : I changed the below from 1000000 to 0045000 and this happend:
                 : How did that get rid of resends ?
  count     %    min t     max t  avg    timeout   dump resent xtra
   7200  52.74  0.003873  0.045  0.015   0.045     0     0     0    Set_LCD_Special_Character_Data
    120  00.87  0.004263  0.038  0.017   0.023     0     0     0    Set_Fan_Power
   2658  19.47  0.005831  0.043  0.021   0.033     0     0     0    Send_Data_to_LCD
   3672  26.90  0.004322  0.040  0.020   0.800     0     0     0    Set_GPIO_Pin
  13650         0.003873  0.045  0.018             0     0     0    All
 CFA635:h1.0,v1.3 avg que depth   1.6334 = ~  0.0288 sec., max   8,  times over 128   0
 13650 good,     0 purged, 0.000 % overload.     7.583 packets/sec, elapse 1800.107

*/          
/*        one second packet ack waiting         
          if (p->ack_d[cmd][3]<1000000) p->ack_d[cmd][3]*=1.1; 
*/
          if (p->ack_d[cmd][3]<  50000) p->ack_d[cmd][3]*=1.1;
#ifdef DEBUG
          debug(RPT_QUE, "%s RSack %02x  q %02x %32s             %02x (%03d) > %.3f",
            __FUNCTION__,  p->q_resent, p->sending, p->pkn[cmd], p->endq, p->q_depth,
	    (double)p->ack_d[cmd][3]/1000000);
#endif
          p->q_resent = -3;
        } // end (p->q_resent ?= -3) conditions
#endif // TIMING_REPORT, 60 lines above?
        p->resend_repeats=0;
        unsigned char next=p->sending+1;

        if (p->priority_packet_need > 0 ) { // priority que jumping packet
          unsigned char need=0;
          if (1&p->priority_packet_need) need=0;
	  else if (2&p->priority_packet_need) need=1;
	  else if (4&p->priority_packet_need) need=2;
#ifdef DEBUG
            debug(RPT_QUE, "%s ack q %02x (%03d) %32s      p %d  %32s %02x %02x (%03d)",
              __FUNCTION__, p->sending, p->q_depth, p->pkn[cmd], p->priority_packet_need,
              p->pkn[p->priority_packet[need].command], next, p->endq, p->q_depth);
#endif  
	    // if it gets resent sendq[p->sending]. must be up to date
            commit_packet(drvthis, &p->priority_packet[need]);
            memcpy(&p->sendq[p->sending].command, &p->priority_packet[need].command, p->priority_packet[need].data_length+2);
            gettimeofday (&p->tsent, NULL);

          // was a good paket, timing data taken, pririty packet sent, no que adjustments, so return ok
          pthread_mutex_unlock(&quemutex); /* ack of respsonse packet -> priority packet sent */ 
          return;
        } // end priority_packet_need > 0 


        else { // no que jumping priority packets 
          if (p->q_depth > 1){
#ifdef DEBUG
            debug(RPT_QUE, "%s ack q %02x (%03d) %32s    next   %32s %02x %02x (%03d)",
              __FUNCTION__, p->sending, p->q_depth, p->pkn[cmd],
              p->pkn[p->sendq[next].command], next, p->endq, p->q_depth-1);
#endif
            /* send the next packet */
            commit_packet(drvthis, &p->sendq[next]);
            /* say sendq[this] is now unimportant  */
            p->sendq[p->sending++].command=CF633_Dummy_Unused;
            p->q_depth--; 
            gettimeofday (&p->tsent, NULL);
            pthread_mutex_unlock(&quemutex); /* ack of respsonse packet -> next sent */ 
            return;
          } // end (p->q_depth > 1)
          else if (p->q_depth == 1 ){ // last packet in que
            p->sendq[p->sending].command=CF633_Dummy_Unused;
            p->q_depth--; 
#ifdef DEBUG
            debug(RPT_QUE, "%s ack q %02x (%03d) %32s       end %32s %02x %02x (%03d)",
              __FUNCTION__, p->sending, p->q_depth+1, p->pkn[cmd],
              p->pkn[p->sendq[p->sending].command],
              next,  p->endq, p->q_depth);
#endif
#ifdef TIMING_REPORT 
	    if (p->q_recover!=0) {
              debug(RPT_QUE, "%s q_depth=%d after high of %d", __FUNCTION__,
                p->q_depth, p->q_recover);
              p->q_recover=0;
              p->deepies++;
            }
#endif // TIMING_REPORT
            pthread_mutex_unlock(&quemutex); /*ack of respsonse packet->que now empty*/ 
            return;
          } // q_depth = ? conditions  
        } // priority que jumper cases
        // was top==1, expected command ack, all queing done, so 
        report(RPT_ERR, "%s some condition not handled return?", __FUNCTION__);
        return; // every case above has return in this block
      } // rec'd wanted ack
/*
      if ( // is an ack & common cmd, but not the ack expected
        (cmd==CF633_Set_LCD_Special_Character_Data) ||
        (cmd==CF633_Set_LCD_Contents_Line_One) ||
        (cmd==CF633_Set_LCD_Contents_Line_Two) ||
        (cmd==CF633_Send_Data_to_LCD) ||
        (cmd==CF633_Set_LCD_And_Keypad_Backlight) ||
        (cmd==CF633_Set_Fan_Power) ||
        (cmd==CF633_Set_Up_Fan_Reporting) ||
        (cmd==CF633_Clear_LCD_Screen) ||
        (cmd==CF633_Configure_Key_Reporting) ||
        (cmd==CF633_Set_LCD_Contrast) ||
        (cmd==CF633_Set_LCD_Cursor_Style) ||
        (cmd==CF633_Reboot) ||
        (cmd==CF633_Set_GPIO_Pin) ||
        (cmd==CF633_Ping_Command)  )
*/
#ifdef DEBUG
        debug(RPT_QUE, "%s ?ck q %02x (%03d) %32s             ?=%32s",
          __FUNCTION__, p->sending,p->q_depth, p->pkn[p->sendq[p->sending].command], 
          p->pkn[cmd] );
#ifdef TIMING_REPORT
          p->ack_c[cmd][3]++;  // tally of extra / unexpected / suprise acks
#endif
        //show_packet (drvthis, RPT_ERR, &p->incoming_command);
#endif
        pthread_mutex_unlock(&quemutex); /* ack of unexpected/uknown respsonse packet */ 
        return;
    case 2 : // top==2, autorep packet or key press
      switch (p->incoming_command.command) {
        case 0x80: // key press or release 
#ifdef DEBUG 
          debug (RPT_QUE, "%s %78s", __FUNCTION__, "Key Press or Release");
#endif
#ifdef TIMING_REPORT
          p->ack_rep[0]++;
#endif
          AddKeyToKeyRing(&p->keyring, p->incoming_command.data[0]);
          if (p->q_depth>5){
            purge_que(drvthis, __FUNCTION__);
          }
          pthread_mutex_unlock(&quemutex); /* rec'd keypress */ 
          return;
	case 0x81 : // fan autoreport packet
#ifdef DEBUG 
          debug (RPT_FANS, "%s %78s %d", __FUNCTION__,
            "Fan RPM Report", p->incoming_command.data[0] +1);
#endif
#ifdef TIMING_REPORT
          p->ack_rep[1]++;
#endif
          memcpy (p->fans[p->incoming_command.data[0]].data, &p->incoming_command.data[1], 3);
          pthread_mutex_unlock(&quemutex); /* rec'd fan report */ 
#ifdef CF635LEDS
          if ((p->model==635) && ((p->fans[p->incoming_command.data[0]].led ==2)
                                  ||(p->fans[p->incoming_command.data[0]].led ==4)))
             compute_fan_rpm (drvthis, p->incoming_command.data[0] ); 
#endif
          return;
	case 0x82 :
#ifdef DEBUG 
          debug (RPT_TEMPS, "%s %78s %d", __FUNCTION__, 
            "Temperature Report", p->incoming_command.data[0] +1);
#endif
#ifdef TIMING_REPORT
          p->ack_rep[2]++;
#endif
          // temperature, store p->incoming_command.data[2,3] here
	  if ((p->ds18!=NULL) && (p->ds18[p->incoming_command.data[0]]!=NULL))
	    memcpy (p->ds18[p->incoming_command.data[0]]->data, &p->incoming_command.data[1], 3);
          return; /* FIXME, why is this unlock after the return? noticed 2012 04 17 */
          pthread_mutex_unlock(&quemutex); /* rec'd temp report */ 
        default : 
#ifdef DEBUG
          report(RPT_WARNING, "%s rec. a ? report packet: %d %s",
            __FUNCTION__, cmd, p->pkn[cmd]);
          show_packet (drvthis, RPT_ERR, &p->incoming_command);
#endif
          pthread_mutex_unlock(&quemutex); /* rec'd unknown report */ 
          return;
      } // end switch cmd for top == 2, types of autoreport packets
    // case 0 :  // outgoing packet here? : Hasn't happened in a long time ...
    default :
      report(RPT_ERR, 
        "%s rec unknown outgoing packet, resend commit packet:", __FUNCTION__ );
      commit_packet(drvthis, &p->incoming_command);
#ifdef DEBUG
      show_packet(drvthis, RPT_ERR, &p->incoming_command);
#endif
      pthread_mutex_unlock(&quemutex); /* rec'd  outgoing packet */ 
      return;
  } // end switch (top==?)

  // is not a key, temp, or fan  ....
  report(RPT_ERR, "%s Unknown / Unhandled incoming packet type %d:",
      __FUNCTION__, cmd);
#ifdef DEBUG
  show_packet (drvthis, RPT_ERR, &p->incoming_command);
#endif
  pthread_mutex_unlock(&quemutex); /* Unknown / Unhandled incoming packet */ 
  return;  // and nothing else to do... 
} // end of sensors packet

/*
 * send fan autorpt mask
 * presumes:   p->fans[fan].autorep 
 *     p->priority_packet[0].data[fan]
 * are properly set.
 */
void set_fan_rpt (Driver *drvthis)
{
  PrivateData * p = drvthis->private_data; 
  unsigned char fan;
  for (fan=0; fan < 4 ; fan ++) {
    p->fans[fan].autorep_mask=( p->fans[fan].f_sens[0]==NULL) ? p->fans[fan].autorep :
      p->fans[fan].autorep*(p->priority_packet[0].data[fan]!=0) ;
  }
  p->priority_packet[1].data[0]=p->fans[0].autorep_mask +
      p->fans[1].autorep_mask*2 +
      p->fans[2].autorep_mask*4 +
      p->fans[3].autorep_mask*8;
  debug(RPT_FANS, "%s fan rep mask upd -> %d %d %d %d = %02d.", 
    __FUNCTION__,  p->fans[0].autorep_mask, p->fans[1].autorep_mask,
    p->fans[2].autorep_mask, p->fans[3].autorep_mask, p->priority_packet[1].data[0]);
  if (p->q_depth==0) {
    send_packet(drvthis, &p->priority_packet[1]);
  } else {
    p->priority_packet_need |=2;
  }
} // end of set_fan_rpt

/*
 * send fan power mask
 * presumes:   p->priority_packet[0].data[fan]
 * is properly set.
 */
void set_fan_pow (Driver *drvthis)
{
    PrivateData * p = drvthis->private_data; 
#ifdef FANS_USEC
  debug(RPT_FANS, "%s fan pow upd -> %3d %3d %3d %3d   %ld.%06ld", 
    __FUNCTION__, p->priority_packet[0].data[0], p->priority_packet[0].data[1], 
    p->priority_packet[0].data[2], p->priority_packet[0].data[3],
    p->tdid_power.tv_sec- FSEC,
    p->tdid_power.tv_usec); // ~s since I wrote this... 
#else
  debug(RPT_FANS, "%s fan pow upd -> %3d %3d %3d %3d", 
    __FUNCTION__, p->priority_packet[0].data[0], p->priority_packet[0].data[1], 
    p->priority_packet[0].data[2], p->priority_packet[0].data[3]); 
#endif // FANS_USEC
  if (p->q_depth==0) {
    send_packet(drvthis, &p->priority_packet[0]);
  } else {
    p->priority_packet_need |=1;
  }
} // end of set_fan_pow

/* 
 * given a temp #, compute degrees celcius from raw data
 */
int compute_temp (Driver *drvthis, unsigned char temp){
  // only compute when (1) log, (2) adjust fan, (3) put on display
  PrivateData * p = drvthis->private_data; 
  unsigned char  msb=p->ds18[temp]->data[1] % 8; 
        // only 3 least significant bits are temp values
  char sign = (p->ds18[temp]->data[1] > 7) ? -1 : 1;
        // > 7 means subzero   
  if (sign == -1) msb = 7 - msb; 
        // maxim-ic.com's ds18b20.pdf p.3 for bit usage 
      // getting bad crc disable this to see...
      // bad crc computed = -15.9375
      // Widows reinstall disabled some or all key presses
      // re-enabled key presses CRC errs dissapeared ? correlated ?


  /* make a new function for handling bad temp crcs ~ 50 lines here */
  static unsigned int bad_crc_rec[32]={0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0};
  if (p->ds18[temp]->data[2] == 0){
    bad_crc_rec[temp]++;
  // test if ((p->ds18[temp]->data[2] == 0) || (temp==8)){
    if (bad_crc_rec[temp] < BAD_CRC_CUT){
      report(RPT_CRIT, "%s sensor %02d %s %02x%02x%02x BAD CRC count %d .",
          __FUNCTION__, temp+1, p->ds18[temp]->name , 
          p->ds18[temp]->romid[1], p->ds18[temp]->romid[2], p->ds18[temp]->romid[3], 
          bad_crc_rec[temp]);
    } // end is early count of bad crc
    if (bad_crc_rec[temp]==BAD_CRC_CUT) {
      // 5 bad crcs this temp
      p->ds18[temp]->abstracto=DEFAULT_DS18_TEMP; // denotes bad CRC
      report(RPT_CRIT, 
        "%s sensor %02d %s %02x%02x%02x has likely failed.",
        __FUNCTION__, temp+1, p->ds18[temp]->name, 
        p->ds18[temp]->romid[1], p->ds18[temp]->romid[2], p->ds18[temp]->romid[3] );
        bad_crc_rec[temp]++; /* false increment to avoid repeating this message */


      /* FIXME do bad crc checking somewhere else? 
        write the bad temp crc for fans in manage fans:
        -- no searching through fans here, manage fans has all the fan data 
        -- temperature routine should not need to know about fans
      */

      int check_fans =0;
      for (check_fans=0; check_fans < 4; check_fans ++ ) {
        if ( (&p->ds18[temp]->abstracto == NULL ) /* has sensor been deconfigured ? */
             || (p->fans[check_fans].f_sens[0]== NULL)
             || (p->fans[check_fans].f_sens[0]->abstvalptr == NULL )
           ) { 
               report(RPT_FANS, "%s fan %d %s _Temp is not used.",
                 __FUNCTION__, check_fans+1,  p->fans[check_fans].name );
        }
            // if a fan depends on this sensor, throttle up fan
        else if (p->fans[check_fans].f_sens[0]->abstvalptr == &p->ds18[temp]->abstracto) {
          report(RPT_CRIT, "%s fan %d %s _Temp is sensor %d %s %02x%02x%02x",
            __FUNCTION__, check_fans+1,  p->fans[check_fans].name,
            temp+1, p->ds18[temp]->name,
            p->ds18[temp]->romid[1], p->ds18[temp]->romid[2],
            p->ds18[temp]->romid[3] );
/*           deconfiguring these caused LCDd to hang or crash, 
             what else was depending on them ?
             manage fans checks if p->fans[check_fans].f_sens[0]==NULL 
typedef struct {
  char *name;               // pointer to name str of s_val
  int *abstvalptr;          // pointer to value
                            // temp * 16 , ds18 abstracto-units
  unsigned char number;     // tmp number for passing to update func
  void (*updates)(Driver *, unsigned char);
                            // pointer to function that updates val
} FAN_SENSOR;

          p->fans[check_fans].f_sens[0]->name=NULL;
          p->fans[check_fans].f_sens[0]->abstvalptr=NULL;
*/
          free (p->fans[check_fans].f_sens[0]);
          p->fans[check_fans].f_sens[0]=NULL;
          if ( p->fans[check_fans].f_sens[1]!= NULL) {
            free (p->fans[check_fans].f_sens[1]);
            p->fans[check_fans].f_sens[1]=NULL;
          }
          p->fans[check_fans].power=100;
          p->priority_packet[0].data[check_fans]=100;
          set_fan_pow (drvthis);
          report(RPT_CRIT, 
            "%s fan %d %s set to full power on f%d_Temp=%d|%s sensor failure.",
            __FUNCTION__, check_fans +1,  p->fans[check_fans].name,
             check_fans +1, temp+1, p->ds18[temp]->name );
        } else { 
          report(RPT_CRIT, "%s fan %d %s _Temp is a different sensor.",
            __FUNCTION__, check_fans+1,  p->fans[check_fans].name );
        } // end check this fan f_sens[0]

        if ( (&p->ds18[temp]->abstracto == NULL ) /* has sensor been deconfigured ? */ 
             || (p->fans[check_fans].f_sens[1]== NULL)
             || (p->fans[check_fans].f_sens[1]->abstvalptr == NULL )
           ) { 
               report(RPT_FANS, "%s fan %d %s _Ref_Temp is not used.",
                 __FUNCTION__, check_fans+1,  p->fans[check_fans].name);
             }
        else if (p->fans[check_fans].f_sens[1]->abstvalptr == &p->ds18[temp]->abstracto) {
          report(RPT_CRIT, "%s fan %d %s  _Ref_Temp is sensor %d %s",
            __FUNCTION__, check_fans+1,  p->fans[check_fans].name,
            temp+1, p->ds18[temp]->name );
          /* deconfiguring these caused LCDd to hang or crash, 
             what else was depending on them ? 
          p->fans[check_fans].f_sens[1]->name=NULL;
          p->fans[check_fans].f_sens[1]->abstvalptr=NULL;
          //is the next line freeing the ds18 struct?  
          free (p->fans[check_fans].f_sens[1]);
          */
          p->fans[check_fans].f_sens[1]=NULL;
          if ( p->fans[check_fans].f_sens[0]!= NULL) {
            //free (p->fans[check_fans].f_sens[0]);
            p->fans[check_fans].f_sens[0]=NULL;
          }
          p->fans[check_fans].power=100;
          p->priority_packet[0].data[check_fans]=100;
          set_fan_pow (drvthis);
          report(RPT_CRIT, 
            "%s fan %d %s set to full power on f%dRef_Temp=%d|%s sensor failure.",
            __FUNCTION__, check_fans +1, p->fans[check_fans].name,
            check_fans +1, temp+1, p->ds18[temp]->name );
        } else {
          report(RPT_CRIT, "%s fan %d %s Ref_Temp is a different sensor.",
            __FUNCTION__, check_fans+1,  p->fans[check_fans].name );
        } // end check this fan f_sens[1]

      } // end check each fan
    } // end is the fifth /* go for consecutive failures ? */ failure
    /* recompute & send sensor reporting mask without this sensor ?
       s_able_m is built in ds init; would have to rebuld */
  } // end is a bad crc


  else {
    if (bad_crc_rec[temp]< BAD_CRC_CUT){ 
      p->ds18[temp]->abstracto=
        ( sign*(p->ds18[temp]->data[0]+256*msb) ) ;
      if (bad_crc_rec[temp]!=0) {
        bad_crc_rec[temp]=0;
        report(RPT_CRIT, "%s sensor %02d %s good CRC, resetting count.",
          __FUNCTION__, temp+1, p->ds18[temp]->name );
      }
    }
  }
  report(RPT_TEMPS, 
     "%s sensor %d update %f",
      __FUNCTION__, temp+1, (float)(p->ds18[temp]->abstracto)/16);
  return (p->ds18[temp]->abstracto);
} // end of compute _temp

/* 
 * given a fan #, compute rpm from raw data
 */
float compute_fan_rpm (Driver *drvthis, unsigned char fan){
  // only compute when (1) log, (2) adjust fan, (3) put on display
  PrivateData * p = drvthis->private_data; 
  if (p->priority_packet[0].data[fan]==0) { // fan recently told to stop 
    p->fans[fan].data[0]=0;
    p->fans[fan].data[1]=0;
  }
  if (p->fans[fan].data[0]<3) {
    // device too slow to signal properly, 0=complete stop
    p->fans[fan].rpm_rpt= (float)p->fans[fan].data[0];
  }
  else {  // "27692308" is a constant 
    p->fans[fan].rpm_rpt=((
    (float)p->fans[fan].data[0]-3)*27692308) 
     / ((float)(*(unsigned short *)
     (&p->fans[fan].data[1]))
     *(float)p->fans
     [fan].divisor);
  }
  debug(RPT_FANS, "%s fan %d %16s    %02x %02x %02x = %f",
    __FUNCTION__, fan+1, p->fans[fan].name, p->fans[fan].data[0], 
    p->fans[fan].data[1], p->fans[fan].data[2], p->fans[fan].rpm_rpt );
#ifdef CF635LEDS
  if ((p->model==635) && ((p->fans[fan].led == 2 ) || (p->fans[fan].led == 4 ))){
    compute_led_by_rpm (drvthis, fan); 
  } /* end 635 && set led(s) by rpm */
#endif // CF635LEDS
  return p->fans[fan].rpm_rpt;
} /* end compute fan rpm */

/*
 * returns (now - past) in usec, seconds * 1000000
 * changed to seconds *1000 for longer intervals  
 */
long
time_diff (Driver *drvthis, struct timeval past, struct timeval now)
{
  //  PrivateData * p = drvthis->private_data; 
  struct timeval tdiff;
  // time_t  tv_sec;
  //   ./linux/types.h:typedef __kernel_time_t         time_t;
  //     ./asm/posix_types.h:typedef long                __kernel_time_t;
  // suseconds_t     tv_usec;
  //   ./linux/types.h:typedef __kernel_suseconds_t    suseconds_t;
  //     ./asm/posix_types.h:typedef long                __kernel_suseconds_t;
  // LONG_MAX     2 147 483 647L
  // if __WORDSIZE == 64 LONG_MAX     9223372036854775807L
  // 2 147 sec = 35 min, 47 sec @ 1000000 multi  
  // 2 147 483 sec = 24 days, 20 hours 31 minutes 23 sec
  
  tdiff.tv_sec=now.tv_sec-past.tv_sec;
  if (tdiff.tv_sec < 0) {
    tdiff.tv_sec=5;
    report (RPT_CRIT, "%s time has moved backwards!", __FUNCTION__);
  }
  if (now.tv_usec < past.tv_usec){
    tdiff.tv_sec --;
    tdiff.tv_usec = 1000000L + now.tv_usec- past.tv_usec;
  }
  else tdiff.tv_usec=now.tv_usec-past.tv_usec;
//  report(RPT_DEBUG, "%s %f = time diff ( %ld.%06ld, %ld.%06ld)", __FUNCTION__,
//      ((float)tdiff.tv_usec + (float)tdiff.tv_sec * 1000000)/ 1000000,
//      past.tv_sec, past.tv_usec,
//      now.tv_sec, now.tv_usec);
/* testing
 * static float minusec=LONG_MAX ;
  if ((float)tdiff.tv_usec < minusec) {
	  if (tdiff.tv_usec !=0 ) {
		  minusec = (float)tdiff.tv_usec ;
		  report(RPT_CRIT, "%s new non-zero minusec = %f", __FUNCTION__, minusec);
  		  report(RPT_CRIT, "%s past %ld now %ld tdiff.tv_usec = %ld", __FUNCTION__,
		  	past.tv_usec,
		  	now.tv_usec,
		  	tdiff.tv_usec);
	  }
	  else report(RPT_CRIT, "%s minusec ==0", __FUNCTION__ );
  }
*/
  return (long)(
      ((long)tdiff.tv_usec + (unsigned long int)tdiff.tv_sec * 1000000L));
}


void query_hw_ver (Driver *drvthis){
  PrivateData * p = drvthis->private_data;
  COMMAND_PACKET hw_querry;  
  int model=0;
  int top;
  struct timespec nod = {0, 015000000}; 

  hw_querry.command=CF633_Get_Hardware_And_Firmware_Version;
  hw_querry.data_length=0;

  while (model==0){
    send_packet(drvthis, &hw_querry);
    while (p->q_depth !=0) nanosleep(&nod, NULL); // hw_querry
    top=(0xC0&(p->incoming_command.command)) >> 6;
    if ((top==2) && (p->incoming_command.data_length==4)) {
      // empty responses, with lag 
      report(RPT_ERR, "%s receiving Fan autoreport packets!", __FUNCTION__);
      continue;
    }
    model=atoi((char*)&p->incoming_command.data[3]);
    if (model==0 ) {
      usleep(1000);  // b4 resending hw querry
    }
  } // end while empty querry response
  snprintf(p->model_str, p->incoming_command.data_length+1,  (char *)p->incoming_command.data);  
  report(RPT_INFO, "%s read: (top %d len %d) \"%s\" : %d", __FUNCTION__,
      top, p->incoming_command.data_length, 
      p->model_str, model);
  if ((model != 631) && (model != 633) && (model != 635)){
    report(RPT_ERR, "%s %s (%d) unrecognized. Will proceed with 633...",
        __FUNCTION__, p->model_str, model);
    model=633;
  }
  p->model=model;
#ifdef TIMING_REPORT
  // set these low and see when resends stop as wait increases
  if(p->model==633) { // 633 serial    /* one second = 1000000   */
/* 20012 04 20 observations suggesting that past         45000
 *   packets will never be acked.
 *   see notes in listen packet. using 635
*/
    p->ack_d[CF633_Set_LCD_Contents_Line_One][3]     =   81000;
    p->ack_d[CF633_Set_LCD_Contents_Line_Two][3]     =   59000;
    p->ack_d[CF633_Set_LCD_Special_Character_Data][3]=   77000;
    p->ack_d[CF633_Set_LCD_And_Keypad_Backlight][3]  =   23000;
    p->ack_d[CF633_Set_Fan_Power][3]                 =   28000;
    p->ack_d[CF633_Set_Up_Fan_Reporting][3]          =   21000;
  } else { // 631 and 635 usb
    p->ack_d[CF633_Set_LCD_Special_Character_Data][3]=   45000;
    p->ack_d[CF633_Set_LCD_And_Keypad_Backlight][3]  =   21000;
    p->ack_d[CF633_Send_Data_to_LCD][3]              =   33000;
    p->ack_d[CF633_Set_Fan_Power][3]                 =   23000;
    p->ack_d[CF633_Set_Up_Fan_Reporting][3]          =   21000;
    p->ack_d[CF633_Set_GPIO_Pin][3]                  =   45000;
  }
#endif
  /* New firmware version?
   * I will try to behave differently for firmware 0.6 or above.
   * Currently this is not in use. can parse from p->model_str
   */
  p->newfirmware = drvthis->config_get_bool(drvthis->name, "NewFirmware", 0, 0);

  /* if 633, turn off autodisplays */
  if (p->model==633) CF63oddtf_no_autodisp(drvthis);

  /* Which size */
  int w, h;
  char size[200] = DEFAULT_SIZE_CF633;
  char *default_size = DEFAULT_SIZE_CF633;
  if (p->model == 631)
    default_size = DEFAULT_SIZE_CF631;
  else if (p->model == 633)
    default_size = DEFAULT_SIZE_CF633;
  else if (p->model == 635)
    default_size = DEFAULT_SIZE_CF635;
  else if (p->model == 631)
    default_size = DEFAULT_SIZE_CF631;
  else {
    default_size = DEFAULT_SIZE_CF633;
    // rpt( do something more informative / intelligent )
  }
  strncpy(size, drvthis->config_get_string (drvthis->name, "Size", 0, default_size), sizeof(size));
  size[sizeof(size)-1] = '\0';
  report(RPT_INFO, "%s Size in config: %s.",  __FUNCTION__, size);
  if ((sscanf(size, "%dx%d", &w, &h) != 2)
      || (w <= 0) || (w > LCD_MAX_WIDTH)
      || (h <= 0) || (h > LCD_MAX_HEIGHT)) {
    report(RPT_WARNING, "%s Cannot read size: %s. Using default value.",  __FUNCTION__,  size);
    sscanf(default_size, "%dx%d", &w, &h);
  }
  if (strcmp(size, default_size)) {
    report(RPT_INFO, "%s Model supports: %s.",  __FUNCTION__, default_size );
  }
  int m_w, m_h;
  sscanf(default_size, "%dx%d", &m_w, &m_h);
  if (   (w > m_w ) || (h > m_h) ) {
    report(RPT_WARNING, "%s Size too large: %s.  Using supported value(s).",  __FUNCTION__, size);
    if (w > m_w) w=m_w;
    if (h > m_h) h=m_h;
  }
  p->width = w;
  p->height = h;
  report(RPT_INFO,"%s Size used: %dx%d.",
       __FUNCTION__, p->width, p->height);

} // end query_hw_ver

void spawn_clones (Driver *drvthis, unsigned char bitmask) {
/* bitmask 0|1 0 nothing  1 do prelim init 
 *         0|2            2 spawn listen
 *         0|4            4 spawn fan
 *         0|8            8 spawn log client
 *         0|16          16 final init
 */
  PrivateData * p = drvthis->private_data;
  int errno_sav=0;
  //pthread_attr_t attr;
  //pthread_attr_init( &attr);
  //pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  if (1&bitmask){
// clones should run as same user as parent 
    strncpy(p->user, drvthis->config_get_string( "Server", "User", 0, "\01" ), sizeof(p->user));

    debug(RPT_CLICRE, "%s read user name: %40s", __FUNCTION__, p->user ); 
/*    
    struct passwd pd;
    p->user_pwd=&pd;
    struct passwd* pdptr; 
    char *buf;
    buf=calloc(1, sysconf(_SC_GETPW_R_SIZE_MAX));
    if ( 
         getpwnam_r (  
                     p->user,
                     p->user_pwd,
                     buf,
                     sysconf(_SC_GETPW_R_SIZE_MAX),
                     &pdptr 
                   ) != 0) {
      report (RPT_ERR, "%s getpwuid_r failed!.",
          __FUNCTION__ );   
    }


    if (strcmp( p->user, "\01" ) !=0  ) {
      //p->user_pwd=getpwnam(p->user);
      struct passwd* user_pwd=getpwnam(p->user);
      //if (p->user_pwd==NULL){
      if (user_pwd==NULL){
        report(RPT_ERR, "%s %.40s is not a valid user name.",
          __FUNCTION__, p->user ); 
      }  
      else {
        debug(RPT_CLICRE, "%s clones to run as %d  %.40s",
         // __FUNCTION__, p->user_pwd->pw_uid, p->user_pwd->pw_name ); 
          __FUNCTION__, user_pwd->pw_uid, user_pwd->pw_name ); 
        p->userid=user_pwd->pw_uid;
      }
    }
    else {
      debug(RPT_CLICRE, "%s children will not change privalages.",
        __FUNCTION__);
      p->user_pwd=NULL;
    }
*/
#ifdef TIMING_REPORT
  int tmp;
  for (tmp=0;tmp < MAX_COMMAND+2; tmp++) {
    p->ack_d[tmp][3]=RESEND_TIMEOUT;
  }
  // set these low and see when resends stop as wait increases
  if(p->model==633) { // 633 serial 
    p->ack_d[CF633_Set_LCD_Contents_Line_One][3]     =  181000;
    p->ack_d[CF633_Set_LCD_Contents_Line_Two][3]     =   59000;
    p->ack_d[CF633_Set_LCD_Special_Character_Data][3]=  177000;
    p->ack_d[CF633_Set_LCD_And_Keypad_Backlight][3]  =   23000;
    p->ack_d[CF633_Set_Fan_Power][3]                 =   28000;
    p->ack_d[CF633_Set_Up_Fan_Reporting][3]          =   21000;
  } else { // 631 and 635 usb
    p->ack_d[CF633_Set_LCD_Special_Character_Data][3]=   77000;
    p->ack_d[CF633_Set_LCD_And_Keypad_Backlight][3]  =   21000;
    p->ack_d[CF633_Send_Data_to_LCD][3]              =   33000;
    p->ack_d[CF633_Set_Fan_Power][3]                 =   23000;
    p->ack_d[CF633_Set_Up_Fan_Reporting][3]          =   21000;
  }
    /* init timing data */
    clear_timing(drvthis);
#ifdef LIVE_TIMING
    int id=pthread_create(&p->timing_pth, NULL, (void *)timing_loop,  drvthis);
    if (id !=0) {
      report(RPT_CRIT, "%s timing pthread_create error %d %s.",
        __FUNCTION__, errno, strerror(errno_sav));
      struct timespec nod = {0, 999000000}; nanosleep(&nod, NULL); // let error get sent 
      CF63oddtf_close (drvthis);  
    }
#endif // LIVE_TIMING
#endif // TIMING_REPORT
#ifdef __linux__
    debug(RPT_INFO, "pthread_self % 15s = %d tid = %d", 
      __FUNCTION__, pthread_self(), syscall(SYS_gettid));
#endif
  } /* end bitmask 0|1 do preliminary init */
   
  if (2&bitmask){
    //int id=pthread_create(&p->listen_pth, &attr, (void *)listen_packets,  drvthis);
    int id=pthread_create(&p->listen_pth, NULL, (void *)listen_packets,  drvthis);
    errno_sav=errno;
    if (id !=0) {
      report(RPT_CRIT, "%s listen pthread_create error %d %s.",
        __FUNCTION__, errno, strerror(errno_sav));
      struct timespec nod = {0, 999000000}; nanosleep(&nod, NULL); // let error get sent 
      CF63oddtf_close (drvthis);  
    }
  } /* end bitmask 0|2 do preliminary init */

  if (4&bitmask){
    if ( (p->fans[0].f_sens[0] == NULL ) && (p->fans[0].power==100)
      && (p->fans[1].f_sens[0] == NULL ) && (p->fans[1].power==100)
      && (p->fans[2].f_sens[0] == NULL ) && (p->fans[2].power==100)
      && (p->fans[3].f_sens[0] == NULL ) && (p->fans[3].power==100)){
      report(RPT_CLICRE, "%s fan loop has no tasks, not starting.", __FUNCTION__);
      p->fan_pth=0;
    } else { // reason(s) to start fan loop
      int id=pthread_create(&p->fan_pth, NULL, (void *)fan_loop,  drvthis);
      //int id=pthread_create(&p->fan_pth, &attr, (void *)fan_loop,  drvthis);
      errno_sav=errno;
      if (id !=0) {
        report(RPT_CRIT, "%s fan pthread_create error %d %s.",
          __FUNCTION__, errno, strerror(errno_sav));
        struct timespec nod = {6, 999900000}; nanosleep(&nod, NULL); // let error get sent 
        CF63oddtf_close (drvthis);  
      }
    } // end start fan loop or no
  } /* end bitmask 0|4 start fan thread*/
  
  if (8&bitmask){
/*
#ifndef TIMING_REPORT
    // timing report wrap function is in this loop
    // so with timing report this loop is needed
    if (p->log_sens==0)
      // FIXME: other reasons to start loop: any client screens 
      ){
      // power==100 checks ... is there a set fan power b4 log client ?
      report(RPT_INFO, "%s log client has no tasks, not starting.", __FUNCTION__);
      return;
    }  
#endif //  ndef TIMING_REPORT
*/
#ifdef CLIENT_CHILD /* in spawn clones before starting csv_ cli */
    csv_init(drvthis);
#endif
      int id=pthread_create(&p->log_pth, NULL, (void *)csv_cli,  drvthis);
      //int id=pthread_create(&p->log_pth, &attr, (void *)csv_cli,  drvthis);
      errno_sav=errno;
      if (id !=0) {
        report(RPT_CRIT, "%s csv log pthread_create error %d %s.",
          __FUNCTION__, errno, strerror(errno_sav));
        struct timespec nod = {0, 999000000}; nanosleep(&nod, NULL); // let error get sent 
        CF63oddtf_close (drvthis);  
      }
  } /* end bitmask 0|8 start log client*/
  if (16&bitmask){
    // do what else ?
  } /* end bitmask 0|16 start log client*/
} // end spawn_clones 

/*
void clones_drop_priv (Driver *drvthis, char *calling_clone, uid_t new_uid){
  //PrivateData * p = drvthis->private_data;
  int result, errno_sav;
  result=setuid(new_uid);
  errno_sav=errno;
  if (result < 0) {
    report(RPT_ERR, "%s by %25s Unable to switch to user %d: %s",
      __FUNCTION__, calling_clone, new_uid, strerror (errno_sav));
  } 
  else {
    report(RPT_CLICRE, "%s by %15s switched to user %d",
      __FUNCTION__, calling_clone, new_uid);
  }

    //setuid seems to have no effect on mepu->pw_name mepg->pw_name
    uid_t meu, meg;
    struct passwd *mepu, *mepg;
    meu = getuid();
    mepu = getpwuid(meu);
    meg = getgid();
    mepg = getpwuid(meg);
    report(RPT_CRIT, "%15s as   %d uid %s    %d gid %s",
      __FUNCTION__,  meu,  mepu->pw_name,  meg, mepg->pw_name );
  return;
} // end clones_drop_priv
*/      

#ifdef TIMING_REPORT
void timing_dump(Driver *drvthis) {
  PrivateData * p = drvthis->private_data;
  int cmd;
  
  // 2 147 sec = 35 min, 47 sec wrap 
  struct timeval now;
  gettimeofday (&now, NULL);
  unsigned long d_t=time_diff(drvthis, p->timing_mk, now);
  
  report(RPT_CRIT, "% 6lu Key  % 6lu   Fan % 6lu  Temp",
    p->ack_rep[0]/2, p->ack_rep[1],  p->ack_rep[2]);
  report(RPT_CRIT, "% 6lu GOOD  % 6lu TRY_AGAIN  % 6lu GIVE_UP",
    p->msg_type_count[GOOD_MSG],
    p->msg_type_count[TRY_AGAIN],
    p->msg_type_count[GIVE_UP]);
  
  long max=0, avg=0, min=1000000000;
  long int total=0, tot_d=0, tot_r=0, tot_x=0;
  for (cmd =0; cmd<  MAX_COMMAND+2 ; cmd++){
    if (p->ack_c[cmd][0] >0){
      if (min > p->ack_d[cmd][0]) min = p->ack_d[cmd][0];
      if (max < p->ack_d[cmd][1]) max = p->ack_d[cmd][1];
      if ( avg==0 ) 
        avg = p->ack_d[cmd][2];
      else 
        avg= (total*avg + p->ack_c[cmd][0]*p->ack_d[cmd][2]) / (total+ p->ack_c[cmd][0]);
      total+=p->ack_c[cmd][0];
      tot_d+=p->ack_c[cmd][1];
      tot_r+=p->ack_c[cmd][2];
      tot_x+=p->ack_c[cmd][3];
    }
  }
  report(RPT_CRIT, " count     %%    min t     max t  avg    timeout   dump resent xtra");
  for (cmd =0; cmd< MAX_COMMAND+2; cmd++)
//#ifdef DEBUG 
    if (p->ack_c[cmd][0] >1){
      int pct_lh=(int)(((float)p->ack_c[cmd][0]/(float)total)*100);
      int pct_rh=(int)(((float)p->ack_c[cmd][0]/(float)total - (float)pct_lh/100)*10000);
      report(RPT_CRIT, "% 6lu  %02d.%02d % .6f % .3f % .3f  % .3f % 5lu % 5lu % 5lu    %s",
        p->ack_c[cmd][0], pct_lh, pct_rh, 
        (double)p->ack_d[cmd][0]/1000000, (double)p->ack_d[cmd][1]/1000000, 
	(double)p->ack_d[cmd][2]/1000000, (double)p->ack_d[cmd][3]/1000000,
         p->ack_c[cmd][1], p->ack_c[cmd][2], p->ack_c[cmd][3], p->pkn[cmd]);
    }
//#endif
    // else report(RPT_CRIT, "zero %s", p->pkn[cmd]);
  //report(RPT_CRIT, "%6lu          %.6f  %.3f  %.3f        %5lu %5lu %5lu %5lu      All",
  report(RPT_CRIT, "% 6lu        % .6f % .3f % .3f         % 5lu % 5lu % 5lu    All",
      total, (double)min/1000000, (double)max/1000000, (double)avg/1000000, 
      tot_d, tot_r, tot_x);
  // Que Depth Average = summ of all que depth measures / packets recd
  float qda   = (float)p->ack_c_t_qd / (float) p->ack_c_t;
  int   qda_lh= (int)qda;
  int   qda_rh= (int)((((float)p->ack_c_t_qd / (float) p->ack_c_t)-(int)((float)p->ack_c_t_qd / (float) p->ack_c_t))*10000);
  // Que Depth Average as seconds of time = qda * average packet time
  double qd_sec = (double)qda * ((double)avg/1000000);
  int qd_seclh= (int) qd_sec;
  int qd_secrh= (int)((qd_sec - (float)qd_seclh)*10000);  
  report(RPT_CRIT, "%s avg que depth % 3d.%04d = ~% 3d.%04d sec., max % 3u,  times over 128 % 3d",
      p->model_str, qda_lh, qda_rh,
      qd_seclh, qd_secrh, p->maxq, p->deepies);
//  report(RPT_CRIT, "listen_packets loop wait:  min %.6f  max %.5f  avg %.5f.",
//      (float)p->listen[0]/1000000,  (float)p->listen[1]/1000000,  (float)p->listen[2]/1000000);
  report(RPT_CRIT, "%5lu good, %5lu purged, %.3f %% overload.     %.3f packets/sec, elapse %.3f",
      p->ack_c_t, p->q_purged,
      p->q_purged / (float)p->ack_c_t  *100,
      (float)p->ack_c_t/((float)d_t/1000000) , (double)d_t/1000000);
} // end of timing_dump

#ifdef LIVE_TIMING
static void
install_signal(Driver *drvthis)
/*
dnl     OK signal from child on spawning threads?
dnl     CFLAGS="$CFLAGS -O0 -ggdb -save-temps"
dnl     CFLAGS="$CFLAGS -O0 -ggdb"
dnl     CFLAGS="$CFLAGS -O0"
dnl     spawning threads signals ok
dnl     CFLAGS="$CFLAGS -O"
        CFLAGS="$CFLAGS -O -ggdb -save-temps"
so, what deos -O0 do that -O does not ?
with install_signal called from child thread, kill - SIGUSR1 works
                    did I have it called from start of init for some reason?
                    did that hep gdb ?
*/
{
  /* Installs signal handlers to generate live timing dumps 
   * can also receive a reload signal.
   * sigaction() is favored over signal() */

  // call live_timing once to give it a (Driver *)
//  live_timing((int)drvthis);
  live_timing(0, (siginfo_t*) NULL, (void*) drvthis);
  // call live_timing_reset once to give it a (Driver *)
//  live_timing_reset((int)drvthis);
  live_timing_reset (0, (siginfo_t*) NULL, (void*) drvthis);
  
  //report(RPT_CRIT, "%s, driver_pointer %d set.",
  //    __FUNCTION__, (int)drvthis);
  
  struct sigaction sa;

  report(RPT_CRIT, "%s, live timing report   kill -SIGUSR1 %d", __FUNCTION__, getpid());
  report(RPT_CRIT, "%s, reset timing start   kill -SIGUSR2 %d", __FUNCTION__, getpid());
  sigemptyset ( &(sa.sa_mask) );
//  sa.sa_flags = SA_RESTART || SA_NODEFER ;
  sa.sa_flags = SA_SIGINFO || SA_RESTART || SA_NODEFER ;
//  sa.sa_handler = live_timing;
  sa.sa_sigaction  = live_timing;
#ifndef WIN32
  sigaction (SIGUSR1, &sa, NULL);    /* kill - 30 10 16*/
#else
        /* Win32 does not support POSIX signals i.e. sigaction(). However, it does 
         * support ANSI signals in mingw.  Heck if I know*/
  signal (SIGUSR1, &sa, NULL);    /* kill - 30 10 16*/
#endif
//  sa.sa_handler = live_timing_reset;
  sa.sa_sigaction = live_timing_reset;
#ifndef WIN32
  sigaction (SIGUSR2, &sa, NULL);    /* kill - 31 12 17*/
#else
  signal (SIGUSR2, &sa, NULL);    /* kill - 31 12 17*/
#endif
} // end install_signal

static void
live_timing(int signal_num, siginfo_t *sigfo, void *v_ptr){
//live_timing(int signal_number){
  static Driver *drvthis=NULL;
  // there is no signal NULL 
  if (drvthis==NULL) {
    drvthis=(Driver *)v_ptr;
    //report(RPT_CRIT, "%s, drvthis %d set & return.",
    //  __FUNCTION__, signal_number);
    return;
  }
  timing_dump(drvthis);
  clear_timing(drvthis);
}  

static void
live_timing_reset(int signal_num, siginfo_t *sigfo, void *v_ptr){
//live_timing_reset(int signal_number){
  static Driver *drvthis=(Driver *)NULL;
  // there is no signal NULL 
  if (drvthis==NULL) {
    drvthis=(Driver *)v_ptr;
    //report(RPT_CRIT, "%s, drvthis %d set & return.",
    //  __FUNCTION__, signal_number);
    return;
  }
  clear_timing(drvthis);
}  
static void timing_loop(Driver *drvthis){
  /* idea is : trapped signal will never interrupt a non-threadsafe function 
   * first line of comments not so clear eh?  what was the purpose ? :
   * install signal traps signals for the enitre family of threads, I think ... 
   * but when an interrupt arrives, only one thread is actually interrupted
   * that will be this one -- which does nothing.  Other threads need no 
   * interrupt / resume handling.  */
  install_signal(drvthis);
#ifdef __linux__
  debug(RPT_INFO, "pthread_self % 15s = %d tid = %d", 
    __FUNCTION__, pthread_self(), syscall(SYS_gettid));
#endif
  struct timespec nod = {9990, 999000000};
  while (1) {
    if (nanosleep(&nod, NULL)) {
      report(RPT_CRIT, "%s interrupted", __FUNCTION__);
    }
  }
}
#endif // ifdef LIVE TIMING functions

void
clear_timing(Driver *drvthis){
    PrivateData * p = drvthis->private_data;
  gettimeofday (&p->timing_mk, NULL);

/*  
 // test timing wrap
  struct timeval tester;
  gettimeofday (&tester, NULL);
  long future;
  
  // time_t  tv_sec;
  //   ./linux/types.h:typedef __kernel_time_t         time_t;
  //     ./asm/posix_types.h:typedef long                __kernel_time_t;
  // suseconds_t     tv_usec;
  //   ./linux/types.h:typedef __kernel_suseconds_t    suseconds_t;
  //     ./asm/posix_types.h:typedef long                __kernel_suseconds_t;
  // LONG_MAX     2 147 483 647L
  // if __WORDSIZE == 64 LONG_MAX     9223372036854775807L
  // 2 147 sec = 35 min, 47 sec @ 1000000 multi
  // so any time difference greater than 2 147 seconds is undandled 
  // with only 3 digits of right hand precision 2 147 483 sec = 24 days, 20 hours 31 minutes 23 sec
  for (future=2140; future < 2150  ; future ++) {
    tester.tv_sec=p->timing_mk.tv_sec + future;
    report(RPT_CRIT, "time diff test %f seconds.", 
      (float)time_diff(drvthis, p->timing_mk, tester)/1000000);
  }
*/

//  Nov  9 14:27:58 localhost LCDd: time diff test 2147.000000 seconds.
//  Nov  9 14:27:58 localhost LCDd: time diff test -2146.967000 seconds.
  
  p->ack_c_t=0;
  p->prevp=0;
  p->ack_c_t_qd=0;
  p->q_purged=0;
  p->listen[0]=10000000;
  p->listen[1]=0;
  p->listen[2]=0;
  p->listen_t=0;
  p->msg_type_count[0]=p->msg_type_count[1]=p->msg_type_count[2]=0;
  p->maxq=0;
  p->deepies=0;
  p->q_recover=0;
  p->ack_rep[0]=p->ack_rep[1]=p->ack_rep[2]=0;
  p->q_resent=-3;
  int tmp;
  for (tmp=0; tmp < MAX_COMMAND+2; tmp ++){
    p->ack_d[tmp][0]=9999999;
    p->ack_d[tmp][1]=0;
    p->ack_d[tmp][2]=0;
    p->ack_c[tmp][0]=0;
    p->ack_c[tmp][1]=0;
    p->ack_c[tmp][2]=0;
    p->ack_c[tmp][3]=0;
   }
} // end clear_ timing  

#endif // ifdef TIMING REPORT functions


int csv_cli (Driver * drvthis){ 
/* FIXME ?
  if there are no client lines
  and p->log_sens ==0
  this whole routine / thread does nothing but 
  print and clear timing data ?
  clear and print timing elsewhere upstream of this ?
  check for starting this before calling clone ?
*/ 
  PrivateData * p = drvthis->private_data;
  int port;

#ifdef __linux__
  debug(RPT_INFO, "pthread_self % 15s = %d tid = %d", 
    __FUNCTION__, pthread_self(), syscall(SYS_gettid));
#endif
  //debug(RPT_INFO, "%s started", __FUNCTION__);
  if (p->user_pwd!=NULL){
    int errno_sav;
    struct timespec nod = {0, 750000000}; 
    port=-1;
    while(port==-1) {
      port=nanosleep(&nod, NULL); // let server get up b4 ... 
      errno_sav=errno;
      if (port==-1) {
         report(RPT_CRIT, "%s waiting to clones_drop_priv %s", __FUNCTION__, strerror(errno_sav));
      }
    }
//    clones_drop_priv (drvthis, (char *)__FUNCTION__ , p->userid );
    /* fixme: put this before logfile is opened */
  }

  /* print initial column labels  */
  if (p->log_sens!=0) {
    csv_header (drvthis);
  } // p->log_sens!=0

#ifdef CLIENT_CHILD /* in csv_cli */
  char host[64];
  port=drvthis->config_get_int( "server", "port", 0, LCDPORT );
  strncpy(host, drvthis->config_get_string( "server", "bind", 0, "localhost" ), sizeof(host));
  cf_dsk = sock_connect( host, port );
  if( cf_dsk <= 0 ) {
    report(RPT_CRIT, "%s bad sock_connect", __FUNCTION__);
    return -1;
  }
  else {
    report(RPT_CLICRE, "%s socket %d opened", __FUNCTION__, cf_dsk);
  }
  report(RPT_DEBUG, "%s sock_send: %d hello", __FUNCTION__, cf_dsk);
  sock_send_string( cf_dsk, "hello\n" );
  if (read_connect_string(drvthis)==-1){
    report(RPT_ERR, "%s hello failed, exiting.", __FUNCTION__);
      struct timespec nod = {2, 0};
      nanosleep(&nod, NULL); // b4 e xit on hello failed 
    exit(EXIT_FAILURE);
  }
  else {
    report(RPT_CLICRE, "%s hello ok", __FUNCTION__);
  }
  
  report(RPT_DEBUG, "%s sock_send: %d client_set -name CF633log",
      __FUNCTION__, cf_dsk );
  sock_send_string( cf_dsk, "client_set -name CF633log\n");
  wait_success(drvthis, 0, NULL);
  if (lines >0) cli_scrn(drvthis);
#endif // CLIENT_CHILD
#ifdef TIMING_REPORT
  /* show init timing data & reset*/
  timing_dump(drvthis);
  clear_timing(drvthis);
#endif
#ifdef CLIENT_CHILD
  if ((lines >0) || (p->log_sens > 0) )client_loop(drvthis);
#else /*lines is  CLIENT_CHILD only */
  if  (p->log_sens > 0) client_loop(drvthis);
#endif
  return 1;
} // end csv_cli

#ifdef DO_ULT 
int ult_init(Driver *drvthis){
  PrivateData * p = drvthis->private_data;
  char buf[128], target[128];
  int tmp;
  p->pjt=NULL;
  p->jiffies[0]=p->jiffies[1]=p->jiffies[2]=NULL;
  /* Log uptime in syslog with fan / temp info or no ? */
  p->log_uptime=drvthis->config_get_bool(drvthis->name, "LogUptime",0,0);
  /* Log % cpu usage in syslog with fan / temp info or no ? */
  p->log_load=drvthis->config_get_bool(drvthis->name, "LogLoad", 0, 0);
  /* Log % system cpu usage in syslog with fan / temp info or no ? */
  p->log_load_s=drvthis->config_get_bool(drvthis->name, "LogLoad_System", 0, 0);
  /* Log % user cpu usage in syslog with fan / temp info or no ? */
  p->log_load_u = drvthis->config_get_bool(drvthis->name, "LogLoad_User", 0, 0);
  /* Log top cpu usage & how many or no ? */
  p->topnum = drvthis->config_get_int(drvthis->name, "TopNum", 0, 0);
  if (p->topnum > MAX_TOP){
    report(RPT_ERR, "%s TopNum %d over maximim of %d, set to %d",
        __FUNCTION__, p->topnum, MAX_TOP);
    p->topnum=MAX_TOP;
  }
  p->top_num=malloc(sizeof( CF_ULT *) *p->topnum);
  for (tmp=0; tmp < p->topnum; tmp++){ 
    p->top_num[tmp]=malloc(sizeof( CF_ULT ) );
    p->top_num[tmp]->name=NULL;
    p->top_num[tmp]->val=0;
  }

  /* Top track by name */
  p->topname=0;
  p->top_name=malloc(sizeof( CF_ULT *) * MAX_TOP);
  int offset=0;
  for (tmp=0; tmp < MAX_TOP; tmp++){
    sprintf(target, "TopName%d", tmp+1);
    snprintf(buf, sizeof(buf), drvthis->config_get_string (drvthis->name, target, 0, NULL));
    if (strlen(buf)==0) {
      report(RPT_CLICRE, "%s TopName%d= not found",
        __FUNCTION__, tmp+1);
      offset++;
      continue;
    }
    p->top_name[tmp-offset]=malloc(sizeof( CF_ULT ));
    p->top_name[tmp-offset]->val=0;
    p->top_name[p->topname]->name = (char *)malloc(strlen(buf)+1);
    sprintf(p->top_name[p->topname]->name, buf);
    if (p->topname==tmp) {
      report(RPT_CLICRE, "%s TopName%d=% 16s",
        __FUNCTION__, p->topname+1, p->top_name[p->topname]->name);
    } else {
      report(RPT_CLICRE, "%s TopName%d=% 16s ( %d became missing %d )",
        __FUNCTION__, tmp+1-offset,  p->top_name[p->topname]->name,
        tmp+1, p->topname+1 );
    }
#ifdef CLIENT_CHILD
    sprintf(target, "TopName%d_Lcd", tmp+1);
    p->top_name[p->topname]->lcd=drvthis->config_get_int(drvthis->name, target, 0, 0);
#endif // CLIENT_CHILD
    p->topname++;
  } // end of one possible topname

  if ((p->log_load !=0 ) || ( p->log_load_s!=0)
        || (p->log_load_u!=0) || (p->log_uptime!=0)
        || (p->topnum>0) || (p->topname > 0)){
    machine_init(); /* in ult_init */
    if ((p->log_load !=0 ) || ( p->log_load_s!=0) || (p->log_load_u!=0)){ 
      machine_get_load(&p->load); 
      /* store load  calls get_load below, 
         so prime machine_get_load's internal elapse timer */ 
      store_load(drvthis);  /* in ult_init */ 
    }
    if((p->topnum>0) || (p->topname > 0)){
      gettimeofday(&p->tdid_top, NULL);
      p->tdid_top.tv_sec-=p->log_sens;
      /* one call to init half the delta t data */
      store_tops_init(drvthis);  /* in ult_init */
    }
  }
  return (1);
} // end ult_init

void store_load(Driver *drvthis){
  PrivateData * p = drvthis->private_data;
  machine_get_load(&p->load);
  if (p->log_load !=0) {
    p->load_t=((float)p->load.user +(float)p->load.system +(float)p->load.nice)
		/(float)p->load.total*100 ;
//    report(RPT_CRIT, "%s total ticks p->load.total %d", __FUNCTION__, p->load.total);
//    report(RPT_CRIT, "%s load percent p->load_t %f", __FUNCTION__, p->load_t);
  }
  if (p->log_load_s !=0) {
    p->load_s=(float)p->load.system/(float)p->load.total*100;
  }
  if (p->log_load_u !=0) {
    p->load_u=(float)p->load.user/(float)p->load.total * 100;
  }
}

static int
sort_jiffpid (void *a, void *b)
{
        procjiffy_type *one, *two;
        if ((a == NULL) || (b == NULL))
                return 0;
        one = (procjiffy_type *) a;
        two = (procjiffy_type *) b;
        return (two->pid < one->pid);
}

int machine_get_jiffies(Driver *drvthis, LinkedList *procs, struct timeval *taken){
/* this was largely copied from machine_Linux.c-int machine_get_procs(LinkedList *procs) */ 
  // Much of this code was ripped from "gmemusage"
  char buf[128];
  DIR *proc;
  FILE *StatFile;
  struct dirent *procdir;
  PrivateData * p = drvthis->private_data;

  char  procName[19], dummychar;
  /* procName has () envelope, so 16+2 + \0 */
  int pid, dummyint ; // , ppid, pgrp;
  unsigned int dummyui;
  unsigned long  dummyul;
  unsigned long int utime, stime; // cutime, cstime, num_threads, singlesum, multisum, grandsum;
  long int num_threads;
  //singlesum=multisum=grandsum=0; 
  if((proc = opendir("/proc")) == NULL)
  {
    report(RPT_CRIT, "%s unable to open /proc.", __FUNCTION__);
    return(-1);
  }
  gettimeofday (taken, NULL); /* try for accuracy : when proc was read */
  while((procdir = readdir(proc)))
  {
    if(!index("1234567890", procdir->d_name[0]))
      continue;
    sprintf(buf, "/proc/%s/stat", procdir->d_name);
    if((StatFile = fopen(buf, "r")) == NULL)
    {
      // Not a serious error; process has finished before we could
      // examine it:
      continue;
    }
    utime= stime =0; // cutime = cstime=0;
    //utime= stime = 0 cutime = cstime=0;
    /* does fscanf gaurantee no buffer overunns ?...  put length spec in format  */
/* 
cstime %ld %ld  (num_threads %ld) 
*/
    fscanf(StatFile,
//      "%d %18s %c %d %d %d %d %d %u %lu %lu %lu %lu %lu %lu", //  %lu %lu",  /* utime stime cutime cstime */
//      "%d %18s %c %d %d %d %d %d %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu",  /* utime stime cutime cstime  . . num_threads*/
      "%d %18s %c %d %d %d %d %d %u %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %ld",  /* utime stime cutime cstime  . . num_threads*/
            &pid, procName, &dummychar, 
      &dummyint, &dummyint, &dummyint, &dummyint, &dummyint,
//      &dummyint, &ppid, &pgrp, &dummyint, &dummyint,
      &dummyui, &dummyul, &dummyul, &dummyul, &dummyul,
//      &utime, &stime); //, &cutime, &cstime); 
      &utime, &stime, &dummyul, &dummyul, &dummyul, &dummyul, &num_threads); 
    fclose(StatFile);
    LL_Rewind(procs);
    p->pjt = malloc(sizeof(procjiffy_type));
    //p->pjt = calloc(1, sizeof(procjiffy_type));
    char *tmpstart, *tmpend;
    tmpstart = strchr(procName, '(') + 1;
    tmpend = strrchr(tmpstart, ')');
    /* gnome-terminal,0.64,gnome-cups-icon�!,0.04,cupsd here ? */
    snprintf(p->pjt->name, tmpend-tmpstart+1 > 16 ? 16 : tmpend-tmpstart+1, tmpstart);
    //int counter;
    //for (counter=0; counter < 16; counter++){
    //  if (counter == strlen (p->pjt->name)) break;
    //  if ( isalnum(p->pjt->name[counter]) || ispunct(p->pjt->name[counter]) ) continue;
    //  report(RPT_CRIT, "%s is %s funky at %d ?", __FUNCTION__, p->pjt->name, counter+1); 
    //}
    p->pjt->utime=utime;
    p->pjt->stime=stime;
    p->pjt->pid=pid;
    p->pjt->num_threads=num_threads;
/*
if (num_threads> 1 ) { 
    multisum +=utime+ stime  + cutime+ cstime;
    report(RPT_ERR, "jiffy % 16s  %2d thrds %d %5lu + %5lu + %5lu + %5lu = %5lu",
        p->pjt->name, num_threads,
        pid, utime, stime, cutime, cstime,
        utime+ stime+ cutime+ cstime);
// WTF only one entry by this, but 5 in ls /proc 
} else {
  singlesum +=utime+ stime+ cutime+ cstime;
}
*/
    // TODO:  Check for errors here?
    if ((procs) && (p->pjt)) LL_Push(procs, (void *) p->pjt);
    p->pjt=NULL;
    //  free(p->pjt) happens when  linked list is freed, unless inturrupted b4 this push then in close 
  } /* end while((procdir = readdir(proc))) read stas for each current pid */
/*
  report(RPT_ERR, "siglesum %5lu  multisum %5lu ",
    singlesum, multisum );
*/
  closedir(proc);
  return(1);
}
/*
 * called once from ult_init to create topnum and topname LL
 * and read in first set of jiffies[] without computing an elapse
*/
void store_tops_init(Driver *drvthis){
  PrivateData * p = drvthis->private_data;
  p->newj=0; /* FIXME just use 0 in init, no need to derefrence p->newj over and over */
  int i;
  /* FIXME: where are other jiffies data initialized? could put that here. */

  gettimeofday(&p->tdid_top, NULL);
  p->jiffies[0] = LL_new();
  machine_get_jiffies(drvthis, p->jiffies[0], &p->delta_t[0]);
  // sort jiffies[0] by pid: only sort performed on any set
  LL_Rewind(p->jiffies[0]);
  LL_Sort(p->jiffies[0], sort_jiffpid);
  LL_Rewind(p->jiffies[0]);
  p->named_jiffs[0]=malloc(sizeof (unsigned long int*) * p->topname );
  p->named_jiffs[1]=malloc(sizeof (unsigned long int*) * p->topname );
  for (i=0; i < p->topname; i++){
    p->named_jiffs[0][i]=malloc(sizeof(unsigned long int));
    p->named_jiffs[1][i]=malloc(sizeof(unsigned long int));
    *p->named_jiffs[0][i]=*p->named_jiffs[1][i]=0;
  }
  p->cpus=0;
#ifdef _SC_NPROCESSORS_ONLN
  if(((p->cpus=sysconf(_SC_NPROCESSORS_ONLN))!=-1) && (p->cpus > 0)){
    p->dt_to_clktk=p->cpus*sysconf(_SC_CLK_TCK)/100;
  }
  else {
    report(RPT_ERR, "%s sysconf(_SC_NPROCESSORS_ONLN) failed.  Assumming 1 cpu." ,__FUNCTION__);
    p->dt_to_clktk=sysconf(_SC_CLK_TCK)/100;
    p->cpus=1;
  }
#else  /* _SC_NPROCESSORS_ONLN */
  p->dt_to_clktk=sysconf(_SC_CLK_TCK)/100;
  report(RPT_ERR, "%s _SC_NPROCESSORS_ONLN undefined.  Assumming 1 cpu." ,__FUNCTION__);
  p->cpus=1;
#endif
    report(RPT_INFO, "%s detected %d cpus, %d clock ticks / second.",
      __FUNCTION__, p->cpus, p->dt_to_clktk * 100);
  machine_get_jiffies(drvthis, p->jiffies[0], &p->delta_t[0]);

  LL_Rewind(p->jiffies[0]);
  LL_Sort(p->jiffies[0], sort_jiffpid);
  LL_Rewind(p->jiffies[0]);

/* countofnamedjiffs is a diagnostic curiosity. like how much cpu spent on 'xterm'(s) */
//  p->countofnamedjiffs=calloc(1, sizeof (int) * p->topname);
  /* FIXME ! why is calloc not zeroing the values ?  I'm failing on ptr arithmetic */
//  for (i=0; i < p->topname; i++){ // for each name tracked
//    p->countofnamedjiffs[i]=0;
//  }

  double uptime =0;
  double idle =0;
  machine_get_uptime(&uptime, &idle);
  p->elapse=(float)uptime*100;
    report(RPT_CRIT, "%s elapse %f seconds  (uptime = %f) = %.0f clock ticks.",
      __FUNCTION__, (float)p->elapse / 100, uptime, (float)p->elapse *  p->dt_to_clktk );
  do {
    /* nothing is being computed on init pass */
    procjiffy_type *fresh = LL_Get(p->jiffies[0]);
    /* garbage in unterminated ->name was coming from machine_get_jiffies read overrun
       unternimated name copied into topnum or topname overwrites  */
    if (p->topname >0  ) { // if doing top by name
      for (i=0; i < p->topname; i++){ // for each name tracked
        if (strlen (fresh->name)+1 < 16) fresh->name[15]='\0';
        else fresh->name[15]='\0';
//        if (i==0) report(RPT_CRIT, "before %5d %20s (%d) + %5lu = %5lu",
//            fresh->pid, fresh->name, p->countofnamedjiffs[i],
//            fresh->utime+fresh->stime, *p->named_jiffs[0][i]);
        if (!strncmp(p->top_name[i]->name, fresh->name, strlen(p->top_name[i]->name))){
          // how to handle jiffies wrap here ?
          *p->named_jiffs[0][i]+=fresh->utime+fresh->stime;
          // is sum of all procs by that name/wildcard
          // can wrap -> time data type or no ?
//          p->countofnamedjiffs[i]++;
//          if (i==0) report(RPT_CRIT, "after %5d %20s (%d) + %5lu = %5lu",
//            fresh->pid, fresh->name, p->countofnamedjiffs[i],
//            fresh->utime+fresh->stime, *p->named_jiffs[0][i]);
        } // end name matched
      } // end for each tracked name
    }  // end there are topnames
//    report(RPT_CRIT, "%s 1stread %d %s %lu %lu",
//        __FUNCTION__, fresh->pid, fresh->name, fresh->utime, fresh->stime);
  } while (LL_Next (p->jiffies[p->newj]) != -1 ); /* done picking topnames from jiffies[0] */
  LL_Rewind(p->jiffies[0]);
  LL_Sort(p->jiffies[0], sort_jiffies);  // sort by cpu load
  LL_Rewind(p->jiffies[0]);
  for (i = 0; i < p->topnum ; i++) {
    /* no div by zero if elapse is 0 */
    procjiffy_type *pj = LL_Get(p->jiffies[0]);
    debug(RPT_CLIUP, "%s Top % 2d % 16s, % 8lu % 8lu =% 2.6f", __FUNCTION__,
        i, pj->name, pj->utime, pj->stime,
        ( (float)pj->utime+(float)pj->stime) /((float)p->elapse/100));
    p->top_num[i]->val=(float)(pj->utime+pj->stime)/((float)p->elapse*p->dt_to_clktk )*100;
    if (p->top_num[i]->name) {
      free (p->top_num[i]->name);
    }
    p->top_num[i]->name=strdup(pj->name);
    LL_Next (p->jiffies[0]);
  }
  if (p->topname!=0) { // if doing top by name
    for (i=0; i < p->topname; i++){ // for each name tracked
      p->top_name[i]->val = (float)(*p->named_jiffs[0][i])/((float)p->elapse * p->dt_to_clktk) * 100;
    } /*  end for each named top */
  } /* end do top by name */
  p->newj=1-p->newj; // new becomes old

//  if (p->countofnamedjiffs) {
//    free(p->countofnamedjiffs);
//    p->countofnamedjiffs=NULL;
//  }
} /* end store_tops_init */



void store_tops(Driver *drvthis){
/* L. T. says in the headers that jiffies wrap artificially at 5 minutes to
 * aid in dbg that wrap.  .. Cope with that or is it hidden from this context ?
 */
  PrivateData * p = drvthis->private_data;
             /* p - > elapse is set in this function only */
             /* p - > elapse scale: time_diff ( ) / 10000 */
             /*          = 1 / 100ths of a sec.           */
             /* p - > elapse * dt_to_clktk = ticks | jiffies elapsed */
  int i;
#ifdef TOPCOMPARE
  procjiffy_type load_sum;
  load_sum.utime=0;
  load_sum.stime=0;
#endif //TOPCOMPARE
    struct timeval now;
    gettimeofday(&now, NULL);
    /* do not do all this too frequently */
/* sync this value with one in store_load  - no separate cycles and measures ?*/
  long since=time_diff(drvthis, p->tdid_top, now);
  if (since  <1000000) {
    report (RPT_CLIUP, "%s return on time_diff < 1000000 %d:%d",
       __FUNCTION__, (int)(since/1000000), since - (int)(since/1000000)*1000000);
    return;
  }
  else {
    report (RPT_CLIUP, "%s time_diff OK %d:%d",
      __FUNCTION__, (int)(since/1000000), since - (int)(since/1000000)*1000000 ) ;
  }
  p->tdid_top=now;

  p->jiffies[p->newj] = LL_new();
    machine_get_jiffies(drvthis, p->jiffies[p->newj], &p->delta_t[p->newj]);
    // sort jiffies[p->newj] by pid: only sort performed on any set
    LL_Rewind(p->jiffies[p->newj]);
    LL_Sort(p->jiffies[p->newj], sort_jiffpid); 
    LL_Rewind(p->jiffies[p->newj]);
  // create jiffies[2] for elapsed values
  p->jiffies[2] = LL_new();
  p->countofnamedjiffs=calloc(1, sizeof (int) * p->topname);
  /* FIXME ! why is calloc not zeroing the values ?  I'm failing on ptr arithmetic */
  for (i=0; i < p->topname; i++){ // for each name tracked
    p->countofnamedjiffs[i]=0;
  }
  LL_Rewind(p->jiffies[1-p->newj]);
  p->elapse=time_diff(drvthis,p->delta_t[1-p->newj], p->delta_t[p->newj])/10000;
//    report(RPT_CRIT, "%s elapse %f seconds = %.0f clock ticks.",
//      __FUNCTION__, (float)p->elapse / 100, (float)p->elapse *  p->dt_to_clktk );
    do {
      procjiffy_type *computed;
      computed = calloc(1, sizeof(procjiffy_type));
      /* this is pushed into LL and freed when LL Destroy the LL it is in */
      procjiffy_type *fresh = LL_Get(p->jiffies[p->newj]);
      computed->pid=fresh->pid;
      /* garbage in -> name was coming from get_jiffies read overrun 
         fresh->name comes from LL_Get(p->jiffies[p->newj] 3 lines previous 
         jiffies[p->newj] filled by machine_get_jiffies(drvthis, p->jiffies[p->newj])
         in init function 
      */
      memcpy(computed->name, fresh->name, strlen (fresh->name)+1 < 16 ? strlen (fresh->name)+1 : 16 );
      if (p->topname >0  ) { // if doing top by name
        for (i=0; i < p->topname; i++){ // for each name tracked 
//          if (i==0) report(RPT_CRIT, "before %5d %20s (%d) + %5lu = %5lu",
//              fresh->pid, fresh->name, p->countofnamedjiffs[i],
//              fresh->utime+fresh->stime, *p->named_jiffs[p->newj][i]); 
          if (!strncmp(p->top_name[i]->name, fresh->name, strlen(p->top_name[i]->name))){
            // how to handle jiffies wrap here ?
            *p->named_jiffs[p->newj][i]+=fresh->utime+fresh->stime; 
              // is sum of all procs by that name/wildcard
              // can wrap -> time data type or no ?

            p->countofnamedjiffs[i]++;
//            if (i==0) report(RPT_CRIT, "after %5d %20s (%d) + %5lu = %5lu",
//              fresh->pid, fresh->name, p->countofnamedjiffs[i],
//              fresh->utime+fresh->stime, *p->named_jiffs[p->newj][i]); 
          } // end name matched 
        } // end for each tracked name 
      }  // end there are topnames      
      computed->num_threads=fresh->num_threads;
        procjiffy_type *test;
        int found=0;
        do { 
          test = LL_Get(p->jiffies[1-p->newj]);
          if (test->pid == computed->pid){
              // when does jiffies wrap ?
              // /usr/include/linux/time.h:172: * The maximum jiffie value is (MAX_INT >> 1).
              // limits. h:#  define INT_MAX       2147483647
              // 2147483647 >> 1 == 1073741823 ?
              // 1 sec  =       100 jiff
              // 1 min  =     6 000 jiff
              // 1 hour =   360 000 jiff
              // 1 day  = 8 640 000 jiff
              // 2147483647 - 248 * 8 640 000 = 2147483647 - 2142720000 = 4766647 remainder 
              //    4766647 -  13 *   360 000 =    4766647 -    4680000 =   86647 remainder
              //      86647 -  14 *     6 000 =      86647 -      84000 =    2647 remainder
              //       2647 =  26.47 sec
              // 248 days, 13 hours, 14 min, 26.47 sec
              // >> 1 so half that? 
            computed->utime= fresh->utime-test->utime;
            computed->stime= fresh->stime-test->stime;
            // ->?time are unsigned long int, so if they wrapped in above, throw out 
            // wrapped unsigned long int results & start over
            // fixme UINT_MAX ULINT_MAX ? 
            if (fresh->utime < test->utime)  {
              computed->utime= fresh->utime - test->utime + (INT_MAX >> 1);
            }
            if (fresh->stime < test->stime) {
              computed->stime= fresh->stime - test->stime + (INT_MAX >> 1);
            }
            found=1;
            break;
          } // end test->pid == computed->pid 
        } while (LL_Next (p->jiffies[1-p->newj]) !=-1); //end seach old list for this pid 
        if (found==0) {  // did not find a match in previous set
            computed->utime =fresh->utime;
            computed->stime =fresh->stime;
            // search jiffies[1-p->newj] from beginning next time
             LL_Rewind(p->jiffies[1-p->newj]);
        }

      /* computed delta values are all filled in now */ 
      // could sum this jiffies[p->newj] into load totalizer here
      // see how it compares to values from machine get load
#ifdef TOPCOMPARE
      load_sum.utime=load_sum.utime+computed->utime;
      load_sum.stime=load_sum.stime+computed->stime;
#endif // TOPCOMPARE        
      if ((p->jiffies[2]) && (computed)) {
        LL_Push (p->jiffies[2], (void *)computed);

/* 
//        block reports on /proc/[LCDd] details, for investigating top values 
//        track to verify with top run in shell 
//        2009 03 28 read vaues appear to be the same as top itself is reading
//        either a buisy-wait or race in my driver ... 
//           or kernel is exagerrating pthreads cpu use ?
//           I contributed to debain bugreports April 2009
//           http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=441166
//           ^^^ was a 2.6.26 kern issue, change in /proc writing, fixed in 2.6.32
//                introduced with changing /proc/[pid]/stat data formats and|or
//                KVM / hypervisor accounting features 2.6.18 - > 2.6.26
//           browsing today, is bug 641905 similar?    
        if (!strncmp(computed->name, "LCDd", 4) 
            || !strncmp(computed->name, "Xorg", 4) ) {
          report(RPT_ERR, "%5d %16s, %2d u %6lu + s %6lu = %6lu  /  %6lu =  %3.2f ",
            computed->pid, computed->name, computed->num_threads,
            computed->utime, computed->stime,
            computed->utime + computed->stime,
            p->elapse * dt_to_clktk / 100, // check the /100   
//p -> load . total is from get _ load, so it spans a different time slice than
//p - > elapse, which is measured here  
            (float)(computed->utime+computed->stime)/
                    ((float)p->elapse*dt_to_clktk );
        } 
*/
      } // end there is a list to read && this one is good
    } while (LL_Next (p->jiffies[p->newj]) != -1 );
      // end feshly read list into jiffies[2]

/*
       /proc/[number]/stat does not include nice by [number]
       noted: LCDd stands out in ps-eLf by LWP != PID 
furniture:/usr/src/lcdproc-CVS-current-20081129# cat /proc/14825/stat | cut -d" "  -f1,5,6
14825 14821 14821
       1 is parent ? 5 and 6 are what ?
furniture:/usr/src/lcdproc-CVS-current-20081129# ps -eLf | grep -e "LCDd" -e "PID"
UID        PID  PPID   LWP  C NLWP STIME TTY          TIME CMD
nobody   14821     1 14821  0    5 20:40 ?        00:00:02 ./server/LCDd -c ./LCDd.conf
nobody   14821     1 14822  0    5 20:40 ?        00:00:00 ./server/LCDd -c ./LCDd.conf
nobody   14821     1 14823  8    5 20:40 ?        00:03:12 ./server/LCDd -c ./LCDd.conf
nobody   14821     1 14824  0    5 20:40 ?        00:00:04 ./server/LCDd -c ./LCDd.conf
nobody   14821     1 14825  0    5 20:40 ?        00:00:13 ./server/LCDd -c ./LCDd.conf
root     14827  3339 14827  0    1 20:40 pts/1    00:00:00 grep LCDd

              The fields, in order, with their proper scanf(3) format specifiers, are:
1             pid %d      The process ID.
              comm %s     The  filename  of  the executable, in parentheses.  This is visible whether or not the
                          executable is swapped out.
              state %c    One character from the string "RSDZTW" where R is running, S is sleeping in an  inter‐
                          ruptible wait, D is waiting in uninterruptible disk sleep, Z is zombie, T is traced or
                          stopped (on a signal), and W is paging.
5             ppid %d     The PID of the parent.
6             pgrp %d     The process group ID of the process.
*/
 
   
#ifdef TOPCOMPARE
//      where is p->load_t set ? how recently  A: in store load
//       p->load_t includes nice numbers , my load_sum does not 
    float myload =  (float)(load_sum.utime+ load_sum.stime ) / 
                    ((float)p->elapse *  dt_to_clktk * 100);
// 2012 03 23 ...  
// a comparison of load from store load, and load from store tops 
//    int load_diff = myload *100 - p->load_t* 100;
//    if ((p->load_t > 99) || (abs(load_diff)  > 100 )){ 
    if (1){ 
      report(RPT_ERR, "%7.2f s %5lu tks = %3.2f%% load;  %3.2f%% by load_t:  %3.4f%% difference",
        (float)p->elapse/100,   // elapse 
        load_sum.utime + load_sum.stime,  // used ticks total load_sum used tallied here & now
        myload*10000, // = %3.2f by s
        p->load_t, 
        myload*10000 - p->load_t  // load by per second, difference of those 2 ... 
        );
    }
#endif //TOPCOMPARE
    //  store  jiffies[2]
    LL_Rewind(p->jiffies[2]);
    LL_Sort(p->jiffies[2], sort_jiffies);  // sort by cpu load
    LL_Rewind(p->jiffies[2]);
    if ( p->elapse != 0 ) for (i = 0; i < p->topnum ; i++) {
      /* no div by zero if elapse is 0 */
      procjiffy_type *pj = LL_Get(p->jiffies[2]);
      debug(RPT_CLIUP, "%s Top % 2d % 16s, % 8lu % 8lu =% 2.6f", __FUNCTION__, 
          i, pj->name, pj->utime, pj->stime, 
          ( (float)pj->utime+(float)pj->stime) /((float)p->elapse/100));
      p->top_num[i]->val=(float)(pj->utime+pj->stime)/((float)p->elapse*p->dt_to_clktk )*100;
      if (p->top_num[i]->name) {
        free (p->top_num[i]->name);
      }
      p->top_num[i]->name=strdup(pj->name);
      LL_Next (p->jiffies[2]);
    }
    if ((p->topname!=0)  && ( p->elapse != 0 ) ){ // if doing top by name
      /* & no div by zero if elapse is 0 */
      for (i=0; i < p->topname; i++){ // for each name tracked
/* if top_name[i] stopped, *p->named_jiffs[p->newj][i] will be zero */
/* instead of something > *p->named_jiffs[1-p->newj][i]        */
/* so *p->named_jiffs[newj][i] - *p->named_jiffs[1-newj][i] will be negative */
/* except it is type ul, so it is a huge positive.      */
/* if new is zero, then zero the old one. Process likely restarted ?  */
          if( ( *p->named_jiffs[p->newj][i] == 0 )  && ( *p->named_jiffs[1-p->newj][i] != 0 ) ){
            report(RPT_CRIT, "               %20s (%d) new value = zero, so old val %5lu zeroed, proc restarted?",
              p->top_name[i]->name, p->countofnamedjiffs[i], *p->named_jiffs[1-p->newj][i] );
            *p->named_jiffs[1-p->newj][i] = 0;
          }
          /* if the new value is greater than all the jiffies in this elapse, and the old  is zero, then zero the old one */
          /* *p->named_jiffs[1-p->newj][i] is zero, *p->named_jiffs[p->newj][i] is huge */
/*
          if (( *p->named_jiffs[p->newj][i] > ((float)p->elapse * p->dt_to_clktk) ) && (*p->named_jiffs[1-p->newj][i]==0))  {
            report(RPT_CRIT, "               %20s (%d) new value %5lu > , and old %5lu is zero? , so old val should be zeroed, how?, FIXME?",  
              p->top_name[i]->name, p->countofnamedjiffs[i], *p->named_jiffs[p->newj][i], *p->named_jiffs[1-p->newj][i] );
            // *p->named_jiffs[p->newj][i] =0;
          } 
*/        
        p->top_name[i]->val=(float)(*p->named_jiffs[p->newj][i]-
            *p->named_jiffs[1-p->newj][i]) / 
            ((float)p->elapse * p->dt_to_clktk) * 100;
/*            
          if ( p->top_name[i]->val > 100) { // if absurd value greater than 100 ...
            report(RPT_CRIT, "assigned ->val %20s (%d) = (%5lu - %5lu = %5lu ) / % 2.6f = % 2.6f",  
              p->top_name[i]->name, p->countofnamedjiffs[i],
              *p->named_jiffs[p->newj][i], *p->named_jiffs[1-p->newj][i],
              *p->named_jiffs[p->newj][i] - *p->named_jiffs[1-p->newj][i],
              (float)p->elapse * p->dt_to_clktk , 
              p->top_name[i]->val);
          }   
*/        
// use to track calcs for topname[0] , topname1 in conf
/*            if (i==0) report(RPT_CRIT, "assigned ->val %20s (%d) = (%5lu - %5lu = %5lu ) / % 2.6f = % 2.6f",
              p->top_name[i]->name, p->countofnamedjiffs[i],
              *p->named_jiffs[p->newj][i], *p->named_jiffs[1-p->newj][i], 
              *p->named_jiffs[p->newj][i] - *p->named_jiffs[1-p->newj][i],
              (float)p->elapse * dt_to_clktk , 
              p->top_name[i]->val);
*/
        /* zero summ of jiffs here */
        *p->named_jiffs[1-p->newj][i]=0;
        p->countofnamedjiffs[i]=0;
      }
    } /* end do top by name */

    p->newj=1-p->newj; // new becomes old
//    report(RPT_CRIT, "%s swaped old and new: p->newj=%d, p->jiffies[p->newj]=%d ; 1-p->newj=%d, p->jiffies[1-p->newj]=%d",  
//      __FUNCTION__, p->newj, p->jiffies[p->newj], 1-p->newj, p->jiffies[1-p->newj]);
    // Now clean up the old one & jiffies[2] ...
      LL_Rewind(p->jiffies[p->newj]);
      do {
        procjiffy_type *tofree = (procjiffy_type *) LL_Get(p->jiffies[p->newj]);
        if (tofree != NULL) { free(tofree); }
      } while (LL_Next(p->jiffies[p->newj]) == 0);
      LL_Destroy(p->jiffies[p->newj]);
      p->jiffies[p->newj]=NULL;
//      report(RPT_CRIT, "%s deleted old : p->newj=%d, p->jiffies[p->newj]=%d ; 1-p->newj=%d, p->jiffies[1-p->newj]=%d", 
//        __FUNCTION__, p->newj, p->jiffies[p->newj], 1-p->newj, p->jiffies[1-p->newj]);

    LL_Rewind(p->jiffies[2]);
    do {
      procjiffy_type *tofree=(procjiffy_type *) LL_Get(p->jiffies[2]);
      if (tofree != NULL) { free(tofree); }
    } while (LL_Next(p->jiffies[2]) == 0);
    LL_Destroy(p->jiffies[2]);
    p->jiffies[2]=NULL;
  if (p->countofnamedjiffs) {
    free(p->countofnamedjiffs);
    p->countofnamedjiffs=NULL;
  }
} /* end of store tops */


static int
sort_jiffies (void *a, void *b)
{
        procjiffy_type *one, *two;
        if ((a == NULL) || (b == NULL))
                return 0;
        one = (procjiffy_type *) a;
        two = (procjiffy_type *) b;

  // by pid or by pid and children ?
        return (two->utime + two->stime  > one->utime + one->stime );
}
#endif // DO_ULT

#ifdef DO_LM
int 
lm_init(Driver *drvthis){
    PrivateData * p = drvthis->private_data;
  p->lmconf=NULL;
  // how big do these need to be ?
  char buf[128], target[128];
  int tmp;
  snprintf(buf, sizeof(buf), drvthis->config_get_string
      (drvthis->name,"lmconf", 0, NULL));
  if (strlen(buf)==0){
    report(RPT_ERR, "%s did not read a lmconf.", __FUNCTION__);
  }
  else {
    p->lmconf = (char *)malloc(strlen(buf)+1);
    sprintf(p->lmconf, buf);
    report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
        "%s read lmconf=%s", __FUNCTION__, p->lmconf);
  }
  if (p->lmconf !=NULL) {
    if((p->lmfile = fopen(p->lmconf, "r")) == NULL)
    {
      report(RPT_ERR, "%s failed to open %s", __FUNCTION__, p->lmconf);
      free (p->lmconf);
      p->lmconf=NULL;
      p->lmnum=0;
      p->lmfile=NULL;
    }
    else{
      report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
          "%s opened %s", __FUNCTION__, p->lmconf);
      if (sensors_init(p->lmfile) == 0){
        report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
            "%s lm sensors_init (%s) good.",
            __FUNCTION__, p->lmconf);
      }
      else {
        report(RPT_ERR, "%s lm sensors_init(%s) failed.",
            __FUNCTION__, p->lmconf);
        free (p->lmconf);
        p->lmconf=NULL;
        p->lmnum=0;
        p->lmfile=NULL;
        return (0);
      }
    }
  } // end try to read file & open

  p->lm=malloc(sizeof( CF_LM* ) *MAX_LM);
  for (tmp=0; tmp < MAX_LM; tmp++){ 
    p->lm[tmp]=NULL; 
    sprintf(target, "lmLabel%d", tmp+1);
    snprintf(buf, sizeof(buf), 
      drvthis->config_get_string
      (drvthis->name,target, 0, NULL));
    if(strlen(buf)==0) {
      report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
        "%s %s not found.", __FUNCTION__, target );
      continue;
    } else {
      p->lm[tmp]=malloc(sizeof( CF_LM ));
/*      report(RPT_CRIT,
        "%s malloc p->lm[%d].", __FUNCTION__, tmp );
*/      p->lm[tmp]->lmfeatu=NULL;
      p->lm[tmp]->chipname=NULL;
      p->lm[tmp]->lmlabel=NULL;
      p->lm[tmp]->chipn=NULL;
      p->lm[tmp]->featn=NULL;
      p->lm[tmp]->lmfeatu=(char *)malloc(strlen(buf)+1);
      sprintf(p->lm[tmp]->lmfeatu, buf);
      sprintf(target, "lmChip%d", tmp+1);  
      snprintf(buf, sizeof(buf), 
      drvthis->config_get_string (drvthis->name,target, 0, NULL));
      if (strlen(buf)>0){
        p->lm[tmp]->chipname=(char *)malloc(strlen(buf)+1);
        sprintf(p->lm[tmp]->chipname, buf);
      }
#ifdef CLIENT_CHILD
      sprintf(target, "lm%d_Lcd", tmp+1);
      p->lm[tmp]->lmline=NULL; 
      p->lm[tmp]->lmlcd=drvthis->config_get_int(drvthis->name, target, 0, 0);
#endif //CLIENT_CHILD
    }
    report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, "%s read lm%d=%s,%s",
        __FUNCTION__, tmp+1,  p->lm[tmp]->chipname, p->lm[tmp]->lmfeatu);
  } // end good file open and each lm-sensors name / chip pair 
  if (!buildlm(drvthis)){
    report(RPT_ERR, "%s could not find requested lm sensors, exit.", __FUNCTION__);
    exit(EXIT_FAILURE);
  }
  return(1);
}

/*
 * for each chip
 *   for each feature
 *      for each requested name {
 *        see if it is a match & set pointer
 *        }
 * return 1 on sucess, 0 otherwise
*/
int buildlm(Driver *drvthis){
  PrivateData * p = drvthis->private_data;
  const struct sensors_chip_name *achip=NULL;
  int chip_num=0;
  double value;
  int a,b;
  char full_desig[128];
  for (a=0; a < 128; a++) full_desig[a]='\0';
  a=b=0;
    report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
      "%s SENSORS_API_VERSION= %x, libsensors_version=%s", 
      __FUNCTION__, SENSORS_API_VERSION, libsensors_version);
    if (SENSORS_API_VERSION !=0x432) 
       report(RPT_NOTICE, "%s SENSORS_API_VERSION is not author's 0x432", __FUNCTION__);
    if (strncmp(libsensors_version, "3.3.2", 5)) 
       report(RPT_NOTICE, "%s libsensors_version is not author's 3.3.2", __FUNCTION__);

  do { /* for each lm chip detected */
    a=0;
    achip=sensors_get_detected_chips(NULL, &chip_num); // 3.0.2
    if (achip==NULL) continue; // will exit at end of while
    sensors_snprintf_chip_name(full_desig, 127, achip);
    report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
      "%s chip_num=%d : %s", __FUNCTION__, chip_num, full_desig);
    char *label=NULL;

    const struct sensors_feature *afeature=NULL;
    while ((afeature=sensors_get_features(achip, &a))) { // 3.0.2
      int subf_num=0;
      const struct sensors_subfeature *asubfeature=NULL;
      while ((asubfeature=sensors_get_all_subfeatures(achip, afeature, &subf_num) )) {
        if ( !(asubfeature->flags & SENSORS_MODE_R) ) continue; 
        if ( !(asubfeature->flags & SENSORS_COMPUTE_MAPPING) ) continue; 
        if (asubfeature->flags & SENSORS_MODE_W) continue; 
        label=sensors_get_label(achip, afeature);
        if ( 0!= sensors_get_value( achip, subf_num -1, &value)) {
           report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
               "%s Can't get %d %s %s %s %s data!\n", __FUNCTION__, 
               chip_num, full_desig, afeature->name, label, asubfeature->name);
           free (label);
           continue;
        }
        /* debugger's print
        report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
            "%s found %d %s, bus %d, addr %d, label %s, subf# %d, subname %s, subflags %d, = %f " ,
                __FUNCTION__, chip_num, full_desig, achip->bus, achip->addr, label, 
                asubfeature->number, asubfeature->name, asubfeature->flags, value);  
        */
        /* users's print */
        report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
            "%s found lmChip %s, feature %s, = %f " ,
                __FUNCTION__, full_desig, label, 
                 value);  

        int one_lm;
	for (one_lm=0; one_lm < MAX_LM; one_lm++) { // check each LCDd.conf entry 
          /* if p->lm[one_lm] is not full enough, skip it */
          if ((p->lm[one_lm]==NULL) || ((p->lm[one_lm]->lmfeatu==NULL) && (p->lm[one_lm]->chipname==NULL))){
            //report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
            //    "%s lm[%d] is not full enough to check, skipping.", __FUNCTION__, one_lm+1);
            continue;
          }
          /* if p->lm[one_lm] has been assigned , skip it */
	  if ((p->lm[one_lm]->lmlabel!=NULL) || (p->lm[one_lm]->chipn!=NULL) || (p->lm[one_lm]->featn!=NULL)){
            //report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
            //    "%s lm[%d] has been assigned, skipping.", __FUNCTION__, one_lm+1);
            continue; 
	  }
          if   (!strncmp(label,p->lm[one_lm]->lmfeatu, strlen(p->lm[one_lm]->lmfeatu))){ /* if the feature matcihes */
            if  ((p->lm[one_lm]->chipname!=NULL) /* and there is a chip name that does not match */ 
                     && (0 != strncmp(p->lm[one_lm]->chipname, full_desig, strlen(p->lm[one_lm]->chipname)-1))){
              report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
                  "%s lm[%d] feature match (%s = %s) chipname missmatch (%s != %s).", __FUNCTION__, 
                  one_lm+1, p->lm[one_lm]->lmfeatu, label, p->lm[one_lm]->chipname, full_desig);
              continue; 
	    } 
            // is a good read match 
            p->lm[one_lm]->lmlabel=strdup(label);
            /* strdup so original can be freed whenever */  
            p->lm[one_lm]->chipn=achip;
            p->lm[one_lm]->featn=afeature;
            if (p->lm[one_lm]->chipname==NULL) p->lm[one_lm]->chipname=strdup(achip->prefix);
            p->lm[one_lm]->subfnum=asubfeature->number;
            /* debugger's print
            report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
                "%s match %d, %s, bus %d, addr %d, label %s, subf# %d, subname %s, subflags %d to lm%d = %f" ,
                __FUNCTION__, chip_num, full_desig, achip->bus, achip->addr, label, asubfeature->number, 
                asubfeature->name, asubfeature->flags, one_lm+1, value );  
            */
            /* users's print */
            report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE, 
                "%s match %s, label %s, lm%d = %f" ,
                __FUNCTION__, full_desig,  label, 
                one_lm+1, value );  
            update_lm(drvthis, one_lm);
	    p->lmnum++;
          } // end match
        } // end check each p->lm[one_lm] for match
        free (label);
      }// end read all subfeatures unmatched
    } // end get each afeature 
  } while (achip !=NULL); // end each chip 

  //  sort into a tidy for ( 0 ; < lmnum; ++ ) & run some checks 
  for (a=0; a < MAX_LM; a++) {
    if  ( (p->lm[a] != NULL) && ((p->lm[a]->lmfeatu != NULL) || (p->lm[a]->chipname != NULL) )) {
      /* if a config was suggested */
      if ((p->lm[a]->chipn==NULL) || (p->lm[a]->featn==NULL) ) {
        /* but nothing matched */
        report(RPT_ERR, "%s could not find requested lmChip%d %s lmLabel%d %s.",
          __FUNCTION__, a+1, p->lm[a]->chipname, a+1, p->lm[a]->lmfeatu);
        if (p->lm[a]->lmfeatu) free(p->lm[a]->lmfeatu);
        if (p->lm[a]->chipname) free (p->lm[a]->chipname);
        free (p->lm[a]);
        p->lm[a]=NULL;
     } 
    }


    if ( (p->lm[a]==NULL) || ( (p->lm[a]->chipn==NULL) && (p->lm[a]->featn==NULL))){
      for (b=a+1; b < MAX_LM; b++) {
        if ((p->lm[b]!=NULL) && (p->lm[b]->chipn!=NULL) && (p->lm[b]->featn!=NULL)){
          report(RPT_CLICRE, "%s lm%d is empty but %d is not - lm%d becomes lm%d",
           __FUNCTION__, a+1, b+1 ,b+1, a+1);
	   p->lm[a]=p->lm[b];
           //if (p->lm[b]->chipname) free (p->lm[b]->chipname);
           //if (p->lm[b]->lmfeatu) free (p->lm[b]->lmfeatu);
           //if (p->lm[b]->lmlabel) free (p->lm[b]->lmlabel);
           //free (p->lm[b]);
	   p->lm[b]=NULL;
	   break;
	} // end found first empty (b) in order
      } // end from current to MAX_LM
    } //end lm[a] is empty
    for (b=0; b!=a; b++) {
      if (    (p->lm[a]!=NULL) 
           && (p->lm[b]!=NULL)  
           && (p->lm[a]->chipn!=NULL) 
           && (p->lm[b]->chipn!=NULL) 
           && (p->lm[a]->chipn==p->lm[b]->chipn)
           && (p->lm[a]->featn!=NULL) 
           && (p->lm[b]->featn!=NULL)
	   && (p->lm[a]->featn==p->lm[b]->featn)
           && (p->lm[a]->lmlabel!=NULL)
           && (p->lm[b]->lmlabel!=NULL)
           && (strlen(p->lm[a]->lmlabel) == strlen(p->lm[b]->lmlabel) )
           && (0 != strncmp(p->lm[a]->lmlabel, p->lm[b]->lmlabel, strlen(p->lm[a]->lmlabel))) 
         ) {
        report(RPT_CLICRE, "%s lm%d (%s %s %s) is identical to lm%d (%s %s %s)",
	    __FUNCTION__, a+1,p->lm[a]->chipn->prefix, 
            p->lm[a]->featn->name, 
            p->lm[a]->lmlabel,
            b+1, p->lm[b]->chipn->prefix, 
            p->lm[b]->featn->name, p->lm[b]->lmlabel);
      }
    } // end check for duplicates
  } //end for each possible  MAX_LM
  report(RPT_CLICRE, "%s p->lmnum=%d", __FUNCTION__, p->lmnum);
  for (a=0; a < MAX_LM; a++) {
    if ((p->lm[a]!=NULL) && (p->lm[a]->chipn!=NULL) && (p->lm[a]->featn!=NULL) ){
      if ( (p->lm[a]->chipname!=NULL) && (p->lm[a]->lmlabel!=NULL)){
      report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
        "%s lm%d=%s, %s = %f", __FUNCTION__, 
         a+1, p->lm[a]->chipname, p->lm[a]->lmlabel, 
         (float)p->lm[a]->lm_abstracto/16);
      } else
        report(RPT_LMS <= RPT_CLICRE ? RPT_LMS : RPT_CLICRE,
        "%s lm%d=odd!", __FUNCTION__,
         a+1);

    }
  }
  //p->lm=realloc(p->lm, sizeof( CF_LM* ) *p->lmnum);
  return 1;
  // free() achip, afeature  ?
}

static void
update_lm (Driver *drvthis, unsigned char lmn)
{
  PrivateData * p = drvthis->private_data;
//  report(RPT_LMS, "%s update  p->lm[%d]->lm_abstracto %f:",
//      __FUNCTION__, lmn, (float)p->lm[lmn]->lm_abstracto/16  );
// could possibly run sensors_get_feature (p->height*8) times every second
// instead of once every p->syslogsensors seconds
  double value;
  int retval= sensors_get_value( p->lm[lmn]->chipn, p->lm[lmn]->subfnum, &value);
  if (retval !=0 ) 
    value=666;  // error on read
  p->lm[lmn]->lm_abstracto=(int)(value*16);
/*
  report(RPT_LMS <= RPT_CLIUP ? RPT_LMS : RPT_CLIUP,
      "%s lm%d prefix %s, bus %d, addr %d, subfnum %d, name %s= %f.",
      __FUNCTION__, lmn+1, p->lm[lmn]->chipn->prefix, p->lm[lmn]->chipn->bus, 
      p->lm[lmn]->chipn->addr, p->lm[lmn]->subfnum, 
      p->lm[lmn]->lmlabel, value );
*/
  report(RPT_LMS <= RPT_CLIUP ? RPT_LMS : RPT_CLIUP,
      "%s lm%d % 16s, % 16s = %f.",
      __FUNCTION__, lmn+1, p->lm[lmn]->chipn->prefix, 
      p->lm[lmn]->lmlabel, value );
} // end update_lm
#endif // DO_LM

void
client_loop(Driver * drvthis)
{
  PrivateData *p = drvthis->private_data;

  int Quit=0;
  struct timeval now;
  gettimeofday(&now,NULL);
#ifdef CLIENT_CHILD /* in client_loop setup */
  int i = 0;
  char buf[64]={0,0,0,0,0,0,0,0,
                0,0,0,0,0,0,0,0,
                0,0,0,0,0,0,0,0,
                0,0,0,0,0,0,0,0,
                0,0,0,0,0,0,0,0,
                0,0,0,0,0,0,0,0,
                0,0,0,0,0,0,0,0,
                0,0,0,0,0,0,0,0}; 
  char *argv[256];
  int argc, newtoken;
  int len;
  // lines and lcd_hgt are globals 
  // assigned values before main _ loop is called
  // (lines-1)/lcd_hgt+1 first computed in "cli_scrn creating LCDd x lines in y screens." 
  char foreground[(lines-1)/lcd_hgt+1];  // is a screen visible
  unsigned int updates[(lines-1)/lcd_hgt+1];  // # updates to current screen 
  char active_screen=0;  // which screen # is active
  char listen_ign=0;
  unsigned int want_success=0;

  for(i=0; i<(lines-1)/lcd_hgt+1; i++) 
    updates[i]=foreground[i]=0;
#endif
  // Main loop
  struct timespec nod = {1, 125000000}; 
//  nanosleep(&nod, NULL); // let temps get filled in b4 first text log 
  nod.tv_sec=0;
  while(!Quit)
  {
    gettimeofday(&now, NULL); // neeeded for both log and timing dump
// replace with dumb counter loop, fewer gettimeofday & time_diff
    if (p->log_sens!=0){
      if (time_diff(drvthis, p->tdid_tlog, now)>=
        p->log_sens * 1000000) {
        csv_log (drvthis);
        gettimeofday(&p->tdid_tlog, NULL);
      }
    } //end of if time to display 
#ifdef TIMING_REPORT
//  Nov  9 14:27:58 localhost LCDd: time diff test 2147.000000 seconds.
//  Nov  9 14:27:58 localhost LCDd: time diff test -2146.967000 seconds.
    //if ( (time_diff(drvthis, p->timing_mk, now)/1000) > 5*60*1000){ 
    //if ( (time_diff(drvthis, p->timing_mk, now)/1000) > 10*60*1000){ 
    if ( (time_diff(drvthis, p->timing_mk, now)/1000) > 30*60*1000){ 
      // /1000 one side _>_ * 1000 other side so no interger overflow
      // (calculated sloppily ... opportunistic use of existing time measures
      // report(RPT_CRIT, "---------TIMING COUNTER WRAP -----------");
      timing_dump(drvthis);
      clear_timing(drvthis);
    }
#endif
    nanosleep(&nod, NULL); // 1/8 sec screen update loop
#ifdef CLIENT_CHILD /* in client_loop loop */
    // Check for server input...
    len = sock_recv( cf_dsk, buf, sizeof(buf)-1); 

    // Handle server input...
    while(len > 0)
    {
/*
      static int maxlen=0;
      if (maxlen < len) {
        maxlen=len;
        debug(RPT_INFO, "%s new sock_recv max len %d/%d", __FUNCTION__, maxlen, sizeof(buf));
      }
*/

/* 
   buf is not print-safe terminated, print only len of buf 
*/
      buf[len+1]='\0'; // always sock_recv (,,-1) to null terminate received 
      debug(RPT_CLIUP, "%s tokenizer got %d :%s", __FUNCTION__, len, buf);
      // Now split the string into tokens...
      argc = 0;
      newtoken = 1;
      for(i = 0; i < len; i++)
      {
        switch(buf[i])
        {
          case ' ':
            newtoken = 1;
            buf[i] = 0;
          break;
          default:  // regular chars, keep tokenizing
            if(newtoken)
              argv[argc++] = buf + i;
            newtoken = 0;
          break;
          case '\0':
          case '\n':
            buf[i] = 0;
            if(argc > 0)
            { 
             debug(RPT_CLIUP, "%s Tokenizer line: argc=%d, argv[0]=%s \n", __FUNCTION__, argc, argv[0]);
              if(0 == strcmp(argv[0], "listen"))
                {
                active_screen=atoi((char *)argv[1]+6);
                debug(RPT_CLIUP, "%s Listen %d", __FUNCTION__, active_screen );
                foreground[(int)active_screen]=1;
                listen_ign=1;
              }
              else if(0 == strcmp(argv[0], "ignore"))
              {
                int ign_scn=atoi((char *)argv[1]+6);
                debug(RPT_CLIUP,"%s Ignore %d", __FUNCTION__, ign_scn);
                foreground[ign_scn]=0;
                updates[ign_scn]=0;
                listen_ign=0;
              }
              else if(0 == strncmp(argv[0], "cf_log", 6))
              {    // if "l/i screen l/i" is chopped up by a success ... ?
                // seems to never happen now
                int scrn_num=atoi((char *)argv[1]+6);
                debug(RPT_CLIUP,"%s foreground cf_log[%d]=%d",
                  __FUNCTION__, scrn_num, listen_ign);
                foreground[scrn_num]=listen_ign;
              }
              else if (0 == strcmp(argv[0], "key"))
              {
                debug(RPT_CLIUP,"%s Key %s", __FUNCTION__,  argv[1]);
              }
              else if(0 == strcmp(argv[0], "menu"))
              {
              }
              else if(0 == strcmp(argv[0], "connect"))
              {
                int a;
                for(a = 1; a < argc; a++)
                {
                  if(0 == strcmp(argv[a], "wid"))
                    lcd_wid = atoi(argv[++a]);
                  else if(0 == strcmp(argv[a], "hgt"))
                    lcd_hgt = atoi(argv[++a]);
                  else if(0 == strcmp(argv[a], "cellwid"))
                    lcd_cellwid = atoi(argv[++a]);
                  else if(0 == strcmp(argv[a], "cellhgt"))
                    lcd_cellhgt = atoi(argv[++a]);
                }
              }
              else if(0 == strcmp(argv[0], "bye"))
              {
                debug(RPT_NOTICE, "%s got : %s : exit", __FUNCTION__, argv[0]);
            struct timespec nod = {2, 0};
            nanosleep(&nod, NULL); // b4 e xit on client got exit 
                exit(EXIT_SUCCESS);
              }
              else if(0 == strcmp(argv[0], "success"))
              {
                want_success--;
                debug(RPT_CLIUP, "%s success: got %s [%u]",
                    __FUNCTION__, argv[0], want_success);
              }
              else
              {
                int j;
                for(j = 0; j < argc; j++)
                  debug(RPT_CLIUP, "%s got ?: %s", __FUNCTION__, argv[j]);
              }
            }

            // Restart tokenizing
            argc = 0;
            newtoken = 1;
          break;
        } // switch( buf[i] )
      } // end tokenizig one buf
      len = sock_recv( cf_dsk, buf, sizeof(buf)-1);
    } // end while len > 0
    // Update screens...
    for (i=active_screen*lcd_hgt; i< active_screen*lcd_hgt+lcd_hgt; i++)  {
      if ( i == lines ) break;
      int screen=i/lcd_hgt; // which screen # this is,  == active_screen
      int screen_line=(i+lcd_hgt)%lcd_hgt+1; // which line # on this screen
      if (foreground[screen]==1) {  // if this screen is visible now
        if (screen_line==1) updates[screen]++;
        if ( ((updates[screen]+7)%8)==0) { // once every 8 loops
          // ( 1 ) = 8 times a second
          // ((updates[screen]+3)%4)==0 = 2 times / second
          // ((updates[screen]+1)%2)==0 = 4 times / second
          updateline[i](drvthis, t_or_f_num[i]);
#ifdef PROTOCOL_CHEATING
          debug(RPT_PCHEAT, "%s client_text[%02d] = %s  server_widget[%02d]= %s",
             __FUNCTION__, i, client_text[i], i, server_widget[i]);
          memcpy(server_widget[i], client_text[i], lcd_wid);
          debug(RPT_PCHEAT, "%s displayed PCHEAT %d =  %s, t/f = %d",
            __FUNCTION__, i, server_widget[i] , t_or_f_num[i]);
#endif
#ifndef PROTOCOL_CHEATING    
          debug(RPT_CLIUP, "%s displayed %d = %s, t/f = %d",
            __FUNCTION__, i, client_text[i] , t_or_f_num[i]);
          snprintf(buf,sizeof(buf),"widget_set cf_log%d line%d 1 %d \"%s\"\n",
            screen,   // screen # (name)
            i,    // total line # (name)
            screen_line,   // line # this screen (value)
            client_text[i]);   //  text
          debug(RPT_CLIUP , "%s line %d rep %u %u update widget: %s",
              __FUNCTION__, i, updates[screen], updates[screen]/8, buf );
          if (sock_send_string( cf_dsk, buf )< strlen(buf) ){
            report(RPT_CRIT, "%s exit on socket write error.", __FUNCTION__);
            struct timespec nod = {2, 0};  nanosleep(&nod, NULL); // b4 exit on error 
            exit(EXIT_FAILURE);
          }
          want_success++;
#endif // #ifndef PROTOCOL_CHEATING
        }
      } // end line [i] is foreground
    } // end for each line [i]
#endif // CLIENT_CHILD
  } // end while(!Quit) 125000000 nsec loop
} // end client_loop

#ifdef DO_HDD
static void hdd_init(Driver *drvthis){
  PrivateData * p = drvthis->private_data;
  
  struct timespec nod={1,000000000};
  int hdd_sock, count=0;
  do {
    hdd_sock = sock_connect ("localhost", 7634);
    if (hdd_sock <=0 ) {
      //if ((count+1)%50==0) 
        report(RPT_CLICRE, "%s waiting for hddtemp daemon ... %d",
            __FUNCTION__, count +1);
      nanosleep(&nod, NULL); // hddtemp connect
      count++;
    }
  } while ((hdd_sock <=0 ) && (count < 4));
  
  if (hdd_sock <=0 ) {
    report(RPT_ERR, "%s failed to open hddtemp on port %d", __FUNCTION__, 7634);
    p->hdd_count=0;
    return;
  }
  else {
    report(RPT_CLICRE, "%s hddtemp daemon good connect on try %d",
          __FUNCTION__, count );
    int len;
    count=0;
    // from daemon_send_message in hddtemp/srs/deamon.c
    // appears 130 bytes hypothetical max per drive
    // I'm getting 26
    char hdd_in[132*MAX_HDD];
    //nod.tv_nsec=50000000;
    p->hdd=malloc(sizeof( CF_HDD* ) *MAX_HDD);
    do {
      nanosleep(&nod, NULL); // hddtemp wakeup
      len = sock_recv( hdd_sock, hdd_in, sizeof (hdd_in));
      count++;
    } while ((len <= 0) && (count < 200));
    if (len <=0 ) {
      report(RPT_ERR, "%s failed to read from hddtemp on port %d", __FUNCTION__, 7634);
      p->hdd_count=0;
      return;
    } 
    else {
      // "good read on try 1, |/dev/hda|ST3120024A|27|C|2c/devices/0-0290/temp1_input"
      if ( len < 512) hdd_in[len]='\0';
      else hdd_in[511]='\0';
      report(RPT_CLICRE, "%s hddtemp daemon good read on try %d, %s",
          __FUNCTION__, count -1, hdd_in);
    }
    if (len >0 ) p->hdd_rep_len=len+1;
    char * ptr=hdd_in;
    int word_of = 0;
    int drive=0;
    while (strstr (ptr, "|" )!=NULL){
      p->hdd[drive]=malloc(sizeof( CF_HDD));
      p->hdd[drive]->hddname=NULL; 
      ptr=strstr (ptr, "|" );
      while ( (word_of <4) && ((index(ptr, '|')!=NULL) && strlen(ptr)>1) )   {
        ptr=index(ptr, '|')+1;
        char aword[128];
        //strncpy(aword, ptr , index(ptr, '|') -ptr);
        strncpy(aword, ptr , index(ptr, '|') !=NULL ? index(ptr, '|')-ptr : strlen(ptr) );
        //aword[index(ptr, '|') -ptr ]='\0';
        aword[index(ptr, '|') !=NULL ? index(ptr, '|')-ptr : strlen(ptr)]='\0';
	if (word_of==0) { // if this is the( device )name
          p->hdd[drive]->hddname=strdup(aword); 
        }
	if (word_of==1) { // if this is the model name
          p->hdd[drive]->hddname=realloc(p->hdd[drive]->hddname, strlen(p->hdd[drive]->hddname) + strlen(aword)+2);
          snprintf (p->hdd[drive]->hddname+strlen(p->hdd[drive]->hddname), strlen(aword)+2, ":%s", aword);
        }
        if (word_of==2) { // if this is the temperature 
          p->hdd_count++;
          p->hdd[drive]->hd_abstracto=atoi(aword)*16;
        }
        word_of ++;
      } 
      if ((p->hdd[drive]!=NULL) && (p->hdd[drive]->hddname!=NULL)) {
        report(RPT_CLICRE, "%s drive %d name %s temp %f", 
          __FUNCTION__, drive, p->hdd[drive]->hddname, (float)p->hdd[drive]->hd_abstracto/16 );
        if (strlen(ptr)>1) {
          ptr=index(ptr, '|') + 1;
        } else {
          debug(RPT_CLICRE, "%s drive %d word %d '%s' and no next", __FUNCTION__,  drive, word_of, ptr );
          sock_close ( hdd_sock );
	  return;
        }
        drive ++;
        word_of=0;
        if (drive == MAX_HDD) ptr=NULL;
      } // end p->hdd[drive]->hddname  got a good drive name 
    } // end good drive read returned better than ptr==NULL
  sock_close ( hdd_sock );
  } // end good connect
  report(RPT_CLICRE, "%s read %d drives", __FUNCTION__, p->hdd_count);
  //p->hdd=malloc(sizeof( CF_HDD* ) *MAX_HDD);
  p->hdd=realloc(p->hdd, sizeof( CF_HDD* ) *p->hdd_count );
} // end hdd_init 

static void hddtemp_temps(Driver *drvthis, unsigned char number){
  PrivateData * p = drvthis->private_data;
  int hdd_sock, count=0;
  int len;
  struct timespec nod={0,250000000};
  do {
    hdd_sock = sock_connect ("localhost", 7634);
    if (hdd_sock <=0 ) {
      report(RPT_ERR, "%s waiting for hddtemp daemon ... %d",
          __FUNCTION__, count );
      nanosleep(&nod, NULL); // hddtemp connect
      count++;
    }
  } while ((hdd_sock <=0 ) && (count < 2));
  
  if (hdd_sock <=0 ) {
    report(RPT_ERR, "%s failed to open hddtemp on port %d", __FUNCTION__, 7634);
    //p->hdd_count=0;
    for (count=0; count < p->hdd_count; count++){
      p->hdd[count]->hd_abstracto=666*16;
    }
    return;
  }
  else {
    debug(RPT_CLIUP, "%s hddtemp daemon good connect on try %d",
          __FUNCTION__, count );
    count=0;
    char hdd_in[p->hdd_rep_len];
    do {
      len = sock_recv( hdd_sock, hdd_in, sizeof (hdd_in));
      if (len <= 0) nanosleep(&nod, NULL); // hddtemp wakeup
      count++;
    } while (len <= 0 && count < 3);
    if (len <=0 ) {
      report(RPT_ERR, "%s hddtemp read fail port %d -- temps not updated", __FUNCTION__, 7634);
      // FIXME: happens often enough.. do not want to linger here..
      //        dropping hdd messes up .csv file,
      //        if drop hdds, print a new header line ?
      //p->hdd_count=0;
      // for now, throw in dummy bad read value
      for (count=0; count < p->hdd_count; count++){
        p->hdd[count]->hd_abstracto=666*16;
      }
      return;
    } 
    else {
      if ( len < p->hdd_rep_len) hdd_in[len]='\0';
      else {
        hdd_in[p->hdd_rep_len-1]='\0';
        debug(RPT_NOTICE, "%s hddtemp daemon could use a larger socket read bufferr.",
          __FUNCTION__);
        }
      debug(RPT_CLIUP, "%s hddtemp daemon good read on try %d, %s",
          __FUNCTION__, count -1, hdd_in);
    }
    char * ptr=hdd_in;
    int word_of = 0;
    int drive=0;
    // why was this crashing when (ptr == NULL) ?
    while ((ptr != NULL) && (strstr (ptr, "|" )!=NULL)){
      ptr=strstr (ptr, "|" );
      do {
        if (strlen(ptr)<2){
          debug(RPT_CLIUP, "%s drive %d word %d '%s' and no next", __FUNCTION__,  drive, word_of, ptr );
          sock_close ( hdd_sock );
	  return;
	}
	ptr=index(ptr, '|')+1;
        if ((word_of==2) && (ptr!=NULL)) { // if this is the temperature 
          char aword[128];
	    strncpy(aword, ptr , index(ptr, '|') -ptr);
            aword[index(ptr, '|') -ptr ]   ='\0';
            if( strlen(aword) >0) {
	      p->hdd[drive]->hd_abstracto=(int)atoi(aword)*16;
              debug(RPT_CLIUP, "%s drive %d temp %f", 
                __FUNCTION__, drive, (float)p->hdd[drive]->hd_abstracto/16);
	      // if this is the spcified drive, force exit
              //if (drive==number) ptr=NULL;
	    } else { // strlen(aword) <=0
              debug(RPT_NOTICE, "%s drive %d value string is too short", 
                __FUNCTION__, drive );
              p->hdd[drive]->hd_abstracto=666*16;
	    }
	} // end is word 2
	if (ptr==NULL){
          debug(RPT_CLIUP, "%s drive %d word %d and pointer is null", __FUNCTION__,  drive, word_of );
	}
        word_of ++;
      } while (( word_of != 4 ) && (ptr  !=NULL));
      // end read if storage is full 
      if (drive+1 == p->hdd_count) {
        debug(RPT_CLIUP, "%s drive %d word %d is last drive.", __FUNCTION__,  drive, word_of );
      }
      if (strlen(ptr)>1) {
        ptr=index(ptr, '|') + 1;
      } else {
        debug(RPT_INFO, "%s drive %d word %d '%s' and no next", __FUNCTION__,  drive, word_of, ptr );
        sock_close ( hdd_sock );
	return;
      }
      drive ++;
      word_of=0;
    } // end good drive read
  sock_close ( hdd_sock );
  } // end good connect
} // end hddtemp_ temps 
#endif // DO_HDD

#ifdef MY_CUSTOM_LINE
static void update_wattsline (Driver *drvthis, unsigned char unused) {
    PrivateData * p = drvthis->private_data;
  static int check_once=0;
  if (check_once==0){
    check_once=1;
    if ( p->tempcount < 8) {
      // simple & easy & not gauranteed to catch a change check 
      report(RPT_ERR, "%s #ifdef MY_CUSTOM_LINE; edit update_wattsline() to something meaningful.",
          __FUNCTION__);

      snprintf(watts, 12, "?: undefined");
      return;
    }
      if (p->tempcount > 6 ) 
        report(RPT_ERR, "%s in 1  %s", __FUNCTION__, p->ds18[6]->name);
      else report(RPT_ERR, "%s there is no sensor 7", __FUNCTION__);

/* sensor quit 2009 09 
      if (p->tempcount > 7 ) 
        report(RPT_ERR, "%s in 2  %s", __FUNCTION__, p->ds18[7]->name);
      else report(RPT_ERR, "%s there is no sensor 8", __FUNCTION__);
*/

/* sensor quit 2008 05 
      if (p->tempcount > 8 ) 
        report(RPT_ERR, "%s in 3  %s", __FUNCTION__, p->ds18[8]->name);
      else report(RPT_ERR, "%s there is no sensor 9", __FUNCTION__);
*/

      if (p->tempcount > 0 ) 
        report(RPT_ERR, "%s out 1 %s", __FUNCTION__, p->ds18[0]->name);
      else report(RPT_ERR, "%s there is no sensor 1", __FUNCTION__);

      if (p->tempcount > 1 ) 
        report(RPT_ERR, "%s out 2 %s", __FUNCTION__, p->ds18[1]->name);
      else report(RPT_ERR, "%s there is no sensor 2", __FUNCTION__);
      if (p->tempcount > 2 ) 
        report(RPT_ERR, "%s out 3 %s", __FUNCTION__, p->ds18[2]->name);
      else report(RPT_ERR, "%s there is no sensor 3", __FUNCTION__);
  } /* end check once */
  if ( p->tempcount < 8) {
    sprintf(watts, "MY_CUSTOM_LINE");
    return;
  }
  // average of 3 temp probes measuring water out of computer into radiator
  // average of 3 temp probes measuring water into psu
  float in =(  (float)p->ds18[6]->abstracto);
//+
//      (float)p->ds18[7]->abstracto) /2;
      // (float)p->ds18[8]->abstracto) / 3;  
  // average of 3 temp probes measuring water out of radiator into computer
  // average of 3 temp probes measuring water out of psu 
  float out=(  (float)p->ds18[0]->abstracto+
      (float)p->ds18[1]->abstracto+ 
      (float)p->ds18[2]->abstracto) / 3;
  // cooling
  // float dt=(in-out)/16; // degrees celcius delta temp
  // heat
  float dt=(out - in)/16; // degrees celcius delta temp
  float mlps=compute_fan_rpm(drvthis, 1); // milliliters of water per second
  float w = dt * mlps *4.2; // 1 calorie = raise 1 ml water 1 deg c
        // 4.2 joules / calorie ; joules / second = wattage
/*
  report(RPT_ERR, "%s in=%f, out=%f, mlps=%f, w=%f", 
    __FUNCTION__, in/16, out/16, mlps, w);
*/
  /* cue the laser lights and chorus;
   *   tapdance with the theory of significant digits
   * and manufacturers' specs for accuracy & repeatability;
   * pay the laboratory staff for measuring the thermal conductivity
   * and thermal capacitance of today's coolant mixture;
   * blow big kisses to degree of certainty & margin of error; */
  int w_lh=(int)w;
  int w_rh=(int) ((w-(float)w_lh)*100);
  //sprintf(watts+2, "% 3d.%02d        ", w_lh, w_rh);
  snprintf(watts+2,p->width-3, "% 3d.%02d        ", w_lh, w_rh);
#ifdef DO_ULT
  float l=p->load_t;
  int l_lh=(int)l;
  int l_rh=(int) ((l-(float)l_lh)*100);
  snprintf(watts+p->width-7, 8, "% 3d.%02d%%",  l_lh, l_rh );
#endif // DO_ULT
// #ifdef PROTOCOL_CHEATING  // FIXME: PROTOCOL_CHEATING for MY_CUSTOM_LINE ?  
//  snprintf((char *)(wattsptr), strlen(watts), "%s", watts);
// #endif
  watts_est =w;
} // end update_wattsline
#endif // MY_CUSTOM_LINE

/*
 * *****************************************************
 * everything below here is in CLIENT_CHILD ifdef
 * *****************************************************
 */
#ifdef CLIENT_CHILD /* cli only func begin */
int csv_init (Driver * drvthis){
    /* valgrind balked 3 places in csv_init */
    /* init is portion of cliont that can be done before server is up and running 
       before client has connected */
    // this uses global p->width, used to use lcd_wid global
    // that was only set after client connected 
  PrivateData * p = drvthis->private_data;
  //debug(RPT_INFO, "%s started", __FUNCTION__);
  char target[128];
  int fan, temp, i, j;
  char *spaces=NULL;
        // add blank line(s) as option to organize lines ?
  sprintf(target, "FastLcd");
  p->prettylcd=drvthis->config_get_int(drvthis->name, target, 0, 0);
  if (p->prettylcd ==0) {
    report(RPT_INFO, "%s FastLcd=0: load & top stats measured favoring logfile precision.",
      __FUNCTION__);
  } else {
    report(RPT_INFO, "%s FastLcd enabled: load & top stats measured favoring fast display.",
      __FUNCTION__);
  }
  sprintf(target, "HeartBeat");
  p->bool_heartbeat=drvthis->config_get_int(drvthis->name, target, 0, 0);
  if (p->bool_heartbeat==0) {
    report(RPT_INFO, "%s HeartBeat=0: driver-client screens without heartbeat.",
      __FUNCTION__);
  } else {
    report(RPT_INFO, "%s HeartBeat=1: driver-client screens with heartbeat.",
      __FUNCTION__);
  }
  spaces=malloc(p->width+1); 
  for (i=0; i<p->width; i++)
    spaces[i]=' ';  
    spaces[p->width]='\0';  
  for (i=0; i<MAXLINES; i++){
     order[i]=-1;
     client_text[i]=NULL;     
#ifdef PROTOCOL_CHEATING
     server_widget[i]=(char *) NULL;
     // =NULL was working once what has changed ?
#endif // PROTOCOL_CHEATING
  }
  for (temp=0; temp< p->tempcount; temp++) {
    p->ds18[temp]->line=NULL;
      sprintf(target, "%s_lcd", p->ds18[temp]->name);  
      int tmp_wid=drvthis->config_get_int
        (drvthis->name, target, 0, -1);
      if (tmp_wid==-1){ 
        sprintf(target, "t%02d_lcd", temp+1);  
        tmp_wid=drvthis->config_get_int
          (drvthis->name, target, 0, -1);
      }
      if (( DEFAULT_WIDGET!=0) && (tmp_wid==-1)){
        p->ds18[temp]->dowidget=500+10*temp;
      }
      else {
        p->ds18[temp]->dowidget=tmp_wid;
      }
      /* valgrind */
      report(RPT_DEBUG, "%s sensor %d name %s dowidget %d.",
          __FUNCTION__, temp+1, p->ds18[temp]->name,
           p->ds18[temp]->dowidget);
  } // end each sensor
  for (fan=0; fan <4; fan++) {
      sprintf(target, "f%d_lcd", fan+1);  
      i=drvthis->config_get_int(drvthis->name, target, 0, -1);
      if ((DEFAULT_WIDGET!=0) && (i==-1))
          p->fans[fan].dowidget=700+10*fan;
      else p->fans[fan].dowidget=i;
     if (p->fans[fan].dowidget!=0) {
      p->fans[fan].line=malloc(p->width+1);
/*       Invalid read of size 1 csv_init (CF63oddtf.c:5979)
      snprintf(p->fans[fan].line, p->width+1, "%s%s",
          p->fans[fan].name, spaces);
*/
      snprintf(p->fans[fan].line, p->width+1, "%s%s\n",
          p->fans[fan].name, spaces);
      snprintf((p->fans[fan].line)+p->width-7, 8, "not-yet");
      order[lines]=(p->fans[fan].dowidget)*10;
      client_text[lines]=p->fans[fan].line;
      t_or_f_num[lines]=fan;
      updateline[lines]=*set_up_fanline;
      report(RPT_CLICRE, "%s %02d f%d       % 4d %s",
          __FUNCTION__, lines+1, fan+1, order[lines], p->fans[fan].line);
      lines++;
      if (p->fans[fan].f_sens[0] == NULL) {
        p->fans[fan].line2=NULL;
      }
      else { // temp associated
        p->fans[fan].line2=malloc(p->width+1);
        order[lines]=(p->fans[fan].dowidget)*10+1;
        client_text[lines]=p->fans[fan].line2;
        // without separate set_up_fan.line2 func
        t_or_f_num[lines]=fan;
        updateline[lines]=*set_up_fanline;
        //if  (p->fans[fan].f_sens[1]->name==NULL) { // simple temp
        if  (p->fans[fan].f_sens[1]==NULL) { // simple temp
          snprintf(p->fans[fan].line2, p->width+1, "%s%s",
              p->fans[fan].f_sens[0]->name, spaces);
        }
        else { // must be a relative temp
          if ( (p->fans[fan].dt_label==NULL) 
               || (strlen(p->fans[fan].dt_label)==0) ) {
            snprintf(p->fans[fan].line2, p->width+1, "%s%s",  
              p->fans[fan].f_sens[0]->name, spaces);
            snprintf(p->fans[fan].line2+4, p->width-3, "%s%s",
              p->fans[fan].f_sens[1]->name, spaces);
          } else // have a good p->fans[fan].dt_label
          {
            snprintf(p->fans[fan].line2, p->width+1, "%s%s",  
              p->fans[fan].dt_label, spaces);
          }
        }
        snprintf((p->fans[fan].line2)+p->width-7, 8, "not-yet");
        report(RPT_CLICRE, "%s %02d f%d l 2   % 4d %s",
            __FUNCTION__, lines+1, fan+1, order[lines], p->fans[fan].line2);
        lines++;
      } // end has assocated temp
     } // end fan has LCD line(s)
     else p->fans[fan].line=NULL;
  } // end each fan
  
  for (temp=0; temp < p->tempcount; temp ++) {
    if ( p->ds18[temp]->dowidget!=0) {  
      p->ds18[temp]->line= malloc(p->width+1);
      order[lines]=(p->ds18[temp]->dowidget)*10;
      client_text[lines]=p->ds18[temp]->line;
      t_or_f_num[lines]=temp;
      updateline[lines]=*set_up_templine;
      snprintf(p->ds18[temp]->line, p->width+1, "%s%s",
          p->ds18[temp]->name, spaces);
      snprintf((p->ds18[temp]->line)+p->width-7, 8, "not-yet");
      report(RPT_CLICRE, "%s %02d T%02d      % 4d %s",
        __FUNCTION__, lines+1, temp+1, order[lines], p->ds18[temp]->line);
      lines++;
      }
    else p->ds18[temp]->line=NULL;
  } // end each temp 
#ifdef DO_ULT
  // use of p->tdid_tlog presumes p->tdid_tlog set earlier in init 
  for (temp=0; temp < p->topname; temp ++) {
    if (p->top_name[temp]->lcd!=0){
      p->top_name[temp]->line=malloc(p->width+1);
      order[lines]=p->top_name[temp]->lcd*10;
      client_text[lines]=p->top_name[temp]->line;
      t_or_f_num[lines]=temp;
      updateline[lines]=*set_up_top_name_line;
      snprintf(p->top_name[temp]->line, p->width+1, "%s%%%s",
          p->top_name[temp]->name, spaces);
      snprintf((p->top_name[temp]->line)+p->width-7, 8, "not-yet");
      report(RPT_CLICRE, "%s %02d TopName%d % 4d %s",
          __FUNCTION__, lines+1, temp+1, order[lines], p->top_name[temp]->line);
      lines++;
    } // end this p->top_name[temp]->name has a line
  } // end for each topname
#endif // DO_ULT
#ifdef DO_LM
  for (temp=0; temp < p->lmnum; temp ++) {
    if (p->lm[temp]->lmlcd!=0){
      p->lm[temp]->lmline=malloc(p->width+1);
      order[lines]=p->lm[temp]->lmlcd*10;
      client_text[lines]=p->lm[temp]->lmline;
      t_or_f_num[lines]=temp;
      updateline[lines]=*set_up_lmline;
      snprintf(p->lm[temp]->lmline, p->width+1, "%s%s",
          p->lm[temp]->lmlabel, spaces);
      snprintf((p->lm[temp]->lmline)+p->width-7, 8, "not-yet");
      report(RPT_CLICRE, "%s %02d lm%d      % 4d %s",
          __FUNCTION__,  lines+1, temp+1, order[lines], p->lm[temp]->lmline);
      lines++;
    } // end p->lm[temp]->lmlcd!=0 
  } // end for each topname
#endif // DO_LM
#ifdef DO_HDD
  for (temp=0; temp < p->hdd_count; temp ++) {
    p->hdd[temp]->hddline=NULL; 
    sprintf(target,  p->hdd[temp]->hddname);
    snprintf(index(target, ':') , 5  , "_Lcd");
    p->hdd[temp]->hddlcd=drvthis->config_get_int(drvthis->name, target, 0, 0);
    if (p->hdd[temp]->hddlcd!=0) {
      p->hdd[temp]->hddline=malloc(p->width+1);
      order[lines]=p->hdd[temp]->hddlcd*10;
      client_text[lines]=p->hdd[temp]->hddline;
      t_or_f_num[lines]=temp;
      updateline[lines]=*set_up_hddline;
      snprintf(p->hdd[temp]->hddline, p->width+1, "%s%s",
        p->hdd[temp]->hddname, spaces);
      snprintf((p->hdd[temp]->hddline)+p->width-8, 9, "       x");
      report(RPT_CLICRE, "%s %02d hdd%d     % 4d %s",
        __FUNCTION__, lines+1, temp+1, order[lines], p->hdd[temp]->hddline);
      lines++;
    } // end this [temp] might have a line
  } // end for each hdd
#endif // DO_HDD
  
/*************************************************************
 * Add any more lines here
 * lines = # of current lines: add line [lines] then lines++
 *          order[lines]=numeric order for this line
 *      client_text[lines]=string in driver client
 * server_widget[lines]=wiget->text that a sock_send would put  client_text[lines] into
 *            order[MAXLINES]
 *      client_text[MAXLINES]     globals at top of Cfontz633_c.c
 *    server_widget[MAXLINES]  
 *       MAXLINES   must be increasd for more lines 
 * Creation of actual widgets & screens is 
 * done below automatically.
 * **********************************************************/

#ifdef MY_CUSTOM_LINE
  watts=malloc(p->width+1);
  snprintf(watts, p->width+1, "w:%s", spaces);
  order[lines]=210;
  client_text[lines]=watts;
  t_or_f_num[lines]=0; //unused here
  updateline[lines]=*update_wattsline;
  lines++;
#endif // MY_CUSTOM_LINE
  
  if (lines==0){
    return 1;
  }
  // sort the lcd lines
  for (i=1; i < lines; i ++) {
    for (j=0; j < i+1; j ++) {
      if (order[j] > order[i]){
        int swaporder=order[j];
        char * swapline=client_text[j];
        unsigned char t_or_f_tmp= t_or_f_num[j];
        void *udl_tmp=updateline[j];
#ifdef PROTOCOL_CHEATING
        char *swapwidget=server_widget[j];
#endif
        
        order[j]=order[i];
        client_text[j]=client_text[i];
        t_or_f_num[j]=t_or_f_num[i];
        updateline[j]=updateline[i];
#ifdef PROTOCOL_CHEATING
        server_widget[j]=server_widget[i];
#endif
        
        order[i]=swaporder;
        client_text[i]=swapline;
        t_or_f_num[i]=t_or_f_tmp;
        updateline[i]=udl_tmp;
#ifdef PROTOCOL_CHEATING
        server_widget[i]=swapwidget;
#endif
      }
    }
  } // done sorting
  for (i=0; i < lines; i++){
    if ( order[i] ==  order[(i-1)] )
      report(RPT_INFO < RPT_CLICRE ? RPT_INFO : RPT_CLICRE, 
          "%s duplicate ordering %d %s; %d %s",
          __FUNCTION__, order[i],  client_text[i], order[(i-1)], client_text[(i-1)]);
  }
  free(spaces);
  return 1;
} // end csv_init

/*
 * wait on socket for "success" from server
 * other = string that also signals sucess, 
 * when server switches to a new screen, the "listen newscreen" 
 * message knocks out / supresses / replaces an expected success
 */
int wait_success(Driver * drvthis, int options, char **other)
{
  char buf[256]; /* checkme what is the lagest that might be sent ? */
                 /* and if this reads it in bits, is that handled */
  int len=0;
  int a;
  char * argv[20];
  int argc;
  short received = 0;
  short timeout = 5000; /* Give the server 5 secs to respond */
  long timeout_mult = 010000000;
  char success=0;
  short repeat=0;
  struct timespec nod={0, timeout_mult}; 
  
  while (repeat < 10000 ) {
    buf[0] = 0;
    while( !received && timeout > 0) {
      len = sock_recv_string ( cf_dsk, buf, sizeof(buf));
      if( len == 0 ) {
        nod.tv_nsec=timeout_mult; nanosleep(&nod, NULL); // waiting for success from server socket
        timeout --;
      } else {
        received = 1;
      }
    }
    if( len <= 0 ) {
      report(RPT_ERR, "%s Received (len <= 0) repeat [%d].", __FUNCTION__, repeat );
    }
    argc = split( buf, ' ', argv, 20 );
    for (a = 0; a < argc; a++) {
      if (0 == strcmp (argv[a], "success")){
        success=1;
        report(RPT_DEBUG, "%s received  expected: %s repeat %d & return",
            __FUNCTION__, argv[a], repeat);
        return 0;
      }
      else if (other!=NULL) {
        int i;
        for (i=0;  i <= options ; i++){
          // report(RPT_CLICRE, "%s alternate check : %s", __FUNCTION__, other[i]);
          if (0 == strcmp (argv[a], other[i])){
            success=1;
            report(RPT_DEBUG, "%s received alternate: %s  repeat %d & return",
                __FUNCTION__, argv[a], repeat);
            return 0;
          } // end checking alternates
        }
      }
      else if (strlen(argv[a]) !=0) {
        report(RPT_DEBUG, "%s received something else ?: %s repeat %d",
            __FUNCTION__, argv[a], repeat);
      }
    }
    if(success==0){
      repeat ++;
      if (repeat %2000==0) {
        report(RPT_WARNING, "%s did not receive a success, repeat [%d].", __FUNCTION__, repeat);
        }
    }
    else if(success==1){
      return 1;
    }
    nod.tv_nsec=1000000; nanosleep(&nod, NULL); // wait_success, cli waiting on socket response
  } // end of repeat for multiple responses
  report(RPT_ERR, "%s did not receive a success, repeat [%d] and fail.", __FUNCTION__, repeat );
  return 0;
} // end of wait_success 

int cli_scrn(Driver * drvthis){
  PrivateData * p = drvthis->private_data;
  /* p is used only once in this function, for heartbeat */
  int  i, j;
  for (i=0; i < lines; i++){
    // this uses global lcd_hgt, not p->height
    report(RPT_CLICRE, "%s scn %02d l %02d  % 5d  %s%s%s",
          __FUNCTION__, i/lcd_hgt +1, i+1, order[i],
          ( (i/lcd_hgt +2) %2 == 0 ) ? " " : "|",
          client_text[i], ( (i/lcd_hgt +2) %2 == 0 ) ? " " : "|" );
  }

  // add the screens
  char *other[]={"fixME", "elsewhere", "listen", "ignore", "cf_log"};
  for (i=0; i < lines; i ++) { // for each line to be created
    char buf[64];
    j=i/lcd_hgt; // j is screen #, starting with 0 
    report(RPT_DEBUG, "%s line %d screen %d", __FUNCTION__, i, j);
    // when server switches to newly added screen,
    // the "listen" appears to knock out an expected success
    if ((i-lcd_hgt*j)==0){ // is 1st in new screen
      snprintf (buf, sizeof(buf)-1, "screen_add cf_log%d\n", j); 
      report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
      sock_send_string( cf_dsk, buf );
      wait_success(drvthis, 5, other);
      if (p->bool_heartbeat == 0 ) {
        /* turn off heartbeat for driver client screens */

        snprintf (buf, sizeof(buf)-1, "screen_set cf_log%d heartbeat off\n", j);
        //report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
        sock_send_string( cf_dsk, buf );
        wait_success(drvthis, 5, other);
      }
      // hide all screens untill all done
      //snprintf (buf, sizeof(buf)-1, "screen_set cf_log%d priority hidden\n", j);
      //report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
      //sock_send_string( cf_dsk, buf );
      //wait_success(drvthis, 5, other);
    } // end is new screen
    snprintf( buf, sizeof(buf)-1, "widget_add cf_log%d line%d string\n",j, i );
    report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
    sock_send_string( cf_dsk, buf );
    wait_success(drvthis, 5, other);
    snprintf( buf, sizeof(buf)-1, "widget_set cf_log%d line%d 1 %d \"%s\"\n",
        j,    // screen #
        i,    // line #
        (i+lcd_hgt)%lcd_hgt+1,   // line # on this screen 
        client_text[i] );  // text
    report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
    sock_send_string( cf_dsk, buf );
    wait_success(drvthis, 5, other);
  } // end for each line to be created
#ifdef PROTOCOL_CHEATING
  Client * cf_client_cheat=NULL;
  int tryme=1 ;
  do {
    // changed from sock to sock -1 CVS-stable-0-5-x-20060415 ?
    // if initing during boot & things are slow in server, we want sock, not sock -1 ????
    // could make oddtf.c gcc dependent on sock.c? server.c and address globals directly ?
    cf_client_cheat= clients_find_client_by_sock (tryme );
    if ( cf_client_cheat!=NULL ){ 
      debug(RPT_PCHEAT, "%s clients_find_client_by_sock (connect %d / find %d) = %s   %d",
          __FUNCTION__, cf_dsk, tryme, (cf_client_cheat==NULL) ? "null" : cf_client_cheat->name, cf_client_cheat); 
      if (strncmp (cf_client_cheat->name, "CF633log", 8 ) != 0 ){
        cf_client_cheat=NULL;
        tryme++;
      }
    }
    else tryme++;
  } while ((cf_client_cheat==NULL) && (tryme != 255)) ;
  report(RPT_PCHEAT, "%s clients_find_client_by_sock (connect %d / find %d) = %s  sock or sock -1 ? %d",
        __FUNCTION__, cf_dsk, cf_client_cheat->sock, cf_client_cheat->name, cf_client_cheat);
  Screen *cf_screen_cheat=NULL;
  struct timespec nod= {0, 010000000};
  for (i=0; i < lines; i ++) { // for each line to be created
    char buf[64];
    j=i/lcd_hgt; // j is screen #, starting with 0 
    //report(RPT_PCHEAT, "%s PROTOCOL_CHEATING line %d screen %d", __FUNCTION__, i, j);
    // when server switches to newly added screen,
    // the "listen" appears to knock out an expected success
    if ((i-lcd_hgt*j)==0){ // is 1st in new screen  
      snprintf (buf, sizeof(buf)-1, "cf_log%d", j);
      cf_screen_cheat=NULL;  
      do {
        cf_screen_cheat=client_find_screen (cf_client_cheat, buf);
        // is hanging when cf_screen_cheat=0, screen not found ?
        // wait_success failed ?
	nanosleep(&nod, NULL); // untill cf_screen_cheat != NULL
      } while (cf_screen_cheat==NULL);
      report(RPT_PCHEAT, "%s (Screen *)cf_screen_cheat=%d",
          __FUNCTION__, cf_screen_cheat);
    } // end is new screen
    Widget * cf_widget_cheat=NULL;
    snprintf (buf, sizeof(buf)-1, "line%d", i);
    do {
      cf_widget_cheat=screen_find_widget (cf_screen_cheat, buf);
      if (cf_widget_cheat==NULL) usleep(10000); // idle until server creates cf_widget_cheat
    } while ( cf_widget_cheat==NULL);
    while (cf_widget_cheat->text==NULL) {
      nanosleep(&nod, NULL); // idle until server creates cf_widget_cheat->text
      report(RPT_PCHEAT, "%s cf_widget_cheat line %02d until server creates cf_widget_cheat->text",
          __FUNCTION__, i);
    }
    while (strncmp(cf_widget_cheat->text, client_text[i], lcd_wid)) {
      nanosleep(&nod, NULL);  // idle untill server fills in cf_widget_cheat->text
      report(RPT_PCHEAT, "%s cf_widget_cheat line %02d not NULL but not filled in SUSPICION ?", __FUNCTION__, i);
    }
    report(RPT_PCHEAT, "%s cf_widget_cheat=%d line %02d    |%s|",
        __FUNCTION__, cf_widget_cheat, i, cf_widget_cheat->text);
    // record pointer to widget.text
    server_widget[i]=cf_widget_cheat->text;
    report(RPT_PCHEAT, "%s PCHEAT (int)server_widget[%02d])=  %d     |%s|",
        __FUNCTION__, i, server_widget[i], server_widget[i] );
    // nod.tv_sec=2; nanosleep(&nod, NULL);  // dbg delay @ end of get cf_widget_cheat per line
  } // end for each line to be created
#endif // PROTOCOL_CHEATING
  report(RPT_CLICRE, "%s creating LCDd %d lines in %d screens.",
      __FUNCTION__, lines, (lines-1)/lcd_hgt+1);
/*  report(RPT_CLICRE, "%s unhiding screens...",__FUNCTION__);
  for (i=0; i < lines; i+=lcd_hgt) {
    j=i/lcd_hgt; // j is screen #, starting with 0 
    char buf[64];
    //  all done
    snprintf (buf, sizeof(buf)-1, "screen_set cf_log%d priority info\n", j);
    report(RPT_DEBUG, "%s sock_send: %d %s", __FUNCTION__, cf_dsk, buf);
    sock_send_string( cf_dsk, buf );
    //wait_success(drvthis, 0, NULL);
    wait_success(drvthis, 5, other);
  }
  for (i=0; i < lines; i+=lcd_hgt) {
    //wait_success(drvthis, 0, NULL);
  }
*/  return 1;  
} // end cli_scrn

int read_connect_string(Driver * drvthis)
{
  char buf[8192];
  int len=-1;
  int a;
  char * argv[20];
  int argc;
  short received = 0;
  short timeout = 50; /* Give the server 5 secs to respond */
  struct timespec nod={0, 90000000};
  buf[0] = 0;

  while( !received && timeout > 0) {
    len = sock_recv_string ( cf_dsk, buf, sizeof(buf));
    if( len == 0 ) {
      nanosleep(&nod, NULL); // read_connect_string len==0, try again
      timeout --;
    } else {
      received = 1;
    }
  }
  report(RPT_CLICRE, "%s : %s", __FUNCTION__, buf);

  if( len <= 0 ) {
    report(RPT_ERR, "%s Did not receive LCDd connect response.", __FUNCTION__);
    return -1;
  }

  argc = split( buf, ' ', argv, 20 );
  for (a = 1; a < argc; a++) {
    if (0 == strcmp (argv[a], "wid"))
      lcd_wid = atoi (argv[++a]);
    else if (0 == strcmp (argv[a], "hgt"))
      lcd_hgt = atoi (argv[++a]);
  }
  if( argc <= 0 || strcmp( argv[0], "connect" ) != 0
      || lcd_wid == 0 || lcd_wid == 0 ) {
          report(RPT_ERR, "%s Received invalid LCDd connect response.", __FUNCTION__);
    return -1;
  }
  report(RPT_CLICRE, "%s Client Connect Received lcd_wid=%d  lcd_hgt=%d.",
        __FUNCTION__, lcd_wid, lcd_hgt);
  return 1;
} // end of read_connect_string

int split( char * str, char delim, char * parts[], int maxparts )
/* Splits a string into parts, to which pointers will be returned in &parts.
 * The return value is the number of parts.
 * maxparts is the maximum number of parts returned. If more parts exist
 * they are (unsplit) in the last part.
 * The parts are split at the character delim.
 * No new space will be allocated, the string str will be mutated !
 */
{
  char * p1 = str;
  char * p2;
  int part_nr = 0;

  /* Find the delim char to end the current part */
  while( part_nr < maxparts - 1 && (p2 = strchr( p1, delim )) ) {

    /* subsequent parts... */
    *p2 = 0;
    parts[part_nr] = p1;

    p1 = p2 + 1; /* Just after the delim char */
    part_nr ++;
  }
  /* and the last part... */
  parts[part_nr] = p1;
  part_nr ++;

  return part_nr;
} // end of split

#ifdef DO_ULT
static void
set_up_top_name_line (Driver *drvthis, unsigned char topn)
{
    PrivateData * p = drvthis->private_data;
  if (p->prettylcd!=0) {
    // could possibly run store tops (p->height*8) times every second
    // instead of once every p->syslogsensors seconds
    store_tops(drvthis);  /* in set_up_top_name_line */
  }
  if ((p->top_name[topn]!=NULL) &&(p->top_name[topn]->line!=NULL)) { 
    // if this topname[topn] has a string for log / lcd line 
    snprintf(p->top_name[topn]->line+strlen(p->top_name[topn]->line)-7, 8, "%f        ", 
        p->top_name[topn]->val );
    debug(RPT_CLIUP, "%s update p->top_name[%02d]->line: %s",
        __FUNCTION__, topn, p->top_name[topn]->line);
  }
  else {
    debug(RPT_CLIUP, "%s update p->top_name[%02d]->line: NULL",
        __FUNCTION__, topn);
  }
} // end set_up_top_name_line
#endif // DO_ULT

#ifdef DO_LM
static void
set_up_lmline (Driver *drvthis, unsigned char lmn)
{
  PrivateData * p = drvthis->private_data;
    //report(RPT_ERR, "%s update 0 p->lm[%d]->lmline %f: |%s| %d",
    //    __FUNCTION__, lmn, (float)p->lm[lmn]->lm_abstracto/16,  p->lm[lmn]->lmline, strlen(p->lm[lmn]->lmline));
  if (p->prettylcd!=0) {
      // could possibly run sensors_get_feature (p->height*8) times every second
      // instead of once every p->syslogsensors seconds
    double value;
    int retval= sensors_get_value( p->lm[lmn]->chipn, p->lm[lmn]->subfnum, &value);
    if ( retval !=0 ) {
      value=666;  // error on read
    }
    p->lm[lmn]->lm_abstracto=(int)value*16;
  }
  if (p->lm[lmn]->lmline!=NULL) { 
    // if this lm[lmn] has a string for log / lcd line 
    snprintf(p->lm[lmn]->lmline+strlen(p->lm[lmn]->lmline)-7, 8, "%f", 
        (float)p->lm[lmn]->lm_abstracto/16 );
    //report(RPT_ERR, "%s update 1 p->lm[%d]->lmline %f: |%s| %d ",
    //    __FUNCTION__, lmn, (float)p->lm[lmn]->lm_abstracto/16,
    //    p->lm[lmn]->lmline, strlen(p->lm[lmn]->lmline));
    debug(RPT_CLIUP, "%s update p->lm[%d]->lmline %f: |%s| %d",
        __FUNCTION__, lmn, (float)p->lm[lmn]->lm_abstracto/16,  p->lm[lmn]->lmline, strlen(p->lm[lmn]->lmline));
  }
  else {
    debug(RPT_CLIUP, "%s update p->lm[%d]->lmline: NULL",
        __FUNCTION__, lmn);
  }
} // end set_up_lmline
#endif // DO_LM

#ifdef DO_HDD
static void
set_up_hddline (Driver *drvthis, unsigned char hdd)
{
  PrivateData * p = drvthis->private_data;
  if (p->hdd[hdd]->hddline!=NULL) {
    if (p->prettylcd!=0) {
      // could possibly run hddtemp_temps (p->height*8) times every second
      // instead of once every p->syslogsensors seconds
      hddtemp_temps(drvthis, UCHAR_MAX);
    }
    // if this has a string for log / lcd line 
    snprintf(p->hdd[hdd]->hddline+strlen(p->hdd[hdd]->hddline)-5, 6, "%f", 
        (float)p->hdd[hdd]->hd_abstracto/16 );
    debug(RPT_CLIUP, "%s update p->hdd[%d]->hddline: %s",
        __FUNCTION__, hdd,   p->hdd[hdd]->hddline);
  }
  else {
    debug(RPT_CLIUP, "%s update p->hdd[%d]->hddline: NULL",
        __FUNCTION__, hdd);
  }
} // end set_up_hddline
#endif // DO_HDD

static void
set_up_templine (Driver *drvthis, unsigned char temp)
{
  PrivateData * p = drvthis->private_data; 
  if ((p->ds18!=NULL) && (p->ds18[temp]->line!=NULL)) { 
    // if this temp has a string for log / lcd line 
    compute_temp(drvthis, temp);
    snprintf(p->ds18[temp]->line+strlen(p->ds18[temp]->line)-7, 8, "%f", 
        (float)(p->ds18[temp]->abstracto)/16 );
    debug(RPT_CLIUP, "%s update p->ds18[%d].line %d: %s",
        __FUNCTION__, temp, &(p->ds18[temp]->line) , p->ds18[temp]->line);
  }
  else {
    debug(RPT_CLIUP, "%s update p->ds18=%d, p->ds18[%d]->line %d: NULL",
        __FUNCTION__, temp, p->ds18, p->ds18[temp]->line);
  }
} // end set_up_templine

static void
set_up_fanline (Driver *drvthis, unsigned char fan)
{
  PrivateData * p = drvthis->private_data; 
  if (p->fans[fan].line!=NULL) { 
    // if this fan has a string for log / lcd line 
    snprintf((p->fans[fan].line)+strlen(p->fans[fan].line)-7, 8, "%f", 
        compute_fan_rpm(drvthis, fan) );
    debug(RPT_CLIUP, "%s update p->fans[%d].line  %d: %s",
        __FUNCTION__, fan, &(p->fans[fan].line), p->fans[fan].line);
  }
  else {
    debug(RPT_CLIUP, "%s update p->fans[%d].line  %d: NULL",
        __FUNCTION__, fan, &(p->fans[fan].line));
  }
  if ((p->fans[fan].line2)!=NULL) {
    // if this fan has an asociated temp widget
    // no updating temps here, fan dislay is of last values used
    // //if (p->fans[fan].f_sens[1]->name==NULL) { // if simple temp 
    if ((p->fans[fan].f_sens[0]!=NULL) && (p->fans[fan].f_sens[1]==NULL)) { 
       // if simple temp 
       snprintf((p->fans[fan].line2) +strlen(p->fans[fan].line2)-7, 8,
          "%f", (float)*p->fans[fan].f_sens[0]->abstvalptr/16 );
    }
    else if ((p->fans[fan].f_sens[0]!=NULL) && (p->fans[fan].f_sens[1]!=NULL)){
       // relative temp
       snprintf( (p->fans[fan].line2) +strlen(p->fans[fan].line2)-7, 8, "%f",
           (float)p->fans[fan].dt/16);
    }
    else if ((p->fans[fan].f_sens[0]==NULL) && (p->fans[fan].f_sens[1]==NULL)){
      // no sensors attacthed, must be a sensor failure */
       snprintf( (p->fans[fan].line2) +strlen(p->fans[fan].line2)-7, 8,
         "BAD CRC");
    }
    debug(RPT_CLIUP, "%s update p->fans[%d].line2 %d: %s",
        __FUNCTION__, fan, &(p->fans[fan].line2), p->fans[fan].line2);
  } 
  else {
    debug(RPT_CLIUP, "%s update p->fans[%d].line2 %d: NULL",
        __FUNCTION__, fan, &(p->fans[fan].line2));
  }
  // end has a temperature line2
} // end set_up_fanline

#endif // CLIENT_CHILD 
 
/* EOF */
